//-------------------------------------------------------------------------------------------------
// <copyright file="WixUserExperience.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

static const LPCWSTR WIXUX_WINDOW_CLASS = L"WixBurnUX";

enum WIXUX_STATE
{
    WIXUX_STATE_INITIALIZING,
    WIXUX_STATE_PREPARING,

    WIXUX_STATE_ACCEPT_LICENSE,
    WIXUX_STATE_SELECT_LOCATION,
    WIXUX_STATE_SELECT_FEATURES,
    WIXUX_STATE_MAINTENANCE,

    WIXUX_STATE_CACHING,
    WIXUX_STATE_CACHED,
    WIXUX_STATE_EXECUTING,
    WIXUX_STATE_EXECUTED,
    WIXUX_STATE_RESTART,
    WIXUX_STATE_ERROR,
};


enum WIXUX_CONTROL
{
    WIXUX_CONTROL_HEADER_IMAGE,
    WIXUX_CONTROL_TITLE_TEXT,
    WIXUX_CONTROL_SUBTITLE_TEXT,
    WIXUX_CONTROL_LICENSE_RICHEDIT,
    WIXUX_CONTROL_ACCEPT_CHECKBOX,
    WIXUX_CONTROL_TOP_MESSAGE_TEXT,
    WIXUX_CONTROL_BOTTOM_MESSAGE_TEXT,

    WIXUX_CONTROL_PRINT_BUTTON,
    WIXUX_CONTROL_ADVANCED_BUTTON,
    WIXUX_CONTROL_INSTALL_BUTTON,
    WIXUX_CONTROL_BACK_BUTTON,
    WIXUX_CONTROL_NEXT_BUTTON,
    WIXUX_CONTROL_CLOSE_BUTTON,

    WIXUX_CONTROL_REPAIR_BUTTON,
    WIXUX_CONTROL_UNINSTALL_BUTTON,

    WIXUX_CONTROL_PROGRESS_BAR,
    WIXUX_CONTROL_MESSAGE_TEXT,

    WIXUX_CONTROL_CANCEL_BUTTON,
};

enum WM_WIXUX
{
    WM_WIXUX_FIRST = WM_USER, // this enum value must always be first.

    WM_WIXUX_PREPARE_PACKAGES,
    WM_WIXUX_CACHE_PACKAGES,
    WM_WIXUX_EXECUTE_PACKAGES,
    WM_WIXUX_MODAL_ERROR,

    WM_WIXUX_LAST, // this enum value must always be last.
};


class CWixUserExperience : public IBurnUserExperience
{
public: // IUnknown
    IFACEMETHODIMP QueryInterface(
        __in const IID& riid,
        __out void** ppvObject
        )
    {
        HRESULT hr = S_OK;

        ExitOnNull(ppvObject, hr, E_INVALIDARG, "Invalid argument ppvObject");
        *ppvObject = NULL;

        if (::IsEqualIID(__uuidof(IBurnUserExperience), riid))
        {
            *ppvObject = static_cast<IBurnUserExperience*>(this);
        }
        else if (::IsEqualIID(IID_IUnknown, riid))
        {
            *ppvObject = static_cast<IUnknown*>(this);
        }
        else // no interface for requested iid
        {
            ExitFunction1(hr = E_NOINTERFACE);
        }

        AddRef();
    LExit:
        return hr;
    }


    IFACEMETHODIMP_(ULONG) AddRef()
    {
        return ::InterlockedIncrement(&this->m_cReferences);
    }


    IFACEMETHODIMP_(ULONG) Release()
    {
        long l = ::InterlockedDecrement(&this->m_cReferences);
        if (0 < l)
        {
            return l;
        }

        delete this;
        return 0;
    }


public: // IBurnUserExperience
    //
    // Initialize - ensure all the necessary objects are created/initialized.
    //
    IFACEMETHODIMP Initialize(
        __in IBurnCore* pCore,
        __in int nCmdShow
        )
    {
        HRESULT hr = S_OK;
        INITCOMMONCONTROLSEX icc = { 0 };

        // Initialize common controls.
        icc.dwSize = sizeof(icc);
        icc.dwICC = ICC_PROGRESS_CLASS /* | ICC_STANDARD_CLASSES (requires 0x0501 == _WIN32_WINNT*/;
        ::InitCommonControlsEx(&icc);

        hr = ThemeInitialize(m_hModule);
        ExitOnFailure(hr, "Failed to initialize theme manager.");

        pCore->AddRef();
        m_pCore = pCore;

        hr = CreateMainWindow();
        ExitOnFailure(hr, "Failed to create main window.");

    LExit:
        return hr;
    }


    //
    // Run - start pumping messages.
    //
    IFACEMETHODIMP Run()
    {
        HRESULT hr = S_OK;
        BOOL fRet = FALSE;
        MSG msg;

        while (0 != (fRet = ::GetMessage(&msg, NULL, 0, 0)))
        {
            if (-1 == fRet)
            {
                hr = E_UNEXPECTED;
                ExitOnFailure(hr, "Unexpected return value from message pump.");
            }
            else if (!::IsDialogMessage(m_hWnd, &msg))
            {
                ::TranslateMessage(&msg);
                ::DispatchMessage(&msg);
            }
        }

    LExit:
        return hr;
    }

    IFACEMETHODIMP_(void) Uninitialize()
    {
    }

    IFACEMETHODIMP_(int) OnPrepareBegin(
        __in DWORD cPackages
        )
    {
        return IDOK;
    }


    IFACEMETHODIMP_(int) OnPreparePackageBegin(
        __in_z LPCWSTR wzPackageId
        )
    {
        Trace(REPORT_STANDARD, "CWixUserExperience::OnPreparePackageBegin()");
        return IDOK; // be a good Boy Scout, always prepare.
    }


    IFACEMETHODIMP_(void) OnPreparePackageComplete(
        __in LPCWSTR wzPackageId,
        __in HRESULT hrStatus,
        __in PACKAGE_STATE state,
        __in REQUEST_STATE requested,
        __in ACTION_STATE execute,
        __in ACTION_STATE rollback
        )
    {
        Trace(REPORT_STANDARD, "CWixUserExperience::OnPreparePackageComplete()");
    }


    IFACEMETHODIMP_(void) OnPrepareComplete(
        __in HRESULT hrStatus
        )
    {
        Trace(REPORT_STANDARD, "CWixUserExperience::OnPrepareComplete()");
        if (SUCCEEDED(hrStatus))
        {
            WIXUX_STATE state = WIXUX_STATE_ACCEPT_LICENSE;  // TODO: factor in the install state of the main Wix.msi and if installed use "WIXUX_STATE_MAINTENANCE"
            SetState(state);

            // Quiet or passive display just goes straight into caching (if installing) or
            // executing (if uninstalling).
            if (BURN_DISPLAY_NONE == m_command.display || BURN_DISPLAY_PASSIVE == m_command.display)
            {
                m_fAcceptedLicense = TRUE;
                ::PostMessageW(m_hWnd, BURN_ACTION_INSTALL == m_command.action ? WM_WIXUX_CACHE_PACKAGES : WM_WIXUX_EXECUTE_PACKAGES, 0, 0);
            }
        }
    }


    IFACEMETHODIMP_(void) OnCacheComplete(
        __in HRESULT hrStatus
        )
    {
        if (SUCCEEDED(hrStatus))
        {
            SetState(WIXUX_STATE_CACHED);

            // After caching the packages, always execute them.
            ::PostMessageW(m_hWnd, WM_WIXUX_EXECUTE_PACKAGES, 0, 0);
        }
    }


    IFACEMETHODIMP_(int) OnExecuteBegin(
        __in DWORD cExecutingPackages
        )
    {
        return IDOK;
    }


    IFACEMETHODIMP_(int) OnExecutePackageBegin(
        __in LPCWSTR wzPackageId,
        __in BOOL fExecute
        )
    {
        Trace(REPORT_STANDARD, "CWixUserExperience::OnExecutePackageBegin()");
        return IDOK; // always execute.
    }


    IFACEMETHODIMP_(int) OnError(
        __in DWORD dwCode,
        __in_z LPCWSTR wzError,
        __in DWORD dwUIHint
        )
    {
        HRESULT hr = S_OK;
        int nModalResult = IDCANCEL; // assume we always cancel on error.

        if (BURN_DISPLAY_FULL == m_command.display)
        {
            nModalResult = ModalPrompt(WIXUX_STATE_ERROR, dwUIHint, wzError);
        }

        return nModalResult;
    }


    IFACEMETHODIMP_(int) OnProgress(
        __in DWORD dwProgressPercentage,
        __in DWORD dwOverallProgressPercentage
        )
    {
        HRESULT hr = S_OK;
        int nModalResult = IDOK;

        hr = ThemeSetProgressControl(m_pTheme, WIXUX_CONTROL_PROGRESS_BAR, dwOverallProgressPercentage);
        ExitOnFailure(hr, "Failed to set progress.");

    LExit:
        if (FAILED(hr))
        {
            nModalResult = IDERROR;
        }

        return nModalResult;
    }


    IFACEMETHODIMP_(void) OnExecutePackageComplete(
        __in LPCWSTR wzPackageId,
        __in HRESULT hrExitCode
        )
    {
        Trace(REPORT_STANDARD, "CWixUserExperience::OnExecutePackageComplete()");
        if (FAILED(hrExitCode))
        {
            m_fFailed = TRUE;
        }
    }


    IFACEMETHODIMP_(void) OnExecuteComplete(
        __in HRESULT hrStatus
        )
    {
        // Failure or quiet or passive display just close at the end, no questions asked.
        if (m_fFailed || BURN_DISPLAY_NONE == m_command.display || BURN_DISPLAY_PASSIVE == m_command.display)
        {
            ::PostMessageW(m_hWnd, WM_CLOSE, 0, 0);
        }
        else
        {
            if (SUCCEEDED(hrStatus))
            {
                OnProgress(100, 100); // ensure we're 100% complete.
            }

            SetState(WIXUX_STATE_EXECUTED);
        }
    }


    IFACEMETHODIMP_(BOOL) OnRestartRequired()
    {
        // If the user already said "OK", don't ask again.
        if (!m_fAllowRestart)
        {
            if (BURN_DISPLAY_FULL == m_command.display && BURN_RESTART_PROMPT == m_command.restart)
            {
                // TODO: Get restart message from somewhere

                int nModalResult = ModalPrompt(WIXUX_STATE_RESTART, MB_OKCANCEL, L"Restart required.  Allow restart?");
                m_fAllowRestart = (IDOK == nModalResult); // only allow restart if the user clicked OK.
            }
            else
            {
                m_fAllowRestart = (BURN_RESTART_ALWAYS == m_command.restart || BURN_RESTART_AUTOMATIC == m_command.restart || BURN_RESTART_PROMPT == m_command.restart);
            }
        }

        return m_fAllowRestart;
    }


    IFACEMETHODIMP GetControlText(
        __in DWORD dwControlID,
        __out LPWSTR* psczText
        )
    {
        HRESULT hr = ThemeGetTextControl(m_pTheme, dwControlID, psczText);

        return hr;
    }


private: // privates
    //
    // CreateMainWindow - creates the main install window.
    //
    HRESULT CreateMainWindow()
    {
        HRESULT hr = S_OK;
        WNDCLASSW wc = { 0 };
        DWORD dwWindowStyle = 0;
        LPWSTR sczThemePath = NULL;
        LPWSTR sczCommandLine = NULL;

        // load theme relative to stdux.dll.
        hr = PathRelativeToModule(&sczThemePath, L"thm.xml", m_hModule);
        ExitOnFailure(hr, "Failed to combine module path with thm.xml.");

        hr = ThemeLoadFromFile(sczThemePath, &m_pTheme);
        ExitOnFailure(hr, "Failed to load theme from thm.xml.");

        hr = m_pCore->GetCommandLineParameters(&sczCommandLine);

        if (S_OK == hr)
        {
            hr = ProcessCommandLine(sczCommandLine);
            ExitOnFailure(hr, "Unknown commandline parameters.");
        }

        if (NULL != m_sczLanguage)
        {
            hr = ThemeLoadLocFromFile(m_pTheme, m_sczLanguage, m_hModule);
            ExitOnFailure(hr, "Failed to localize from /lang.");
        }
        else
        {
            hr = ThemeLoadLocFromFile(m_pTheme, L"en-us.wxl", m_hModule);
            ExitOnFailure(hr, "Failed to localize from default language.");
        }

        // Register the window class and create the window.
        wc.style = 0;
        wc.lpfnWndProc = CWixUserExperience::WndProc;
        wc.cbClsExtra = 0;
        wc.cbWndExtra = 0;
        wc.hInstance = m_hModule;
        wc.hIcon = reinterpret_cast<HICON>(m_pTheme->hIcon);
        wc.hCursor = ::LoadCursor(NULL, IDC_ARROW);
        wc.hbrBackground = m_pTheme->rgFonts[m_pTheme->dwFontId].hBackground;
        wc.lpszMenuName = NULL;
        wc.lpszClassName = WIXUX_WINDOW_CLASS;
        if (!::RegisterClassW(&wc))
        {
            ExitWithLastError(hr, "Failed to register window.");
        }

        m_fRegistered = TRUE;

        // Calculate the window style based on the theme style and command display value.
        dwWindowStyle = m_pTheme->dwStyle;
        if (BURN_DISPLAY_NONE == m_command.display)
        {
            dwWindowStyle &= ~WS_VISIBLE;
        }

        m_hWnd = ::CreateWindowExW(0, wc.lpszClassName, m_pTheme->wzCaption, dwWindowStyle, CW_USEDEFAULT, CW_USEDEFAULT, m_pTheme->nWidth, m_pTheme->nHeight, HWND_DESKTOP, NULL, m_hModule, this);
        ExitOnNullWithLastError(m_hWnd, hr, "Failed to create window.");

    LExit:
        ReleaseStr(sczThemePath);
        ReleaseStr(sczCommandLine);
        ReleaseStr(m_sczLanguage);

        if (FAILED(hr))
        {
            DestroyMainWindow();
        }

        return hr;
    }


    //
    // DestroyMainWindow - clean up all the window registration.
    //
    void DestroyMainWindow()
    {
        if (m_hWnd)
        {
            ::CloseWindow(m_hWnd);
            m_hWnd = NULL;
        }

        if (m_fRegistered)
        {
            ::UnregisterClassW(WIXUX_WINDOW_CLASS, m_hModule);
            m_fRegistered = FALSE;
        }

        if (m_pTheme)
        {
            ThemeFree(m_pTheme);
            m_pTheme = NULL;
        }
    }


    //
    // ProcessCommandLine - process the provided command line arguments.
    //
    HRESULT ProcessCommandLine(
        __in_z LPCWSTR wzCommandLine
        )
    {
        HRESULT hr = S_OK;
        LPWSTR sczName = NULL;

        int argc = 0;
        LPWSTR* argv = ::CommandLineToArgvW(wzCommandLine, &argc);
        ExitOnNullWithLastError(argv, hr, "Failed to get command line.");

        for (int i = 0; i < argc; ++i)
        {
            if (argv[i][0] == L'-' || argv[i][0] == L'/')
            {
                if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"lang", -1))
                {
                    if (i + 1 >= argc)
                    {
                        ExitOnRootFailure(hr = E_INVALIDARG, "Must specify a language.");
                    }

                    ++i;

                    hr = StrAllocString(&m_sczLanguage, &argv[i][0], 0);
                    ExitOnFailure(hr, "Failed to copy language.");

                    hr = StrAllocConcat(&m_sczLanguage, L".wxl", 0);
                    ExitOnFailure(hr, "Failed to concatenate wxl extension.");
                }
            }
        }

    LExit:
        ReleaseStr(sczName);

        if (argv)
        {
            ::LocalFree(argv);
        }

        return hr;
    }


    //
    // ModalPrompt - prompt the user with a message box and return their choice.
    //
    int ModalPrompt(
        __in WIXUX_STATE modalState,
        __in DWORD dwUIHint,
        __in_z_opt LPCWSTR wzMessage
        )
    {
        switch (modalState)
        {
        case WIXUX_STATE_RESTART:
            dwUIHint |= MB_ICONWARNING;
            break;

        case WIXUX_STATE_ERROR:
            dwUIHint |= MB_ICONERROR;
            break;

        default:
            AssertSz(FALSE, "Unexpected modal state.");
        }

        return ::MessageBoxW(m_hWnd, wzMessage, m_pTheme->wzCaption, dwUIHint);
    }


    //
    // WndProc - standard windows message handler.
    //
    static LRESULT CALLBACK WndProc(
        __in HWND hWnd,
        __in UINT uMsg,
        __in WPARAM wParam,
        __in LPARAM lParam
        )
    {
        LRESULT lres = 0;
#pragma warning(suppress:4312)
        CWixUserExperience* pUX = reinterpret_cast<CWixUserExperience*>(::GetWindowLongPtr(hWnd, GWLP_USERDATA));

        switch (uMsg)
        {
        case WM_NCCREATE:
            {
            LPCREATESTRUCT lpcs = reinterpret_cast<LPCREATESTRUCT>(lParam);
            pUX = reinterpret_cast<CWixUserExperience*>(lpcs->lpCreateParams);
#pragma warning(suppress:4244)
            ::SetWindowLongPtr(hWnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(pUX));
            }
            break;

        case WM_NCHITTEST:
            if (pUX->m_pTheme->dwStyle & WS_POPUP)
            {
                return HTCAPTION; // allow windows with no frame to be moved by grabbing any non-control.
            }
            break;

        case WM_NCDESTROY:
            lres = ::DefWindowProc(hWnd, uMsg, wParam, lParam);
            ::SetWindowLongPtr(hWnd, GWLP_USERDATA, 0);
            return lres;

        case WM_CREATE:
            if (!pUX->OnCreate(hWnd))
            {
                return -1;
            }
            break;

        case WM_CLOSE:
            pUX->OnClose(hWnd);
            return 0;

        case WM_DESTROY:
            ::PostQuitMessage(0);
            break;

        case WM_DRAWITEM:
            ThemeDrawControl(pUX->m_pTheme, reinterpret_cast<LPDRAWITEMSTRUCT>(lParam));
            return TRUE;

        case WM_CTLCOLORSTATIC:
            {
            HBRUSH hBrush = NULL;
            if (ThemeSetControlColor(pUX->m_pTheme, reinterpret_cast<HDC>(wParam), reinterpret_cast<HWND>(lParam), &hBrush))
            {
                return reinterpret_cast<LRESULT>(hBrush);
            }
            }
            break;

        case WM_SETCURSOR:
            ThemeHoverControl(pUX->m_pTheme, hWnd, reinterpret_cast<HWND>(wParam));
            break;

        case WM_PAINT:
            // If there is anything to update, do so.
            if (::GetUpdateRect(hWnd, NULL, FALSE))
            {
                PAINTSTRUCT ps;
                ::BeginPaint(hWnd, &ps);
                ThemeDrawBackground(pUX->m_pTheme, &ps);
                ::EndPaint(hWnd, &ps);
            }
            return 0;

        case WM_WIXUX_PREPARE_PACKAGES:
            pUX->OnPrepare();
            return 0;

        case WM_WIXUX_CACHE_PACKAGES:
            pUX->OnCache();
            return 0;

        case WM_WIXUX_EXECUTE_PACKAGES:
            pUX->OnExecute();
            return 0;

        case WM_COMMAND:
            switch (LOWORD(wParam))
            {
            case WIXUX_CONTROL_ACCEPT_CHECKBOX:
                pUX->OnClickAcceptCheckbox();
                break;

            case WIXUX_CONTROL_INSTALL_BUTTON:
                pUX->OnClickInstallButton();
                break;

            case WIXUX_CONTROL_REPAIR_BUTTON:
                pUX->OnClickRepairButton();
                break;

            case WIXUX_CONTROL_UNINSTALL_BUTTON:
                pUX->OnClickUninstallButton();
                break;

            case WIXUX_CONTROL_CANCEL_BUTTON: __fallthrough; // cancel is the same as click the close button or "X" in the top right corner.
            case WIXUX_CONTROL_CLOSE_BUTTON:
                pUX->OnClickCloseButton();
                break;

            default:
                break;
            }
            return 0;
        }

        return ::DefWindowProc(hWnd, uMsg, wParam, lParam);
    }


    //
    // OnCreate - finishes loading the theme and tells the core we're ready for data.
    //
    BOOL OnCreate(
        __in HWND hWnd
        )
    {
        HRESULT hr = S_OK;

        hr = ThemeLoadControls(m_pTheme, hWnd);
        ExitOnFailure(hr, "Failed to load theme controls.");

        hr = ThemeLoadRichEditFromFile(m_pTheme, WIXUX_CONTROL_LICENSE_RICHEDIT, L"License.rtf", m_hModule);
        ExitOnFailure(hr, "Failed to load license.");

#ifndef BCM_SETSHIELD // TODO: don't define this here, use the correct headers to get the define.
#define BCM_SETSHIELD            (0x1600 + 0x000C)
#endif
        ::SendMessage(m_pTheme->rgControls[WIXUX_CONTROL_INSTALL_BUTTON].hWnd, BCM_SETSHIELD, 0, TRUE); // mark the install button as requiring elevation.

        // Okay, we're ready for packages now.
        SetState(WIXUX_STATE_PREPARING);

        ::PostMessageW(hWnd, WM_WIXUX_PREPARE_PACKAGES, 0, 0);

    LExit:
        return SUCCEEDED(hr);
    }


    //
    // OnClose - called on close and cancel, prompts for cancel if appropriate.
    //
    void OnClose(
        __in HWND hWnd
        )
    {
        if (WIXUX_STATE_EXECUTED == m_state || m_fFailed || m_fCanceled)
        {
            ::DestroyWindow(hWnd);
        }
        else // for all non-complete states, prompt before cancelling.
        {
            // Check with the user to verify they want to cancel.
            //PromptCancel(hWnd);
            m_fCanceled = TRUE; // TODO: let the prompt set this instead.

            // If the user chose cancel and we're not caching or executing, close the window now. Otherwise,
            // we'll have to let the progress system kick off a rollback which will eventually cause the
            // window to close.
            if (m_fCanceled && (WIXUX_STATE_CACHING > m_state || WIXUX_STATE_EXECUTING < m_state))
            {
                ::DestroyWindow(hWnd);
            }
        }
    }


    //
    // OnPrepare - start the processing of packages.
    //
    void OnPrepare()
    {
        HRESULT hr = S_OK;

        Trace(REPORT_STANDARD, "CWixUserExperience::OnPrepare()");

        // Tell the core we're ready for the packages to be processed now.
        hr = m_pCore->Prepare(NULL, BURN_ACTION_UNINSTALL == m_command.action ? REQUEST_STATE_ABSENT : REQUEST_STATE_PRESENT);
        ExitOnFailure(hr, "Failed to prepare chain.");

    LExit:
        return;
    }


    //
    // OnCache - cache the prepared changes.
    //
    void OnCache()
    {
        HRESULT hr = S_OK;

        Trace(REPORT_STANDARD, "CWixUserExperience::OnCache()");

        hr = m_pCore->Cache(NULL);
        ExitOnFailure(hr, "Failed to cache install packages.");

        SetState(WIXUX_STATE_CACHING);

    LExit:
        return;
    }


    //
    // OnExecute - execute the cached packages.
    //
    void OnExecute()
    {
        HRESULT hr = S_OK;

        Trace(REPORT_STANDARD, "CWixUserExperience::OnExecute()");

        hr = m_pCore->Execute(REQUEST_STATE_NONE);
        ExitOnFailure(hr, "Failed to execute install.");

        SetState(WIXUX_STATE_EXECUTING);

    LExit:
        return;
    }


    //
    // OnClickInstallButton - start the install by caching the packages.
    //
    void OnClickInstallButton()
    {
        HRESULT hr = S_OK;

        // Ensure the elevated process is started.
        do
        {
            hr = m_pCore->Elevate(m_hWnd);
            if (FAILED(hr))
            {
                int nResult = this->ModalPrompt(WIXUX_STATE_ERROR, MB_RETRYCANCEL, L"Failed to elevate.");
                if (IDRETRY == nResult)
                {
                    hr = S_FALSE;
                }
            }
        } while (S_FALSE == hr);

        if (SUCCEEDED(hr))
        {
            this->OnCache();
        }
        else
        {
            m_fFailed = TRUE;
            ::PostMessageW(m_hWnd, WM_CLOSE, 0, 0);
        }
    }


    //
    // OnClickRepairButton - start the repair.
    //
    void OnClickRepairButton()
    {
        HRESULT hr = S_OK;

        hr = m_pCore->Execute(REQUEST_STATE_REPAIR);
        ExitOnFailure(hr, "Failed to execute repair.");

        SetState(WIXUX_STATE_EXECUTING);

    LExit:
        return;
    }


    //
    // OnClickUninstallButton - start the uninstall.
    //
    void OnClickUninstallButton()
    {
        this->OnExecute();
    }


    //
    // OnClickCloseButton - close the application.
    //
    void OnClickCloseButton()
    {
        ::PostMessageW(m_hWnd, WM_CLOSE, 0, 0);
    }


    //
    // OnClickAcceptCheckbox - allow the install to continue.
    //
    void OnClickAcceptCheckbox()
    {
        m_fAcceptedLicense = ThemeIsControlChecked(m_pTheme, WIXUX_CONTROL_ACCEPT_CHECKBOX);

        //::EnableWindow(m_pTheme->rgControls[WIXUX_CONTROL_ADVANCED_BUTTON].hWnd, m_fAcceptedLicense);
        ::EnableWindow(m_pTheme->rgControls[WIXUX_CONTROL_INSTALL_BUTTON].hWnd, m_fAcceptedLicense);
        return;
    }


    //
    // SetState
    //
    void SetState(
        __in WIXUX_STATE state
        )
    {
        if (m_state != state)
        {
            Trace2(REPORT_STANDARD, "CWixUserExperience::SetState() changing from state %d to %d", m_state, state);
            m_state = state;

            int nShowTopMessage = (WIXUX_STATE_SELECT_LOCATION == m_state) ? SW_SHOW : SW_HIDE;
            int nShowBottomMessage = (WIXUX_STATE_ACCEPT_LICENSE == m_state || WIXUX_STATE_SELECT_FEATURES == m_state) ? SW_SHOW : SW_HIDE;

            int nShowLicense = (WIXUX_STATE_ACCEPT_LICENSE == m_state) ? SW_SHOW : SW_HIDE;
            int nShowAcceptLicense = (WIXUX_STATE_ACCEPT_LICENSE == m_state) ? SW_SHOW : SW_HIDE;

            int nShowPrint = (WIXUX_STATE_ACCEPT_LICENSE == m_state) ? SW_SHOW : SW_HIDE;
            int nShowAdvanced = (WIXUX_STATE_ACCEPT_LICENSE == m_state) ? SW_SHOW : SW_HIDE;
            int nShowInstall = (WIXUX_STATE_ACCEPT_LICENSE == m_state || WIXUX_STATE_SELECT_FEATURES == m_state) ? SW_SHOW : SW_HIDE;
            int nShowCancel = (WIXUX_STATE_ACCEPT_LICENSE <= m_state && WIXUX_STATE_EXECUTED > m_state) ? SW_SHOW : SW_HIDE;

            int nShowBack = (WIXUX_STATE_SELECT_LOCATION == m_state || WIXUX_STATE_SELECT_FEATURES == m_state) ? SW_SHOW : SW_HIDE;
            int nShowNext = (WIXUX_STATE_SELECT_LOCATION == m_state) ? SW_SHOW : SW_HIDE;

            int nShowRepair = (WIXUX_STATE_MAINTENANCE == m_state) ? SW_SHOW : SW_HIDE;
            int nShowUninstall = (WIXUX_STATE_MAINTENANCE == m_state) ? SW_SHOW : SW_HIDE;
            int nShowClose = (BURN_DISPLAY_FULL == m_command.display && WIXUX_STATE_EXECUTED == m_state) ? SW_SHOW : SW_HIDE;

            int nShowProgressBar = ((WIXUX_STATE_CACHING <= m_state && WIXUX_STATE_EXECUTED > m_state) || WIXUX_STATE_ERROR == m_state) ? SW_SHOW : SW_HIDE;
            int nShowMessage = ((WIXUX_STATE_CACHING <= m_state && WIXUX_STATE_EXECUTED > m_state) || WIXUX_STATE_ERROR == m_state) ? SW_SHOW : SW_HIDE;

            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_TOP_MESSAGE_TEXT].hWnd, nShowTopMessage);
            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_BOTTOM_MESSAGE_TEXT].hWnd, nShowBottomMessage);

            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_LICENSE_RICHEDIT].hWnd, nShowLicense);
            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_ACCEPT_CHECKBOX].hWnd, nShowAcceptLicense);

            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_PRINT_BUTTON].hWnd, nShowPrint);
            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_ADVANCED_BUTTON].hWnd, nShowAdvanced);
            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_INSTALL_BUTTON].hWnd, nShowInstall);
            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_BACK_BUTTON].hWnd, nShowBack);
            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_NEXT_BUTTON].hWnd, nShowNext);

            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_REPAIR_BUTTON].hWnd, nShowRepair);
            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_UNINSTALL_BUTTON].hWnd, nShowUninstall);
            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_CLOSE_BUTTON].hWnd, nShowClose);

            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_PROGRESS_BAR].hWnd, nShowProgressBar);
            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_MESSAGE_TEXT].hWnd, nShowMessage);

            ::ShowWindow(m_pTheme->rgControls[WIXUX_CONTROL_CANCEL_BUTTON].hWnd, nShowCancel);
        }
    }


public:
    //
    // Constructor - intitialize member variables.
    //
    CWixUserExperience(
        __in HMODULE hModule,
        __in BURN_COMMAND* pCommand
        )
    {
        m_hModule = hModule;
        memcpy_s(&m_command, sizeof(m_command), pCommand, sizeof(BURN_COMMAND));

        m_cReferences = 1;
        m_pTheme = NULL;
        m_fRegistered = FALSE;
        m_hWnd = NULL;
        m_hwndHover = NULL;
        m_pCore = NULL;

        m_fCanceled = FALSE;
        m_fFailed = FALSE;
        m_state = WIXUX_STATE_INITIALIZING;

        m_fAcceptedLicense = FALSE;
        m_fAllowRestart = FALSE;
        m_sczLanguage = NULL;
    }


    //
    // Destructor - release member variables.
    //
    ~CWixUserExperience()
    {
        DestroyMainWindow();

        if (m_pTheme)
        {
            ThemeFree(m_pTheme);
            m_pTheme = NULL;
        }

        ThemeUninitialize();

        ReleaseNullObject(m_pCore);
    }

private:
    long m_cReferences;

    HMODULE m_hModule;
    BURN_COMMAND m_command;

    IBurnCore* m_pCore;
    THEME* m_pTheme;
    BOOL m_fRegistered;
    HWND m_hWnd;
    HWND m_hwndHover;

    BOOL m_fCanceled;
    BOOL m_fFailed;
    WIXUX_STATE m_state;

    BOOL m_fAcceptedLicense;
    BOOL m_fAllowRestart;
    LPWSTR m_sczLanguage;
};


//
// CreateUserExperience - creates a new IBurnUserExperience object.
//
HRESULT CreateUserExperience(
    __in HMODULE hModule,
    __in BURN_COMMAND* pCommand,
    __out IBurnUserExperience **ppUX
    )
{
    HRESULT hr = S_OK;
    CWixUserExperience* pUX = NULL;

    pUX = new CWixUserExperience(hModule, pCommand);
    ExitOnNull(pUX, hr, E_OUTOFMEMORY, "Failed to create new WiX UX object.");

    *ppUX = pUX;
    pUX = NULL;

LExit:
    ReleaseObject(pUX);
    return hr;
}
