//-------------------------------------------------------------------------------------------------
// <copyright file="burnexe.cs" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
// 
// <summary>
// The bootstrapper bundle builder application.
// </summary>
//-------------------------------------------------------------------------------------------------

namespace Microsoft.Tools.WindowsInstallerXml.Tools
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Globalization;
    using System.IO;
    using System.Reflection;
    using System.Resources;
    using System.Runtime.InteropServices;
    using System.Text;
    using System.Xml;
    using System.Xml.Schema;
    using System.Xml.XPath;
    using Microsoft.Tools.WindowsInstallerXml.Cab;
    using Microsoft.Deployment.WindowsInstaller;

    /// <summary>
    /// The main entry point for burnexe
    /// </summary>
    public sealed class Burn
    {
        public enum SourceType
        {
            External,
            Embedded,
            Download
        }

        private Hashtable parameters;
        private Dictionary<XmlElement, FileInfo> embeddedPayloads;
        private Dictionary<XmlElement, FileInfo> externalPayloads;
        private List<string> basePaths;
        private List<string> includeSearchPaths;
        private List<string> invalidArgs;
        private string sourceFile;
        private string outputFile;
        private string outputDirectory;
        private string stubFile;
        private CompressionLevel compressionLevel;
        private bool showLogo;
        private bool showHelp;
        private bool suppressSchema;
        private bool onlyResolveManifest;
        private XmlSchema sourceSchema;
        private ConsoleMessageHandler messageHandler;

        private const string BurnNamespace = "http://schemas.microsoft.com/wix/2008/Burn";

        /// <summary>
        /// Prevents a default instance of the Burn class from being created.
        /// </summary>
        private Burn()
        {
            this.invalidArgs = new List<string>();
            this.embeddedPayloads = new Dictionary<XmlElement, FileInfo>();
            this.externalPayloads = new Dictionary<XmlElement, FileInfo>();
            this.parameters = new Hashtable();
            this.includeSearchPaths = new List<string>();
            this.showLogo = true;
            this.compressionLevel = CompressionLevel.Mszip;
            this.messageHandler = new ConsoleMessageHandler("BURN", "burn.exe");
            this.basePaths = new List<string>();
            this.basePaths.Add("."); // always check working directory

            // load the schema so we can validate
            using (Stream resourceStream = Assembly.GetExecutingAssembly().GetManifestResourceStream("Microsoft.Tools.WindowsInstallerXml.Tools.Xsd.source.xsd"))
            {
                this.sourceSchema = XmlSchema.Read(resourceStream, null);
            }
        }

        /// <summary>
        /// The main entry point for burnexe.
        /// </summary>
        /// <param name="args">Commandline arguments for the application.</param>
        /// <returns>Returns the application error code.</returns>
        [MTAThread]
        public static int Main(string[] args)
        {
            AppCommon.PrepareConsoleForLocalization();
            Burn burn = new Burn();
            return burn.Run(args);
        }

        /// <summary>
        /// Returns a namespace manager that includes the Burn namespace.
        /// </summary>
        /// <param name="manifest">Manifest document with packages to resolve.</param>
        /// <returns>Returns the Burn namespace manager.</returns>
        private static XmlNamespaceManager GetBurnNamespaceManager(XmlDocument manifest)
        {
            XmlNamespaceManager namespaceManager = new XmlNamespaceManager(manifest.NameTable);
            namespaceManager.AddNamespace("burn", Burn.BurnNamespace);
            return namespaceManager;
        }

        /// <summary>
        /// Write the (resolved) manifest to a stream.
        /// </summary>
        /// <param name="stream">Stream to write the manifest to.</param>
        /// <param name="manifest">Manifest to write.</param>
        private static void WriteManifest(Stream stream, XmlDocument manifest)
        {
            // write without the source/line-number processing instructions that Preprocessor inserts
            using (ManifestWriter writer = new ManifestWriter(stream))
            {
                manifest.Save(writer);
            }
        }

        /// <summary>
        /// Returns a SourceType enum indicating where the payload comes from.
        /// </summary>
        /// <param name="element">Manifest element with a Packaging attribute (or none, for the default of embedded).</param>
        /// <returns>Returns a SourceType.</returns>
        private static SourceType PayloadSourceType(XmlElement element)
        {
            string packaging = element.GetAttribute("Packaging");
            if (String.IsNullOrEmpty(packaging) || "embedded" == packaging)
            {
                return SourceType.Embedded;
            }
            else if ("external" == packaging)
            {
                return SourceType.External;
            }
            else
            {
                return SourceType.Download;
            }
        }

        /// <summary>
        /// Main running method for the application.
        /// </summary>
        /// <param name="args">Commandline arguments to the application.</param>
        /// <returns>Returns the application error code.</returns>
        private int Run(string[] args)
        {
            try
            {
                // parse the command line
                this.ParseCommandLine(args);

                // exit if there was an error parsing the command line (otherwise the logo appears after error messages)
                if (this.messageHandler.EncounteredError)
                {
                    return this.messageHandler.LastErrorNumber;
                }

                // gotta have a source manifest
                if (String.IsNullOrEmpty(this.sourceFile))
                {
                    this.showHelp = true;
                    this.messageHandler.Display(this, BurnmanErrors.MissingSourceManifest());
                }

                if (this.showLogo)
                {
                    AppCommon.DisplayToolHeader();
                }

                if (this.showHelp)
                {
                    Console.WriteLine(BurnexeStrings.HelpMessage);
                    AppCommon.DisplayToolFooter();
                    return this.messageHandler.LastErrorNumber;
                }

                foreach (string parameter in this.invalidArgs)
                {
                    this.messageHandler.Display(this, WixWarnings.UnsupportedCommandLineArgument(parameter));
                }

                this.invalidArgs = null;

                string sourceFileName = Path.GetFileName(this.sourceFile);
                string targetFile;

                // target is either a resolved manifest file (-man) or the bundle .exe (setup.exe by default)
                if (null != this.outputFile)
                {
                    targetFile = this.outputFile;
                }
                else if (null != this.outputDirectory)
                {
                    targetFile = Path.Combine(this.outputDirectory, "setup.exe");
                }
                else
                {
                    targetFile = "setup.exe";
                }

                // print friendly message saying what file is being compiled
                Console.WriteLine(sourceFileName);

                // do the first half of our job
                XmlDocument resolvedManifest = this.ResolveSourceManifest(this.sourceFile);

                // if we got this far without error, finish the back half of the job
                if (!this.messageHandler.EncounteredError)
                {
                    // if user wants only a resolved manifest, give it
                    if (this.onlyResolveManifest)
                    {
                        using (FileStream stream = File.Create(targetFile))
                        {
                            Burn.WriteManifest(stream, resolvedManifest);
                        }
                    }
                    else
                    {
                        this.BuildBundle(targetFile, resolvedManifest);

                        // now copy any external payloads into the output directory
                        foreach (KeyValuePair<XmlElement, FileInfo> payload in this.externalPayloads)
                        {
                            XmlElement element = payload.Key;
                            FileInfo fileInfo = payload.Value;
                            string targetDirectory = Path.GetDirectoryName(targetFile);
                            string relativeDirectory = Path.Combine(targetDirectory, element.GetAttribute("RelativeDirectory"));
                            string externalTargetFile = Path.Combine(relativeDirectory, fileInfo.Name);
                            this.messageHandler.Display(this, BurnmanVerboses.CopyingExternalPayload(fileInfo.FullName, relativeDirectory));
                            Directory.CreateDirectory(relativeDirectory);
                            if (File.Exists(externalTargetFile))
                            {
                                File.SetAttributes(externalTargetFile, File.GetAttributes(externalTargetFile) & ~System.IO.FileAttributes.ReadOnly);
                            }
                            File.Copy(fileInfo.FullName, externalTargetFile, true);
                        }
                    }
                }
            }
            catch (WixException we)
            {
                this.messageHandler.Display(this, we.Error);
            }
            catch (Exception e)
            {
                this.messageHandler.Display(this, WixErrors.UnexpectedException(e.Message, e.GetType().ToString(), e.StackTrace));
                if (e is NullReferenceException || e is SEHException)
                {
                    throw;
                }
            }

            return this.messageHandler.LastErrorNumber;
        }

        /// <summary>
        /// Builds a bundle, given a resolved manifest that points to optional 
        /// UX DLL, UX resources, and embedded packages.
        /// </summary>
        /// <param name="bundleFile">Burn bundle being built.</param>
        /// <param name="resolvedManifest">Resolved manifest to embed and that contains optional UX and package information.</param>
        private void BuildBundle(string bundleFile, XmlDocument resolvedManifest)
        {
            // default to burnstub.exe if the source manifest didn't say otherwise
            if (String.IsNullOrEmpty(this.stubFile))
            {
                string burnmanDirectory = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
                this.stubFile = Path.Combine(burnmanDirectory, "burnstub.exe");
            }

            // start with a writable copy of the stub .exe
            this.messageHandler.Display(this, BurnmanVerboses.GeneratingBundle(bundleFile, this.stubFile));
            File.Copy(this.stubFile, bundleFile, true);
            File.SetAttributes(bundleFile, File.GetAttributes(bundleFile) & ~System.IO.FileAttributes.ReadOnly);

            // create the container (cabinet)
            int containerId = 1;
            string containerFile = Path.GetTempFileName();
            string manifestFile = Path.GetTempFileName();
            using (WixCreateCab cab = new WixCreateCab(Path.GetFileName(containerFile), Path.GetDirectoryName(containerFile), 0, 0, this.compressionLevel))
            {
                foreach (KeyValuePair<XmlElement, FileInfo> kvp in this.embeddedPayloads)
                {
                    XmlElement element = kvp.Key;
                    FileInfo fileInfo = kvp.Value;
                    string embeddedId = containerId.ToString(CultureInfo.InvariantCulture);
                    this.messageHandler.Display(this, BurnmanVerboses.LoadingPayload(fileInfo.FullName));
                    element.SetAttribute("EmbeddedId", embeddedId);
                    cab.AddFile(fileInfo.FullName, embeddedId);
                    ++containerId;
                }

                // then generate the manifest file as payload 0 in the container (so there's always an embedded container)
                using (Stream manifestStream = File.OpenWrite(manifestFile))
                {
                    this.messageHandler.Display(this, BurnmanVerboses.ResolvingManifest(manifestFile));
                    Burn.WriteManifest(manifestStream, resolvedManifest);
                    cab.AddFile(manifestFile, "0");
                }
                
                cab.Complete();
            }

            // embed the container and update the .wixburn section to point to it
            try
            {
                using (Stream bundleStream = File.Open(bundleFile, FileMode.Open, FileAccess.ReadWrite))
                {
                    FileInfo containerInfo = new FileInfo(containerFile);
                    SectionWriter.WriteContainerHeader(bundleStream, (uint)containerInfo.Length, this.messageHandler, bundleFile);

                    // append the container
                    this.messageHandler.Display(this, BurnmanVerboses.EmbeddingContainer(containerFile, containerInfo.Length, this.compressionLevel.ToString().ToLower(CultureInfo.InvariantCulture)));
                    bundleStream.Seek(0, SeekOrigin.End);
                    byte[] fileData = File.ReadAllBytes(containerFile);
                    bundleStream.Write(fileData, 0, fileData.Length);
                }

            }
            finally
            {
                File.Delete(containerFile);
                File.Delete(manifestFile);
            }

        }

        /// <summary>
        /// Preprocesses the given source manifest file, validates it, and resolves its
        /// constituents, by extracting metadata about each source file referred to.
        /// </summary>
        /// <param name="sourceManifest">Source manifest path.</param>
        /// <returns>Returns the resolved manifest document.</returns>
        private XmlDocument ResolveSourceManifest(string sourceManifest)
        {
            // let the WiX preprocessor handle variables, ?include?, etc.
            Preprocessor preprocessor = new Preprocessor();
            preprocessor.Message += new MessageEventHandler(this.messageHandler.Display);
            for (int i = 0; i < this.includeSearchPaths.Count; ++i)
            {
                preprocessor.IncludeSearchPaths.Add(this.includeSearchPaths[i]);
            }

            XmlDocument resolvedDocument = preprocessor.Process(Path.GetFullPath(sourceManifest), this.parameters);

            if (!this.suppressSchema)
            {
                this.ValidateSourceManifest(resolvedDocument);
            }

            if (!this.messageHandler.EncounteredError)
            {
                this.ResolveStub(resolvedDocument);
                this.ResolveUX(resolvedDocument);
                this.ResolvePackages(resolvedDocument);
            }

            return resolvedDocument;
        }

        /// <summary>
        /// Validates a source manifest document.
        /// </summary>
        /// <param name="sourceDocument">Source manifest document to validate.</param>
        private void ValidateSourceManifest(XmlDocument sourceDocument)
        {
            try
            {
                sourceDocument.Schemas.Add(this.sourceSchema);
                sourceDocument.Validate(null);
            }
            catch (XmlSchemaValidationException e)
            {
                string message = e.Message.Replace(String.Concat(this.sourceSchema.TargetNamespace, ":"), String.Empty);
                this.messageHandler.Display(this, WixErrors.SchemaValidationFailed(null, message, e.LineNumber, e.LinePosition));
            }
        }

        /// <summary>
        /// Resolves a source file by searching the list of base paths.
        /// </summary>
        /// <param name="file">Relative path to source file.</param>
        /// <returns>Returns the path (possibly still relative) of the source file on disk.</returns>
        private string FindFileInBasePaths(string file)
        {
            foreach (string basePath in this.basePaths)
            {
                string path = Path.Combine(basePath, file);
                if (File.Exists(path))
                {
                    return path;
                }
            }

            // failed (epically) to find file
            return null;
        }

        /// <summary>
        /// Resolves attributes common to everything that takes a SourceFile attribute.
        /// </summary>
        /// <param name="element">Manifest element with a SourceFile.</param>
        /// <returns>Returns the resolved path to the file.</returns>
        private FileInfo ResolveSourceFile(XmlElement element)
        {
            // look for matching source files on the basepath
            string sourceFile = element.GetAttribute("SourceFile");
            string path = this.FindFileInBasePaths(sourceFile);
            FileInfo info = null;

            if (null == path)
            {
                this.messageHandler.Display(this, WixErrors.FileNotFound(Preprocessor.GetSourceLineNumbers(element), sourceFile, element.LocalName));
            }
            else
            {
                info = new FileInfo(path);
                element.RemoveAttribute("SourceFile");

                // stubs are never payloads
                if ("Stub" != element.LocalName)
                {
                    // add payload-related attributes
                    element.SetAttribute("FileName", info.Name);
                    element.SetAttribute("FileSize", info.Length.ToString(CultureInfo.InvariantCulture));

                    // track payloads for later processing
                    SourceType sourceType = Burn.PayloadSourceType(element);
                    if (SourceType.Embedded == sourceType)
                    {
                        this.embeddedPayloads.Add(element, info);
                    }
                    else if (SourceType.External == sourceType)
                    {
                        this.externalPayloads.Add(element, info);
                    }
                }
            }

            return info;
        }

        /// <summary>
        /// Resolves the Stub element.
        /// </summary>
        /// <param name="sourceManifest">Source manifest document to be resolved.</param>
        private void ResolveStub(XmlDocument sourceManifest)
        {
            // for the one Stub, resolve its SourceFile attribute
            foreach (XmlElement element in sourceManifest.DocumentElement.SelectNodes("burn:Stub", Burn.GetBurnNamespaceManager(sourceManifest)))
            {
                FileInfo info = this.ResolveSourceFile(element);
                if (null != info)
                {
                    this.stubFile = info.FullName;
                }
            }
        }

        /// <summary>
        /// Resolves the SourceFile attribute of the UX DLL and its Resources.
        /// </summary>
        /// <param name="sourceManifest">Source manifest document with UX to resolve.</param>
        private void ResolveUX(XmlDocument sourceManifest)
        {
            // for each UX and UX Resource, resolve its SourceFile attribute
            XmlNamespaceManager namespaceManager = Burn.GetBurnNamespaceManager(sourceManifest);
            XmlNodeList uxNodes = sourceManifest.DocumentElement.SelectNodes("burn:UX", namespaceManager);

            if (0 < uxNodes.Count)
            {
                foreach (XmlElement element in uxNodes)
                {
                    this.ResolveSourceFile(element);
                }

                foreach (XmlElement element in sourceManifest.DocumentElement.SelectNodes("burn:UX/burn:Resource", namespaceManager))
                {
                    this.ResolveSourceFile(element);
                }
            }
            else
            {
                // default UX is embedded stdux
                XmlElement uxElement = sourceManifest.CreateElement("UX");
                uxElement.SetAttribute("SourceFile", "stdux.dll");
                sourceManifest.DocumentElement.AppendChild(uxElement);
                this.ResolveSourceFile(uxElement);

                foreach (string uxFile in new string[] { "thm.xml", "en-us.wxl", "stdux.png", "tou.htm" })
                {
                    XmlElement element = sourceManifest.CreateElement("Resource");
                    element.SetAttribute("SourceFile", uxFile);
                    uxElement.AppendChild(element);
                    this.ResolveSourceFile(element);
                }
            }
        }

        /// <summary>
        /// Resolves the SourceFile attribute of each package in the chain.
        /// </summary>
        /// <param name="sourceManifest">Source manifest document with packages to resolve.</param>
        private void ResolvePackages(XmlDocument sourceManifest)
        {
            // for each package element, resolve its SourceFile attribute
            XmlNamespaceManager namespaceManager = Burn.GetBurnNamespaceManager(sourceManifest);
            XmlNodeList nodes = sourceManifest.DocumentElement.SelectNodes("burn:Chain/*", namespaceManager);
            foreach (XmlElement element in nodes)
            {
                FileInfo info = this.ResolveSourceFile(element);
                switch (element.LocalName)
                {
                    case "MsiPackage":
                        this.ResolveMsiPackage(element, info);
                        break;
                    case "ExePackage":
                        this.ResolveExePackage(element, info);
                        break;
                }
            }
        }

        /// <summary>
        /// Resolves an ExePackage chain element.
        /// </summary>
        /// <param name="element">ExePackage element.</param>
        /// <param name="info">FileInfo for resolved source file for the package.</param>
        private void ResolveExePackage(XmlElement element, FileInfo info)
        {
            // nothing to do -- everything's handled in ResolveSourceFile
        }

        /// <summary>
        /// Resolves an MsiPackage chain element.
        /// </summary>
        /// <param name="element">MsiPackage element.</param>
        /// <param name="info">FileInfo for resolved source file for the package.</param>
        private void ResolveMsiPackage(XmlElement element, FileInfo info)
        {
            // peek into the .msi database
            string productCode = null;
            string productVersion = null;
            bool isPerMachine = false;

            if (null != info)
            {
                using (SummaryInfo sumInfo = new SummaryInfo(info.FullName, false))
                {
                    // 8 is the Word Count summary information stream bit that means
                    // "Elevated privileges are not required to install this package."
                    // in MSI 4.5 and below, if this bit is 0, elevation is required.
                    isPerMachine = 0 == (sumInfo.WordCount & 8);
                }

                using (Database db = new Database(info.FullName))
                {
                    try
                    {
                        productCode = db.ExecuteScalar("SELECT `Value` FROM `Property` WHERE `Property` = 'ProductCode'").ToString();
                        productVersion = db.ExecuteScalar("SELECT `Value` FROM `Property` WHERE `Property` = 'ProductVersion'").ToString();

                        // add all external cabinets as package resources
                        foreach (string cabinet in db.ExecuteStringQuery("SELECT `Cabinet` FROM `Media`"))
                        {
                            if (!String.IsNullOrEmpty(cabinet) && !cabinet.StartsWith("#"))
                            {
                                XmlElement resourceElement = element.OwnerDocument.CreateElement("Resource", Burn.BurnNamespace);
                                element.AppendChild(resourceElement);
                                
                                resourceElement.SetAttribute("SourceFile", Path.Combine(info.DirectoryName, cabinet));
                                this.ResolveSourceFile(resourceElement);
                            }
                        }
                    }
                    catch (InstallerException e)
                    {
                        this.messageHandler.Display(this, WixErrors.UnexpectedException(e.Message, "InstallerException", e.StackTrace));
                    }
                }

                element.SetAttribute("PerMachine", isPerMachine ? "yes" : "no");
                element.SetAttribute("ProductCode", productCode);
                element.SetAttribute("ProductVersion", productVersion);
            }
        }

        /// <summary>
        /// Parse the commandline arguments.
        /// </summary>
        /// <param name="args">Commandline arguments.</param>
        private void ParseCommandLine(string[] args)
        {
            for (int i = 0; i < args.Length; ++i)
            {
                string arg = args[i];
                if (null == arg || 0 == arg.Length) 
                {
                    // skip blank arguments
                    continue;
                }

                if (1 == arg.Length) 
                {
                    // treat '-' and '@' as filenames when by themselves.
                    this.sourceFile = String.Join(" ", AppCommon.GetFiles(arg, "SourceManifest"));
                    continue;
                }

                if ('-' == arg[0] || '/' == arg[0])
                {
                    string parameter = arg.Substring(1);
                    if ("b" == parameter)
                    {
                        string path = CommandLine.GetDirectory(parameter, this.messageHandler, args, ++i);

                        if (String.IsNullOrEmpty(path))
                        {
                            return;
                        }

                        this.basePaths.Add(path);
                    }
                    else if (parameter.StartsWith("dcl:"))
                    {
                        string newCompressionLevel = arg.Substring(5).ToLower(CultureInfo.InvariantCulture);
                        switch (newCompressionLevel)
                        {
                            case "low":
                                this.compressionLevel = CompressionLevel.Low;
                                break;
                            case "medium":
                                this.compressionLevel = CompressionLevel.Medium;
                                break;
                            case "high":
                                this.compressionLevel = CompressionLevel.High;
                                break;
                            case "none":
                                this.compressionLevel = CompressionLevel.None;
                                break;
                            case "mszip":
                                this.compressionLevel = CompressionLevel.Mszip;
                                break;
                            default:
                                throw new WixException(WixErrors.IllegalCompressionLevel(newCompressionLevel));
                        }
                    }
                    else if ('d' == parameter[0])
                    {
                        if (1 >= parameter.Length || '=' == parameter[1])
                        {
                            this.messageHandler.Display(this, WixErrors.InvalidVariableDefinition(arg));
                            return;
                        }

                        parameter = arg.Substring(2);

                        string[] value = parameter.Split("=".ToCharArray(), 2);

                        if (this.parameters.ContainsKey(value[0]))
                        {
                            this.messageHandler.Display(this, WixErrors.DuplicateVariableDefinition(value[0], (1 == value.Length) ? String.Empty : value[1], (string)this.parameters[value[0]]));
                            return;
                        }

                        if (1 == value.Length)
                        {
                            this.parameters.Add(value[0], String.Empty);
                        }
                        else
                        {
                            this.parameters.Add(value[0], value[1]);
                        }
                    }
                    else if ('I' == parameter[0])
                    {
                        this.includeSearchPaths.Add(parameter.Substring(1));
                    }
                    else if ("man" == parameter)
                    {
                        this.onlyResolveManifest = true;
                    }
                    else if ("nologo" == parameter)
                    {
                        this.showLogo = false;
                    }
                    else if ("o" == parameter || "out" == parameter)
                    {
                        string path = CommandLine.GetFileOrDirectory(parameter, this.messageHandler, args, ++i);

                        if (!String.IsNullOrEmpty(path))
                        {
                            if (path.EndsWith("\\") || path.EndsWith("/"))
                            {
                                this.outputDirectory = path;
                            }
                            else
                            {
                                this.outputFile = path;
                            }
                        }
                        else
                        {
                            return;
                        }
                    }
                    else if ("ss" == parameter)
                    {
                        this.suppressSchema = true;
                    }
                    else if ("v" == parameter)
                    {
                        this.messageHandler.ShowVerboseMessages = true;
                    }
                    else if ("?" == parameter || "help" == parameter)
                    {
                        this.showHelp = true;
                        return;
                    }
                    else
                    {
                        this.invalidArgs.Add(parameter);
                    }
                }
                else if ('@' == arg[0])
                {
                    this.ParseCommandLine(CommandLineResponseFile.Parse(arg.Substring(1)));
                }
                else
                {
                    this.sourceFile = String.Join(" ", AppCommon.GetFiles(arg, "SourceManifest"));
                }
            }

            return;
        }

        /// <summary>
        /// An XmlWriter that doesn't write out the source/line-number processing instructions
        /// from the WiX preprocessor.
        /// </summary>
        private class ManifestWriter : XmlTextWriter
        {
            /// <summary>
            /// Initializes a new instance of the ManifestWriter class.
            /// </summary>
            /// <param name="stream">The stream to write to.</param>
            public ManifestWriter(Stream stream)
                : base(stream, System.Text.Encoding.UTF8)
            {
            }

            /// <summary>
            /// Overrides WriteProcessingInstruction to <b>not</b> write processing instruction.
            /// </summary>
            /// <param name="name">The processing instruction name.</param>
            /// <param name="text">The processing instruction text.</param>
            public override void WriteProcessingInstruction(string name, string text)
            {
                if ("ln" != name)
                {
                    base.WriteProcessingInstruction(name, text);
                }
            }
        }

        unsafe private static class SectionWriter
        {
            private const uint IMAGE_DOS_SIGNATURE = 0x5A4D;
            private const uint IMAGE_NT_SIGNATURE = 0x00004550;
            private const uint SIZEOF_BURN_SECTION = 16/*sizeof(GUID)*/ + 4 * sizeof(uint);
            private const uint dwMagic = 0x00f14300;
            private const uint dwVersion = 0x00000001;

            public static bool WriteContainerHeader(Stream stubStream, uint cbContainer, ConsoleMessageHandler messageHandler, string stubFilename)
            {
                byte[] bytes = new byte[stubStream.Length];
                stubStream.Seek(0, SeekOrigin.Begin);
                stubStream.Read(bytes, 0, bytes.Length);

                fixed (byte* data = bytes)
                {
                    // make sure we have an "MZ" .exe
                    IMAGE_DOS_HEADER* pDosHeader = (IMAGE_DOS_HEADER*)data;
                    if (IMAGE_DOS_SIGNATURE != pDosHeader->e_magic)
                    {
                        messageHandler.Display(null, BurnmanErrors.InvalidStubExe(stubFilename));
                        return false;
                    }

                    // make sure we have a PE file
                    IMAGE_NT_HEADERS32* pNtHeader = (IMAGE_NT_HEADERS32*)(data + pDosHeader->e_lfanew);
                    if (IMAGE_NT_SIGNATURE != pNtHeader->Signature)
                    {
                        messageHandler.Display(null, BurnmanErrors.InvalidStubExe(stubFilename));
                        return false;
                    }

                    // find the .wixburn section
                    IMAGE_SECTION_HEADER* pImageSectionHeader = (IMAGE_SECTION_HEADER*)((byte*)pNtHeader + sizeof(uint) + sizeof(IMAGE_FILE_HEADER) + pNtHeader->FileHeader.SizeOfOptionalHeader);
                    ulong ulWixBurn = 0x6E7275627869772E; // ".wixburn", as a qword.
                    ushort cSections = pNtHeader->FileHeader.NumberOfSections;
                    ushort iSection = 0;
                    for (; iSection < cSections; ++iSection)
                    {
                        if (pImageSectionHeader->ulName == ulWixBurn)
                        {
                            break;
                        }
                        ++pImageSectionHeader;
                    }

                    // didn't find a .wixburn section, so it's not a valid Burn stub.
                    if (cSections <= iSection)
                    {
                        messageHandler.Display(null, BurnmanErrors.StubMissingWixburnSection(stubFilename));
                        return false;
                    }

                    // we need 32 bytes for the manifest header, which is always going to fit in 
                    // the smallest alignment (512 bytes), but just to be paranoid...
                    if (pImageSectionHeader->SizeOfRawData < SIZEOF_BURN_SECTION)
                    {
                        messageHandler.Display(null, BurnmanErrors.StubMissingWixburnSection(stubFilename));
                        return false;
                    }

                    // update the bytes in memory
                    // 0-3:  magic number
                    // 4-7:  version
                    // 8-11:  offset to start of container
                    // 12-15: byte count of container
                    // 16-31: bundle GUID
                    pImageSectionHeader->VirtualSize = SIZEOF_BURN_SECTION;

                    uint* pContainerSection = (uint*)((byte*)data + pImageSectionHeader->PointerToRawData);
                    *pContainerSection = dwMagic;
                    *++pContainerSection = dwVersion;
                    *++pContainerSection = (uint)stubStream.Length;
                    *++pContainerSection = cbContainer;

                    byte* pGuid = (byte*)++pContainerSection;
                    Guid bundleGuid = Guid.NewGuid();
                    messageHandler.Display(null, BurnmanVerboses.BundleGuid(bundleGuid.ToString("B")));
                    foreach (byte guidByte in bundleGuid.ToByteArray())
                    {
                        *pGuid++ = guidByte;
                    }

                    // and then write the bytes back to disk
                    stubStream.Seek(0, SeekOrigin.Begin);
                    stubStream.Write(bytes, 0, bytes.Length);

                    return true;
                }
            }

            [StructLayout(LayoutKind.Sequential)]
            private struct IMAGE_SECTION_HEADER
            {
                // fixed byte Name[8];
                public ulong ulName;
                public uint VirtualSize;
                public uint VirtualAddress;
                public uint SizeOfRawData;
                public uint PointerToRawData;
                public uint PointerToRelocations;
                public uint PointerToLinenumbers;
                public short NumberOfRelocations;
                public short NumberOfLinenumbers;
                public uint Characteristics;
            }

            [StructLayout(LayoutKind.Explicit)]
            private struct IMAGE_DOS_HEADER
            {
                [FieldOffset(0)]
                public ushort e_magic;
                [FieldOffset(60)]
                public short e_lfanew;
            }

            [StructLayout(LayoutKind.Explicit)]
            private struct IMAGE_NT_HEADERS32
            {
                [FieldOffset(0)]
                public uint Signature;
                [FieldOffset(4)]
                public IMAGE_FILE_HEADER FileHeader;
            }

            [StructLayout(LayoutKind.Sequential)]
            private struct IMAGE_FILE_HEADER
            {
                public ushort Machine;
                public ushort NumberOfSections;
                public uint TimeDateStamp;
                public uint PointerToSymbolTable;
                public uint NumberOfSymbols;
                public ushort SizeOfOptionalHeader;
                public ushort Characteristics;
            }
        }
    }
}
