//-------------------------------------------------------------------------------------------------
// <copyright file="BurnCacheServices.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

//
// EnsureWorkingDirectory - ensures the working directory for a payload is valid.
//
HRESULT CBurnCacheServices::EnsureWorkingDirectory(
    __in_z LPCWSTR wzWorkingPath,
    __out_z_opt LPWSTR* psczWorkingDir
    )
{
    HRESULT hr = S_OK;
    LPWSTR sczWorkingDir = NULL;

    hr = StrAllocString(&sczWorkingDir, wzWorkingPath, FileFromPath(wzWorkingPath) - wzWorkingPath);
    ExitOnFailure(hr, "Failed to get working directory for payload.");

    hr = DirEnsureExists(sczWorkingDir, NULL);
    ExitOnFailure(hr, "Failed create working directory for payload.");

    if (psczWorkingDir)
    {
        hr = StrAllocString(psczWorkingDir, sczWorkingDir, 0);
        ExitOnFailure(hr, "Failed to copy working directory.");
    }

LExit:
    ReleaseStr(sczWorkingDir);
    return hr;
}


//
// CompletePayload - move a payload from its working path to its completed path.
//
HRESULT CBurnCacheServices::CompletePayload(
    __in IBurnPayload* pPayload
    )
{
    HRESULT hr = S_OK;
    BOOL fIsCached = FALSE;
    LPWSTR sczWorkingPath = NULL;
    LPWSTR sczCompletedPath = NULL;
    LPWSTR sczCompletedDir = NULL;

    hr = pPayload->IsCached(&fIsCached);
    ExitOnFailure(hr, "Failed to find whether payload is already cached.");

    if (!fIsCached)
    {
        hr = pPayload->GetWorkingPath(&sczWorkingPath);
        ExitOnFailure(hr, "Failed to get working path for payload.");

        hr = pPayload->GetCompletedPath(&sczCompletedPath);
        ExitOnFailure(hr, "Failed to get completed path for payload.");

        hr = StrAllocString(&sczCompletedDir, sczCompletedPath, FileFromPath(sczCompletedPath) - sczCompletedPath);
        ExitOnFailure(hr, "Failed to get completed directory for payload.");

        hr = DirEnsureExists(sczCompletedDir, NULL);
        ExitOnFailure(hr, "Failed create completed directory for payload.");

        LogStringLine(REPORT_STANDARD, "Moving payload from working path '%ls' to completed path '%ls'", sczWorkingPath, sczCompletedPath);
        hr = FileEnsureMove(sczWorkingPath, sczCompletedPath, TRUE, TRUE);
        ExitOnFailure2(hr, "Failed to move %ls to %ls", sczWorkingPath, sczCompletedPath);
    }

LExit:
    ReleaseStr(sczWorkingPath);
    ReleaseStr(sczCompletedPath);
    ReleaseStr(sczCompletedDir);

    return hr;
}
