//-------------------------------------------------------------------------------------------------
// <copyright file="BurnDownloadEngine.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

static const DWORD64 BURN_DOWNLOAD_ENGINE_TWO_GIGABYTES = DWORD64(2) * 1024 * 1024 * 1024;
static LPCWSTR BURN_DOWNLOAD_ENGINE_HEADER_FORM_DATA = L"Content-Type: application/x-www-form-urlencoded";
static LPCWSTR BURN_DOWNLOAD_ENGINE_HEADER_FINAL_BYTE = L"Range: bytes=-1";
static LPCWSTR BURN_DOWNLOAD_ENGINE_ACCEPT_TYPES[] = { L"*/*", NULL };

struct BURN_DOWNLOAD_ENGINE_INFO
{
    HINTERNET hConnect;

    LPWSTR sczUrl;
    INTERNET_SCHEME scheme;
    LPWSTR sczResource;
    DWORD64 dw64Length;
    FILETIME ftServerLastModified;

    BOOL fRangeRequestsSupported;
    DWORD64 dw64ResumeOffset;
};

class CBurnDownloadEngine : public CUnknownImpl<IBurnCacheEngine>
{
public: // IBurnCacheEngine
    //
    // CachePayload - cache a payload (UX or package payloads).
    //
    STDMETHODIMP CachePayload(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in IBurnView* pView
        )
    {
        HRESULT hr = S_OK;
        LPWSTR sczSourceUrl = NULL;
        LPWSTR sczWorkingPath = NULL;
        BURN_DOWNLOAD_ENGINE_INFO* pDownloadInfo = NULL;
        HANDLE hFile = INVALID_HANDLE_VALUE;

        hr = GetSourceAndWorkingPath(pPayload, &sczSourceUrl, &sczWorkingPath);
        ExitOnFailure(hr, "Failed to get source and working path.");

        hr = Connect(pPackage, pPayload, pView, sczSourceUrl, &pDownloadInfo);
        ExitOnFailure1(hr, "Failed to connect to URL: %ls", sczSourceUrl);

        hFile = ::CreateFileW(sczWorkingPath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_DELETE, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
        if (INVALID_HANDLE_VALUE == hFile)
        {
            ExitWithLastError1(hr, "Failed to working file to download into: %ls", sczWorkingPath);
        }

        if (pDownloadInfo->fRangeRequestsSupported)
        {
            // TODO: read the resume offset
        }

        hr = ExecuteDownload(pPackage, pPayload, pView, pDownloadInfo, hFile);
        ExitOnFailure1(hr, "Failed to execute download URL: %ls", sczSourceUrl);

    LExit:
        ReleaseFileHandle(hFile);

        if (pDownloadInfo)
        {
            DestroyDownloadInfo(pDownloadInfo);
        }

        ReleaseStr(sczWorkingPath);
        ReleaseStr(sczSourceUrl);

        return hr;
    }

    //
    // Complete - move a payload from its working path to its completed path
    //
    STDMETHODIMP Complete(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in IBurnView* pView
        )
    {
        HRESULT hr = S_OK;

        DBG_UNREFERENCED_PARAMETER(pPackage);
        DBG_UNREFERENCED_PARAMETER(pView);

        hr = CBurnCacheServices::CompletePayload(pPayload);
        ExitOnFailure(hr, "Failed to complete payload.");

    LExit:
        return hr;
    }


public:
    //
    // Constructor - intitialize member variables.
    //
    CBurnDownloadEngine(
        __out HRESULT* pHR
        )
    {
        HRESULT hr = S_OK;

        m_hSession = ::InternetOpenW(L"Burn", INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, 0);
        ExitOnNullWithLastError(m_hSession, hr, "Failed to open internet session");

    LExit:
        *pHR = hr;
    }

    //
    // Destructor - release member variables.
    //
    ~CBurnDownloadEngine()
    {
        ReleaseInternet(m_hSession);
    }

private:
    //
    // Download - manage the steps to download the source URL to the working directory.
    //
    HRESULT GetSourceAndWorkingPath(
        __in IBurnPayload* pPayload,
        __out_z LPWSTR* psczSourceUrl,
        __out_z LPWSTR* psczWorkingPath
        )
    {
        HRESULT hr = S_OK;
        SOURCE_TYPE sourceType = SOURCE_TYPE_NONE;

        hr = pPayload->GetSource(&sourceType, psczSourceUrl);
        ExitOnFailure(hr, "Failed to get source path for payload.");

        AssertSz(SOURCE_TYPE_DOWNLOAD == sourceType, "Invalid source type when using download engine.");

        hr = pPayload->GetWorkingPath(psczWorkingPath);
        ExitOnFailure(hr, "Failed to get working path for payload.");

        hr =  CBurnCacheServices::EnsureWorkingDirectory(*psczWorkingPath, NULL);
        ExitOnFailure(hr, "Failed create working directory for payload.");

    LExit:
        return hr;
    }

    //
    // Connect- connect to the remote resource.
    //
    HRESULT Connect(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in IBurnView* pView,
        __in LPCWSTR wzSourceUrl,
        __out BURN_DOWNLOAD_ENGINE_INFO** ppDownloadInfo
        )
    {
        HRESULT hr = S_OK;
        BURN_DOWNLOAD_ENGINE_INFO* pDownloadInfo = NULL;
        LPWSTR sczUrl = NULL;
        BOOL fRangesAccepted = FALSE;
        FILETIME ftServerLastModified = { 0 };

        HINTERNET hConnect = NULL;
        HINTERNET hUrl = NULL;

        INTERNET_SCHEME scheme = INTERNET_SCHEME_UNKNOWN;
        LPWSTR sczHostName = NULL;
        INTERNET_PORT port = INTERNET_INVALID_PORT_NUMBER;
        LPWSTR sczUser = NULL;
        LPWSTR sczPassword = NULL;
        LPWSTR sczResource = NULL;
        LPWSTR sczQueryString = NULL;

        BOOL fAcceptsHeadRequest = TRUE; // assume the server will accept HTTP HEAD requests.
        LPCWSTR wzMethod = NULL;
        LPCWSTR wzHeader = NULL;
        LPWSTR sczContentRange = NULL;

        // Start with the provided URL, we may get redirected to a different URL.
        hr = StrAllocString(&sczUrl, wzSourceUrl, 0);
        ExitOnFailure1(hr, "Failed to copy URL: %S", wzSourceUrl);

        // Try to open the URL.
        BOOL fRetry;
        do
        {
            fRetry = FALSE;

            hr = pView->WaitIfPaused();
            ExitOnFailure(hr, "User canceled while paused.");

            // If the URL was opened close it, so we can reopen it again.
            ReleaseInternet(hUrl);
            ReleaseInternet(hConnect);

            // Open the url.
            ReleaseNullStr(sczHostName);
            ReleaseNullStr(sczUser);
            ReleaseNullStr(sczPassword);
            ReleaseNullStr(sczQueryString);
            ReleaseNullStr(sczResource);

            hr = UriCrack(sczUrl, &scheme, &sczHostName, &port, &sczUser, &sczPassword, &sczResource, &sczQueryString);
            ExitOnFailure(hr, "Failed to break URL into server and resource parts.");

            if (sczQueryString && *sczQueryString)
            {
                hr = StrAllocConcat(&sczResource, sczQueryString, 0);
                ExitOnFailure(hr, "Failed to append query strong to resource from URI.");
            }

            hConnect = ::InternetConnectW(m_hSession, sczHostName, port, sczUser, sczPassword, INTERNET_SCHEME_FTP == scheme ? INTERNET_SERVICE_FTP : INTERNET_SERVICE_HTTP, 0, 0);
            ExitOnNullWithLastError1(hConnect, hr, "Failed to connect to URL: %S", sczUrl);

            // Make a range request for the last byte to see if the server will accept range requests and if so tell us
            // the total length of the content without actually downloading it.
            wzMethod = fAcceptsHeadRequest ? L"HEAD" : L"GET";
            wzHeader = BURN_DOWNLOAD_ENGINE_HEADER_FINAL_BYTE;

            hr = OpenRequest(hConnect, wzMethod, scheme, sczResource, wzHeader, &hUrl);
            ExitOnFailure1(hr, "Failed to open internet URL: %S", sczUrl);

            if (!::HttpSendRequestW(hUrl, NULL, 0, NULL, 0))
            {
                ExitWithLastError1(hr, "Failed to send request to URL: %S", sczUrl);
            }

            // Check the http status code.
            LONG lCode = 0;
            hr = InternetQueryInfoNumber(hUrl, HTTP_QUERY_STATUS_CODE, &lCode);
            ExitOnFailure1(hr, "Failed to get HTTP status code for URL: %S", sczUrl);

            LogStringLine(REPORT_STANDARD, "Download engine HTTP %d %ls to %ls", lCode, wzMethod, sczUrl);

            switch (lCode)
            {
            case 200: // OK but range requests don't work.
                fRangesAccepted = FALSE;
                hr = S_OK;
                break;

            case 206: // Partial content means that range requests work!
                fRangesAccepted = TRUE;
                hr = S_OK;
                break;

            // redirection cases
            case 301:  __fallthrough; // file moved
            case 302:  __fallthrough; // temporary
            case 303:  // redirect method
                hr = InternetQueryInfoString(hUrl, HTTP_QUERY_CONTENT_LOCATION, &sczUrl);
                ExitOnFailure1(hr, "Failed to get redirect location for url: %S", sczUrl);

                fRetry = TRUE;
                break;

            // error cases
            case 400:
                hr = HRESULT_FROM_WIN32(ERROR_BAD_ARGUMENTS);
                break;

            case 401:  __fallthrough; // unauthorized
            case 403:  // access denied
                hr = HRESULT_FROM_WIN32(ERROR_ACCESS_DENIED);
                break;

            case 405:
                // If we just tried the HEAD request but were told that method isn't accepted,
                // fallback to a GET request.
                if (fAcceptsHeadRequest)
                {
                    fAcceptsHeadRequest = FALSE;
                    fRetry = TRUE;
                    hr = S_OK;
                }
                else
                {
                    hr = HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED);
                }
                break;

            case 404:  __fallthrough; // file not found
            case 502:  __fallthrough; // server (through a gateyway) was not found
            case 503:  // server unavailable
                hr = HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
                break;

            case 408:  __fallthrough; // request timedout
            case 504:   // gateway timeout
                hr = HRESULT_FROM_WIN32(WAIT_TIMEOUT);
                break;

            default:
                {
                    hr = E_UNEXPECTED;
#ifdef DEBUG
                    CHAR sz[INTERNET_MAX_URL_LENGTH];
                    StringCchPrintfA(sz, countof(sz), "unhandled HTTP status %d, unknown status code for URL: %S", lCode, sczUrl);
                    AssertSz(FALSE, sz);
#endif
                }
                break;
            }
        } while (S_OK == hr && fRetry);
        ExitOnRootFailure1(hr, "Failed to connect to URL: %S", sczUrl);

        // Get the size of the download.
        DWORD64 dw64Length = 0;
        if (fRangesAccepted)
        {
            hr = InternetQueryInfoString(hUrl, HTTP_QUERY_CONTENT_RANGE, &sczContentRange);
            ExitOnFailure(hr, "Failed to get content range from download.");

            LPWSTR wz = wcschr(sczContentRange, L'/');
            ExitOnNull(wz, hr, HRESULT_FROM_WIN32(ERROR_INVALID_DATA), "Failed to locate size in content range.");

            dw64Length = _wcstoui64(wz + 1, NULL, 10);
        }
        else
        {
            LONGLONG llLength = 0;
            hr = InternetGetSizeByHandle(hUrl, &llLength);
            ExitOnFailure(hr, "Failed to get size of download.");

            dw64Length = static_cast<DWORD64>(llLength);
        }

        // Get the last modified time from the server, we'll use that as our downloaded time here. If
        // the server time isn't provided then use the local system time.
        hr = InternetGetCreateTimeByHandle(hUrl, &ftServerLastModified);
        if (HRESULT_FROM_WIN32(ERROR_HTTP_HEADER_NOT_FOUND) == hr)
        {
            ::GetSystemTimeAsFileTime(&ftServerLastModified);
            hr = S_OK;
        }
        ExitOnFailure(hr, "Failed to get time for resource from server.");

        // Okay, we're all ready to start downloading. Update the download information.
        pDownloadInfo = static_cast<BURN_DOWNLOAD_ENGINE_INFO*>(MemAlloc(sizeof(BURN_DOWNLOAD_ENGINE_INFO), TRUE));
        ExitOnNull(pDownloadInfo, hr, E_OUTOFMEMORY, "Failed to allocate download info.");

        pDownloadInfo->hConnect = hConnect;
        pDownloadInfo->sczUrl = sczUrl;
        pDownloadInfo->scheme = scheme;
        pDownloadInfo->sczResource = sczResource;
        pDownloadInfo->dw64Length = dw64Length;
        pDownloadInfo->ftServerLastModified = ftServerLastModified;
        pDownloadInfo->fRangeRequestsSupported = fRangesAccepted;
        pDownloadInfo->dw64ResumeOffset = 0;

        hConnect = NULL;
        sczUrl = NULL;
        sczResource = NULL;

        *ppDownloadInfo = pDownloadInfo;
        pDownloadInfo = NULL;

    LExit:
        if (pDownloadInfo)
        {
            DestroyDownloadInfo(pDownloadInfo);
        }

        ReleaseStr(sczContentRange);
        ReleaseStr(sczHostName);
        ReleaseStr(sczUser);
        ReleaseStr(sczPassword);
        ReleaseStr(sczQueryString);
        ReleaseStr(sczResource);

        ReleaseInternet(hUrl);
        ReleaseInternet(hConnect);
        ReleaseStr(sczUrl);

        return hr;
    }

    //
    // ExecuteDownload - actually download bits.
    //
    HRESULT ExecuteDownload(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in IBurnView* pView,
        __in BURN_DOWNLOAD_ENGINE_INFO* pDownloadInfo,
        __in HANDLE hFile
        )
    {
        HRESULT hr = S_OK;
        HANDLE hUrl = NULL;

        DWORD cbMaxData = 64 * 1024; // 64 KB
        BYTE* pbData = NULL;

        // Allocate a memory block on a page boundary in case we want to do optimal writing.
        pbData = static_cast<BYTE*>(::VirtualAlloc(NULL, cbMaxData, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE));
        ExitOnNullWithLastError(pbData, hr, "Failed to allocate buffer to download files into.");

        while (pDownloadInfo->dw64ResumeOffset < pDownloadInfo->dw64Length)
        {
            hr = DownloadRequest(pDownloadInfo, &hUrl);
            ExitOnFailure(hr, "Failed to make a range request.");

            hr = DownloadFile(pPackage, pPayload, pView, pDownloadInfo, hUrl, hFile, pbData, cbMaxData);
            ExitOnFailure(hr, "Failed while reading from internet.");
        }

    LExit:
        if (pbData)
        {
            ::VirtualFree(pbData, 0, MEM_RELEASE);
        }

        ReleaseInternet(hUrl);

        return hr;
    }

    //
    // DownloadRequest - make a request for the specified URL.
    //
    HRESULT DownloadRequest(
        __in BURN_DOWNLOAD_ENGINE_INFO* pDownloadInfo,
        __out HINTERNET* phUrl
        )
    {
        HRESULT hr = S_OK;
        LPWSTR sczHeaders = NULL;
        HINTERNET hUrl = NULL;

        // If we can try to resume, let's see if there is any reason to build up a range
        // request header.
        if (pDownloadInfo->fRangeRequestsSupported)
        {
            // If the remaining length is less that 2GB we'll be able to ask for everything.
            DWORD64 dw64RemainingLength = pDownloadInfo->dw64Length - pDownloadInfo->dw64ResumeOffset;
            if (BURN_DOWNLOAD_ENGINE_TWO_GIGABYTES > dw64RemainingLength)
            {
                // If we have a resume offset, let's download everything from there. Otherwise, we'll
                // just get everything with no headers in the way.
                if (0 < pDownloadInfo->dw64ResumeOffset)
                {
                    hr = StrAllocFormatted(&sczHeaders, L"Range: bytes=%I64u-", pDownloadInfo->dw64ResumeOffset);
                    ExitOnFailure(hr, "Failed to add range read header.");
                }
            }
            else // we'll have to download in chunks.
            {
                hr = StrAllocFormatted(&sczHeaders, L"Range: bytes=%I64u-%I64u", pDownloadInfo->dw64ResumeOffset, pDownloadInfo->dw64ResumeOffset + dw64RemainingLength - 1);
                ExitOnFailure(hr, "Failed to add range read header.");
            }
        }

        hr = OpenRequest(pDownloadInfo->hConnect, NULL, pDownloadInfo->scheme, pDownloadInfo->sczResource, sczHeaders, &hUrl);
        ExitOnFailure(hr, "Failed to open internet URL.");

        if (!::HttpSendRequestW(hUrl, NULL, 0, NULL, 0))
        {
            ExitWithLastError(hr, "Failed to send request.");
        }

        *phUrl = hUrl;
        hUrl = NULL;

    LExit:
        ReleaseInternet(hUrl);
        ReleaseStr(sczHeaders);

        return hr;
    }

    //
    // DownloadFile - read from the internet, write to the cache.
    //
    HRESULT DownloadFile(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in IBurnView* pView,
        __in BURN_DOWNLOAD_ENGINE_INFO* pDownloadInfo,
        __in HINTERNET hUrl,
        __in HANDLE hFile,
        __in LPBYTE pbData,
        __in DWORD cbData
        )
    {
        HRESULT hr = S_OK;
        DWORD cbReadData = 0;

        hr = FileSetPointer(hFile, pDownloadInfo->dw64ResumeOffset, NULL, FILE_BEGIN);
        ExitOnFailure(hr, "Failed to seek to start point in file.");

        do
        {
            hr = pView->WaitIfPaused();
            ExitOnFailure(hr, "User canceled while paused.");

            BOOL fContinue = pView->OnCacheProgress(pPackage, pPayload, pDownloadInfo->dw64ResumeOffset, pDownloadInfo->dw64Length);
            if (!fContinue)
            {
                hr = HRESULT_FROM_WIN32(ERROR_INSTALL_USEREXIT);
                ExitOnFailure(hr, "User canceled during progress.");
            }

            // Read bits from the internet.
            if (!::InternetReadFile(hUrl, static_cast<void*>(pbData), cbData, &cbReadData))
            {
                ExitWithLastError(hr, "Failed while reading from internet.");
            }

            // Write bits to disk (if there are any).
            if (cbReadData)
            {
                DWORD cbTotalWritten = 0;
                DWORD cbWritten = 0;
                do
                {
                    if (!::WriteFile(hFile, pbData + cbTotalWritten, cbReadData - cbTotalWritten, &cbWritten, NULL))
                    {
                        ExitWithLastError(hr, "Failed to write data from internet.");
                    }

                    cbTotalWritten += cbWritten;
                } while (cbWritten && cbTotalWritten < cbReadData);

                pDownloadInfo->dw64ResumeOffset += cbTotalWritten;

                // TODO: update the resume offest for this payload.
            }
        } while (cbReadData);

    LExit:
        return hr;
    }

    //
    // OpenRequest - opens an HTTP request.
    //
    HRESULT OpenRequest(
        __in HINTERNET hConnect,
        __in_z_opt LPCWSTR wzMethod,
        __in INTERNET_SCHEME scheme,
        __in_z LPCWSTR wzResource,
        __in_z_opt LPCWSTR wzHeader,
        __out HINTERNET* phUrl
        )
    {
        HRESULT hr = S_OK;
        HINTERNET hUrl = NULL;
        DWORD dwRequestFlags = INTERNET_FLAG_KEEP_CONNECTION | INTERNET_FLAG_NO_CACHE_WRITE | INTERNET_FLAG_NO_UI | INTERNET_FLAG_RELOAD;

        if (INTERNET_SCHEME_HTTPS == scheme)
        {
            dwRequestFlags |= INTERNET_FLAG_SECURE;
        }

        hUrl = ::HttpOpenRequestW(hConnect, wzMethod, wzResource, NULL, NULL, BURN_DOWNLOAD_ENGINE_ACCEPT_TYPES, dwRequestFlags, NULL);
        ExitOnNullWithLastError(hUrl, hr, "Failed to open internet request.");

        if (wzHeader && *wzHeader)
        {
            if (!::HttpAddRequestHeadersW(hUrl, wzHeader, static_cast<DWORD>(-1), HTTP_ADDREQ_FLAG_COALESCE))
            {
                ExitWithLastError(hr, "Failed to add header to HTTP request.");
            }
        }

        *phUrl = hUrl;
        hUrl = NULL;

    LExit:
        ReleaseInternet(hUrl);
        return hr;
    }

    void DestroyDownloadInfo(
        __in BURN_DOWNLOAD_ENGINE_INFO* pDownloadInfo
        )
    {
        ReleaseInternet(pDownloadInfo->hConnect);
        ReleaseStr(pDownloadInfo->sczResource);
        ReleaseStr(pDownloadInfo->sczUrl);
        MemFree(pDownloadInfo);
    }

private:
    HINTERNET m_hSession;
};


//
// BurnCreateDownloadEngine - creates a new IBurnCacheEngine object for downloading from the internet.
//
HRESULT BurnCreateDownloadEngine(
    __out IBurnCacheEngine **ppEngine
    )
{
    HRESULT hr = S_OK;

    CBurnDownloadEngine* pDownloadEngine = new CBurnDownloadEngine(&hr);
    ExitOnNull(pDownloadEngine, hr, E_OUTOFMEMORY, "Failed to create new download engine.");
    ExitOnFailure(hr, "Failed to initialize download engine.");

    *ppEngine = pDownloadEngine;
    pDownloadEngine = NULL;

LExit:
    ReleaseObject(pDownloadEngine);
    return hr;
}
