//-------------------------------------------------------------------------------------------------
// <copyright file="BurnExeEngine.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


struct BURN_EXECUTABLE_ENGINE_INFO
{
    IBurnExePackage* pExePackage;
    IBurnView* pView;

    ACTION_STATE action;
    BOOL fExecutingRollback;

    LPWSTR sczArguments;
    LPWSTR sczExecutablePath;
};


class CBurnExeEngine : public CUnknownImpl<IBurnExecuteEngine>
{
public: // IBurnExecuteEngine
    //
    // Detect - gets the current state of a package.
    //
    STDMETHODIMP Detect(
        __in IBurnPackage* pPackage,
        __in IBurnProperties* pProperties,
        __out PACKAGE_STATE* pState,
        __out DWORD64* pqwInstalledVersion
        )
    {
        HRESULT hr = S_OK;
        IBurnExePackage* pExePackage = NULL;
        LPWSTR sczDetectCondition = NULL;
        BOOL fDetected = FALSE;

        hr = pPackage->QueryInterface(IID_PPV_ARGS(&pExePackage));
        ExitOnFailure(hr, "Was not passed a valid EXE package interface.");

        // If there is a detection condition, use that to tell if we've installed otherwise
        // we have to assume that the package is not installed already.
        hr = pExePackage->GetDetectCondition(&sczDetectCondition);
        if (SUCCEEDED(hr))
        {
            hr = pProperties->EvaluateCondition(sczDetectCondition, &fDetected);
            ExitOnFailure(hr, "Failed to evaluate executable package detect condition.");
        }
        else if (E_NOTIMPL == hr)
        {
            fDetected = FALSE;
            hr = S_OK;
        }
        ExitOnFailure(hr, "Failed to get detect condition.");

        *pState = fDetected ? PACKAGE_STATE_PRESENT : PACKAGE_STATE_ABSENT;
        *pqwInstalledVersion = 0; // TODO: should we have a search mechanism to get the installed version ?

    LExit:
        ReleaseStr(sczDetectCondition);
        ReleaseObject(pExePackage);
        return hr;
    }


    //
    // Plan - calculates the execute and rollback state for the requested package state.
    //
    STDMETHODIMP Plan(
        __in IBurnPackage* pPackage,
        __in IBurnProperties* pProperties,
        __in REQUEST_STATE requested,
        __out ACTION_STATE* pExecute,
        __out ACTION_STATE* pRollback
        )
    {
        HRESULT hr = S_OK;
        IBurnExePackage* pExePackage = NULL;
        PACKAGE_STATE state = PACKAGE_STATE_UNKNOWN;

        // TODO: Currently we ignore the version information because we can't guarantee that the package will
        // actually set any version information in detect. We should revisit this decision some time.
        pPackage->GetInstalledState(&state, NULL);

        if (PACKAGE_STATE_PRESENT == state)
        {
            hr = pPackage->QueryInterface(IID_PPV_ARGS(&pExePackage));
            ExitOnFailure(hr, "Was not passed a valid EXE package interface.");

            if (REQUEST_STATE_REPAIR == requested && pExePackage->IsRepairAllowed())
            {
                *pExecute = ACTION_STATE_RECACHE;
                *pRollback = ACTION_STATE_NONE;
            }
            else if (REQUEST_STATE_PRESENT == requested || REQUEST_STATE_REPAIR == requested)
            {
                *pExecute = ACTION_STATE_NONE;
                *pRollback = ACTION_STATE_NONE;
            }
            else if (PACKAGE_STATE_ABSENT == requested)
            {
                *pExecute = ACTION_STATE_UNINSTALL;
                *pRollback = ACTION_STATE_INSTALL;
            }
        }
        else // package not installed.
        {
            if (REQUEST_STATE_PRESENT == requested || REQUEST_STATE_REPAIR == requested)
            {
                *pExecute = ACTION_STATE_INSTALL;
                *pRollback = ACTION_STATE_UNINSTALL;
            }
            else
            {
                *pExecute = ACTION_STATE_NONE;
                *pRollback = ACTION_STATE_NONE;
            }
        }

    LExit:
        ReleaseObject(pExePackage);
        return hr;
    }


    //
    // Execute - execute the state changes for a package.
    //
    STDMETHODIMP Execute(
        __in IBurnPackage* pPackage,
        __in IBurnView* pView
        )
    {
        HRESULT hr = S_OK;
        BURN_EXECUTABLE_ENGINE_INFO* pInfo = NULL;

        hr = StartExecute(pPackage, pView, FALSE, &pInfo);
        ExitOnFailure(hr, "Failed to start executing executable package.");

        hr = DoExecute(m_pInstallationAbstractionLayer, pInfo);
        ExitOnFailure(hr, "Failed to execute executable package.");

    LExit:
        if (pInfo)
        {
            StopExecute(pInfo);
        }

        return hr;
    }


    //
    // Rollback - rollback the state changes for a package.
    //
    STDMETHODIMP Rollback(
        __in IBurnPackage* pPackage,
        __in IBurnView* pView
        )
    {
        HRESULT hr = S_OK;
        BURN_EXECUTABLE_ENGINE_INFO* pInfo = NULL;

        hr = StartExecute(pPackage, pView, TRUE, &pInfo);
        ExitOnFailure(hr, "Failed to start executing executable package.");

        hr = DoExecute(m_pInstallationAbstractionLayer, pInfo);
        ExitOnFailure(hr, "Failed to execute executable package.");

    LExit:
        if (pInfo)
        {
            StopExecute(pInfo);
        }

        return hr;
    }

    STDMETHODIMP BeginTransaction(
        __out HANDLE* phChangeOfOwnerEvent
        )
    {
        HRESULT hr = E_NOTIMPL;
        return hr;
    }

    STDMETHODIMP JoinTransaction(
        __out HANDLE* phChangeOfOwnerEvent
        )
    {
        HRESULT hr = E_NOTIMPL;
        return hr;
    }

    STDMETHODIMP EndTransaction(
        __in INSTALL_TRANSACTION transactionState
        )
    {
        HRESULT hr = E_NOTIMPL;
        return hr;
    }


private:
    static HRESULT CBurnExeEngine::StartExecute(
        __in IBurnPackage* pPackage,
        __in IBurnView* pView,
        __in BOOL fRollback,
        __out BURN_EXECUTABLE_ENGINE_INFO** ppInfo
        )
    {
        HRESULT hr = S_OK;
        IBurnExePackage* pExePackage = NULL;
        IBurnPayload* pPayload = NULL;
        ACTION_STATE execute = ACTION_STATE_NONE;
        ACTION_STATE rollback = ACTION_STATE_NONE;
        ACTION_STATE action = ACTION_STATE_NONE;
        LPWSTR sczExecutablePath = NULL;
        LPWSTR sczArguments = NULL;
        BURN_EXECUTABLE_ENGINE_INFO* pInfo = NULL;

        hr = pPackage->QueryInterface(IID_PPV_ARGS(&pExePackage));
        ExitOnFailure(hr, "Was not passed a valid EXE package interface.");

        // Get the correct command-line switch for the package action state.
        pExePackage->GetActionState(&execute, &rollback);
        action = fRollback ? rollback : execute;

        switch (action)
        {
        case ACTION_STATE_INSTALL:
            hr = pExePackage->GetInstallSwitch(&sczArguments);
            break;

        case ACTION_STATE_UNINSTALL:
            hr = pExePackage->GetUninstallSwitch(&sczArguments);
            break;

        default:
            hr = E_NOTIMPL;
        }
        ExitOnFailure(hr, "Failed to get action arguments for executable package.");

        // Always use the first payload as the installation package.
        hr = pExePackage->GetPayload(1, &pPayload);
        ExitOnFailure(hr, "Failed to get executable installation payload.");

        hr = pPayload->GetCompletedPath(&sczExecutablePath);
        ExitOnFailure(hr, "Failed to get completed path for executable installation payload.");

        // Finally, populate the engine information with the gathered data.
        pInfo = static_cast<BURN_EXECUTABLE_ENGINE_INFO*>(MemAlloc(sizeof(BURN_EXECUTABLE_ENGINE_INFO), TRUE));
        ExitOnNull(pInfo, hr, E_OUTOFMEMORY, "Failed to allocate executable info.");

        if (sczArguments && *sczArguments)
        {
            hr = StrAllocFormatted(&pInfo->sczArguments, L"\"%s\" %s", sczExecutablePath, sczArguments);
        }
        else
        {
            hr = StrAllocFormatted(&pInfo->sczArguments, L"\"%s\"", sczExecutablePath);
        }
        ExitOnFailure(hr, "Failed to format arguments string into executable engine info.");

        pInfo->sczExecutablePath = sczExecutablePath;
        sczExecutablePath = NULL;

        pInfo->action = action;
        pInfo->fExecutingRollback = fRollback;

        pInfo->pExePackage = pExePackage;
        pExePackage = NULL;

        pView->AddRef();
        pInfo->pView = pView;

        *ppInfo = pInfo;
        pInfo = NULL;

    LExit:
        if (pInfo)
        {
            StopExecute(pInfo);
        }

        ReleaseStr(sczArguments);
        ReleaseStr(sczExecutablePath);
        ReleaseObject(pPayload);
        ReleaseObject(pExePackage);
        ReleaseMem(pInfo);

        return hr;
    }


    static HRESULT CBurnExeEngine::DoExecute(
        __in IInstallationAbstractionLayer* pInstallationAbstractionLayer,
        __in BURN_EXECUTABLE_ENGINE_INFO* pInfo
        )
    {
        HRESULT hr = S_OK;
        STARTUPINFOW si = { 0 };
        PROCESS_INFORMATION pi = { 0 };
        DWORD dwExitCode = ERROR_SUCCESS;
        HRESULT hrExitCode = S_OK;

        si.cb = sizeof(si); // TODO: hookup the stdin/stdout/stderr pipes for logging purposes?
        hr = pInstallationAbstractionLayer->CreateProcess(pInfo->sczExecutablePath, pInfo->sczArguments, NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &si, &pi);
        ExitOnFailure1(hr, "Failed to CreateProcess on path: %S", pInfo->sczExecutablePath);

        if (WAIT_OBJECT_0 != ::WaitForSingleObject(pi.hProcess, INFINITE))
        {
            ExitOnFailure1(hr = E_UNEXPECTED, "Unexpected terminated process: %S", pInfo->sczExecutablePath);
        }

        hr = pInstallationAbstractionLayer->GetExitCodeProcess(pi.hProcess, &dwExitCode);
        ExitOnFailure1(hr, "Failed to get process exit code: %S", pInfo->sczExecutablePath);

        hr = ProcessExitCode(pInfo, dwExitCode, &hrExitCode);
        ExitOnFailure(hr, "Failed to process exit code.");

        hr = hrExitCode;

    LExit:
        if (pi.hProcess)
        {
            ::CloseHandle(pi.hProcess);
        }

        if (pi.hThread)
        {
            ::CloseHandle(pi.hThread);
        }

        return hr;
    }


    static HRESULT CBurnExeEngine::ProcessExitCode(
        __in BURN_EXECUTABLE_ENGINE_INFO* pInfo,
        __in DWORD dwExitCode,
        __out HRESULT* phrExitCode
        )
    {
        HRESULT hr = S_OK;
        BURN_EXE_EXIT_CODE_TYPE typeCode = BURN_EXE_EXIT_CODE_TYPE_NONE;
        DWORD cCodes = 0;

        cCodes = pInfo->pExePackage->ExitInfoCount();
        for (DWORD i = 0; i < cCodes; ++i)
        {
            BURN_EXE_EXIT_CODE_TYPE type;
            DWORD dwCode;
            BOOL fWildcard;

            hr = pInfo->pExePackage->GetExitInfo(i, &type, &dwCode, &fWildcard);
            ExitOnFailure(hr, "Failed to get exit information.");

            // If this is a wildcard, use the last one we come across.
            if (fWildcard)
            {
                typeCode = type;
            }
            else if (dwExitCode == dwCode) // If we have an exact match on the error code use that and stop looking.
            {
                typeCode = type;
                break;
            }
        }

        // If we didn't find a matching code then say that 0 == success and
        // everything else is an error.
        if (BURN_EXE_EXIT_CODE_TYPE_NONE == typeCode)
        {
            if (0 == dwExitCode)
            {
                typeCode = BURN_EXE_EXIT_CODE_TYPE_SUCCESS;
            }
            else
            {
                typeCode = BURN_EXE_EXIT_CODE_TYPE_ERROR;
            }
        }

        switch (typeCode)
        {
        case BURN_EXE_EXIT_CODE_TYPE_SUCCESS:
            *phrExitCode = S_OK;
            break;

        case BURN_EXE_EXIT_CODE_TYPE_ERROR:
            *phrExitCode = HRESULT_FROM_WIN32(ERROR_INSTALL_FAILURE);
            break;

        case BURN_EXE_EXIT_CODE_TYPE_SCHEDULE_REBOOT:
            *phrExitCode = HRESULT_FROM_WIN32(ERROR_SUCCESS_REBOOT_REQUIRED);
            break;

        case BURN_EXE_EXIT_CODE_TYPE_FORCE_REBOOT:
            *phrExitCode = HRESULT_FROM_WIN32(ERROR_INSTALL_SUSPEND);
            break;

        default:
            hr = E_UNEXPECTED;
            break;
        }

    LExit:
        return hr;
    }


    static void CBurnExeEngine::StopExecute(
        __in BURN_EXECUTABLE_ENGINE_INFO* pInfo
        )
    {
        ReleaseStr(pInfo->sczArguments);
        ReleaseStr(pInfo->sczExecutablePath);
        ReleaseObject(pInfo->pExePackage);
        ReleaseObject(pInfo->pView);
        ReleaseMem(pInfo);
    }


public:
    //
    // Constructor - intitialize member variables.
    //
    CBurnExeEngine(
        __in IInstallationAbstractionLayer* pInstallationAbstractionLayer
        )
    {
        pInstallationAbstractionLayer->AddRef();
        m_pInstallationAbstractionLayer = pInstallationAbstractionLayer;
    }

    //
    // Destructor - release member variables.
    //
    ~CBurnExeEngine()
    {
        ReleaseObject(m_pInstallationAbstractionLayer);
    }

private:
    IInstallationAbstractionLayer* m_pInstallationAbstractionLayer;
};


//
// BurnCreateExeEngine - creates a new IBurnExecuteEngine object.
//
HRESULT BurnCreateExeEngine(
    __in IInstallationAbstractionLayer* pInstallationAbstractionLayer,
    __out IBurnExecuteEngine **ppExeEngine
    )
{
    HRESULT hr = S_OK;

    CBurnExeEngine* pExeEngine = new CBurnExeEngine(pInstallationAbstractionLayer);
    ExitOnNull(pExeEngine, hr, E_OUTOFMEMORY, "Failed to create new EXE execute engine object.");

    *ppExeEngine = pExeEngine;
    pExeEngine = NULL;

LExit:
    ReleaseObject(pExeEngine);
    return hr;
}
