//-------------------------------------------------------------------------------------------------
// <copyright file="BurnExtractEngine.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

class CBurnExtractEngine : public CUnknownImpl<IBurnCacheEngine>
{
public: // IBurnCacheEngine
    //
    // CachePayload - cache a payload (UX or package payloads).
    //
    STDMETHODIMP CachePayload(
        __in IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in IBurnView* pView
        )
    {
        HRESULT hr = S_OK;
        LPWSTR sczSourcePath = NULL;
        LPWSTR sczWorkingPath = NULL;
        LPWSTR sczWorkingDir = NULL;
        LPWSTR sczEmbeddedId = NULL;
        LPWSTR sczExtractedPath = NULL;
        SOURCE_TYPE sourceType = SOURCE_TYPE_NONE;
        DWORD64 dwEmbeddedContainerOffset = 0;

        DBG_UNREFERENCED_PARAMETER(pView);

        hr = pPayload->GetSource(&sourceType, &sczSourcePath);
        ExitOnFailure(hr, "Failed to get source path for payload.");

        hr = pPayload->GetWorkingPath(&sczWorkingPath);
        ExitOnFailure(hr, "Failed to get working path for payload.");

        hr =  CBurnCacheServices::EnsureWorkingDirectory(sczWorkingPath, &sczWorkingDir);
        ExitOnFailure(hr, "Failed create working directory for payload.");

        switch (sourceType)
        {
        case SOURCE_TYPE_EMBEDDED:
            LogStringLine(REPORT_STANDARD, "Extracting embedded payload from source '%ls' to working path '%ls'", sczSourcePath, sczWorkingPath);

            hr = pPayload->GetEmbeddedSource(&dwEmbeddedContainerOffset, &sczEmbeddedId);
            ExitOnFailure1(hr, "Failed to query embedded payload information from: %S.", sczSourcePath);

            hr = CabExtract(sczSourcePath, sczEmbeddedId, sczWorkingDir, NULL, NULL, dwEmbeddedContainerOffset);
            ExitOnFailure1(hr, "Failed to extract embedded payload from: %S.", sczSourcePath);

            // the file is extracted to the specified path with a name of its cabinet token, so rename it
            hr = PathConcat(sczWorkingDir, sczEmbeddedId, &sczExtractedPath);
            ExitOnFailure(hr, "Failed to create path to extracted payload.");

            hr = FileEnsureMove(sczExtractedPath, sczWorkingPath, TRUE, TRUE);
            ExitOnFailure(hr, "Failed to rename extracted payload to its working path.");

            break;

        case SOURCE_TYPE_EXTERNAL:
            LogStringLine(REPORT_STANDARD, "Copying external payload from source '%ls' to working path '%ls'", sczSourcePath, sczWorkingPath);

            // copying external payloads to a temporary working path ensures that a multi-disk
            // installation won't lock files on disk#1.
            hr = FileEnsureCopy(sczSourcePath, sczWorkingPath, TRUE);
            ExitOnFailure2(hr, "Failed to copy payload from %S to %S", sczSourcePath, sczWorkingPath);

            break;

        case SOURCE_TYPE_DOWNLOAD:
            LogStringLine(REPORT_STANDARD, "Extract engine cannot cache source '%ls' marked for download.", sczSourcePath);
            hr = E_INVALIDARG;

            break;
        }

    LExit:
        ReleaseStr(sczSourcePath);
        ReleaseStr(sczWorkingPath);
        ReleaseStr(sczWorkingDir);
        ReleaseStr(sczEmbeddedId);
        ReleaseStr(sczExtractedPath);

        return hr;
    }

    //
    // Complete - move a payload from its working path to its completed path
    //
    STDMETHODIMP Complete(
        __in IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in IBurnView* pView
        )
    {
        HRESULT hr = S_OK;

        DBG_UNREFERENCED_PARAMETER(pView);

        hr = CBurnCacheServices::CompletePayload(pPayload);
        ExitOnFailure(hr, "Failed to complete payload.");

    LExit:
        return hr;
    }


public: // internals


public:
    //
    // Constructor - intitialize member variables.
    //
    CBurnExtractEngine()
    {
        m_cReferences = 1;
        CabInitialize(FALSE);
    }

    //
    // Destructor - release member variables.
    //
    ~CBurnExtractEngine()
    {
        CabUninitialize();
    }

private:
};


//
// BurnCreateExtractEngine - creates a new IBurnExtractEngine object from a filename.
//
HRESULT BurnCreateExtractEngine(
    __out IBurnCacheEngine **ppEngine
    )
{
    HRESULT hr = S_OK;

    CBurnExtractEngine* pExtractEngine = new CBurnExtractEngine;
    ExitOnNull(pExtractEngine, hr, E_OUTOFMEMORY, "Failed to create new extract cache engine object.");
    ExitOnFailure(hr, "Failed to initialize extract engine.");

    *ppEngine = pExtractEngine;
    pExtractEngine = NULL;

LExit:
    ReleaseObject(pExtractEngine);
    return hr;
}
