//-------------------------------------------------------------------------------------------------
// <copyright file="BurnInstallationAbstractionLayer.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

class CBurnInstallationAbstractionLayer : public CUnknownImpl<IInstallationAbstractionLayer>
{
public: // IInstallationAbstractionLayer
    STDMETHODIMP CreateProcess(
        __in LPCWSTR lpApplicationName,
        __inout LPWSTR lpCommandLine,
        __in_opt LPSECURITY_ATTRIBUTES lpProcessAttributes,
        __in_opt LPSECURITY_ATTRIBUTES lpThreadAttributes,
        __in BOOL bInheritHandles,
        __in DWORD dwCreationFlags,
        __in_opt LPVOID lpEnvironment,
        __in_opt LPCWSTR lpCurrentDirectory,
        __in LPSTARTUPINFOW lpStartupInfo,
        __out LPPROCESS_INFORMATION lpProcessInformation
        )
    {
        HRESULT hr = S_OK;

        if (!::CreateProcessW(lpApplicationName, lpCommandLine, lpProcessAttributes, lpThreadAttributes, bInheritHandles, dwCreationFlags, lpEnvironment, lpCurrentDirectory, lpStartupInfo, lpProcessInformation))
        {
            ExitWithLastError(hr, "Failed to create process.");
        }

    LExit:
        return hr;
    }


    STDMETHODIMP GetExitCodeProcess(
        __in HANDLE hProcess,
        __out LPDWORD lpExitCode
        )
    {
        HRESULT hr = S_OK;

        if (!::GetExitCodeProcess(hProcess, lpExitCode))
        {
            ExitWithLastError(hr, "Failed to get process exit code.");
        }

    LExit:
        return hr;
    }


    //
    // IsElevated - determines if current process is elevated. E_NOTIMPL is returned if the
    //              operating system does not support elevation.
    //
    IFACEMETHODIMP IsElevated()
    {
        HRESULT hr = S_OK;
        HANDLE hToken = NULL;
        TOKEN_ELEVATION tokenInfo = { };
        DWORD cbData = sizeof(tokenInfo);

        if (!::OpenProcessToken(::GetCurrentProcess(), TOKEN_QUERY, &hToken))
        {
            ExitWithLastError(hr, "Failed to get process token to check for elevation.");
        }

        if (!::GetTokenInformation(hToken, TokenElevation, &tokenInfo, cbData, &cbData))
        {
            // If the TokenElevation argument is not recognized, we're on a "downlevel" OS.
            DWORD er = ::GetLastError();
            if (ERROR_INVALID_PARAMETER == er)
            {
                ExitFunction1(hr = E_NOTIMPL);
            }

            hr = HRESULT_FROM_WIN32(er);
            ExitOnRootFailure(hr, "Failed to get elevation token information about current process.");
        }

        hr = tokenInfo.TokenIsElevated ? S_OK : S_FALSE;

    LExit:
        if (hToken)
        {
            ::CloseHandle(hToken);
        }

        return hr;
    }


    STDMETHODIMP MsiGetProductInfo(
        __in LPCWSTR szProduct,
        __in LPCWSTR szProperty,
        __out LPWSTR* psczValue
        )
    {
        return WiuGetProductInfo(szProduct, szProperty, psczValue);
    }


    //
    // ShellExecute - executes a shell command.
    //
    STDMETHODIMP ShellExecute(
        __in SHELLEXECUTEINFOW* pInfo
        )
    {
        HRESULT hr = S_OK;

        if (!::ShellExecuteExW(pInfo))
        {
            ExitWithLastError(hr, "Failed to shell execute process.");
        }

    LExit:
        return hr;
    }


    STDMETHODIMP_(INSTALLSTATE) MsiQueryFeatureState(
        __in LPCWSTR szProduct,
        __in LPCWSTR szFeature
        )
    {
        INSTALLSTATE state = ::MsiQueryFeatureStateW(szProduct, szFeature);
        return state;
    }


    STDMETHODIMP_(INSTALLUILEVEL) MsiSetInternalUI(
        __in INSTALLUILEVEL dwUILevel,
        __inout HWND* phWnd
        )
    {
        INSTALLUILEVEL level = ::MsiSetInternalUI(dwUILevel, phWnd);
        return level;
    }


    STDMETHODIMP MsiSetExternalUI(
        __in INSTALLUI_HANDLERW puiHandler,
        __in DWORD dwMessageFilter,
        __in LPVOID pvContext,
        __out_opt INSTALLUI_HANDLERW* ppuiPrevHandler
        )
    {
        INSTALLUI_HANDLERW puiPrevHandler = ::MsiSetExternalUIW(puiHandler, dwMessageFilter, pvContext);
        if (ppuiPrevHandler)
        {
            *ppuiPrevHandler = puiPrevHandler;
        }

        return S_OK;
    }


    STDMETHODIMP MsiSetExternalUIRecord(
        __in INSTALLUI_HANDLER_RECORD puiHandler,
        __in DWORD dwMessageFilter,
        __in LPVOID pvContext,
        __out_opt PINSTALLUI_HANDLER_RECORD ppuiPrevHandler
        )
    {
        HRESULT hr = S_OK;

        if (m_pfnMsiSetExternalUIRecord)
        {
            DWORD er = (*m_pfnMsiSetExternalUIRecord)(puiHandler, dwMessageFilter, pvContext, ppuiPrevHandler);
            hr = HRESULT_FROM_WIN32(er);
        }
        else
        {
            hr = E_NOTIMPL;
        }
        ExitOnFailure(hr, "Failed to set external UI record handler.");

    LExit:
        return hr;
    }


    STDMETHODIMP MsiEnableLog(
        __in DWORD dwLogMode,
        __in LPCWSTR szLogFile,
        __in DWORD dwLogAttributes
        )
    {
        HRESULT hr = S_OK;

        DWORD er = ::MsiEnableLogW(dwLogMode, szLogFile, dwLogAttributes);
        hr = HRESULT_FROM_WIN32(er);
        ExitOnFailure(hr, "Failed to enable log.");

    LExit:
        return hr;
    }


    STDMETHODIMP MsiBeginTransaction(
        __in LPCWSTR szName,
        __in DWORD dwTransactionAttributes,
        __out MSIHANDLE* phTransactionHandle,
        __out HANDLE* phChangeOfOwnerEvent
        )
    {
        HRESULT hr = S_OK;

        if (m_pfnMsiBeginTransaction)
        {
            DWORD er = (*m_pfnMsiBeginTransaction)(szName, dwTransactionAttributes, phTransactionHandle, phChangeOfOwnerEvent);
            hr = HRESULT_FROM_WIN32(er);
            ExitOnFailure(hr, "Failed to begin multi package installation.");
        }
        else
        {
            hr = E_NOTIMPL;
            ExitOnFailure(hr, "Multi package installation not supported.");
        }

    LExit:
        return hr;
    }

    STDMETHODIMP MsiEndTransaction(
        __in DWORD dwTransactionState
        )        
    {
        HRESULT hr = S_OK;

        if (m_pfnMsiEndTransaction)
        {
            DWORD er = (*m_pfnMsiEndTransaction)(dwTransactionState);
            hr = HRESULT_FROM_WIN32(er);
            ExitOnFailure(hr, "Failed to end multi package installation.");
        }
        else
        {
            hr = E_NOTIMPL;
            ExitOnFailure(hr, "Multi package installation not supported.");
        }

    LExit:
        return hr;
    }


    STDMETHODIMP MsiJoinTransaction(
        __in MSIHANDLE hTransactionID,
        __in DWORD dwTransactionAttributes,
        __out HANDLE* phChangeOfOwnerEvent
        )
    {
        HRESULT hr = S_OK;

        if (m_pfnMsiJoinTransaction)
        {
            DWORD er = (*m_pfnMsiJoinTransaction)(hTransactionID, dwTransactionAttributes, phChangeOfOwnerEvent);
            hr = HRESULT_FROM_WIN32(er);
            ExitOnFailure(hr, "Failed to join multi package installation.");
        }
        else
        {
            hr = E_NOTIMPL;
            ExitOnFailure(hr, "Multi package installation not supported.");
        }

    LExit:
        return hr;
    }


    STDMETHODIMP MsiInstallProduct(
        __in LPCWSTR szPackagePath,
        __in LPCWSTR szCommandLine
        )
    {
        HRESULT hr = S_OK;

        DWORD er = ::MsiInstallProductW(szPackagePath, szCommandLine);
        hr = HRESULT_FROM_WIN32(er);
        ExitOnFailure(hr, "Failed to install product.");

    LExit:
        return hr;
    }


    STDMETHODIMP MsiConfigureProductEx(
        __in LPCWSTR szProduct,
        __in int iInstallLevel,
        __in INSTALLSTATE eInstallState,
        __in LPCWSTR szCommandLine
        )
    {
        HRESULT hr = S_OK;

        DWORD er = ::MsiConfigureProductExW(szProduct, iInstallLevel, eInstallState, szCommandLine);
        hr = HRESULT_FROM_WIN32(er);
        ExitOnFailure(hr, "Failed to configure product.");

    LExit:
        return hr;
    }


    //
    // ScheduleRestart - reboots the local machine using "software installation" as the reason.
    //
    STDMETHODIMP ScheduleRestart()
    {
        HRESULT hr = S_OK;
        TOKEN_PRIVILEGES priv = { 0 };

        priv.PrivilegeCount = 1;
        priv.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
        if (!::LookupPrivilegeValueW(NULL, L"SeShutdownPrivilege", &priv.Privileges[0].Luid))
        {
            ExitOnLastError(hr, "Failed to get shutdown privilege LUID.");
        }

        if (!::AdjustTokenPrivileges(::GetCurrentProcess(), FALSE, &priv, sizeof(TOKEN_PRIVILEGES), NULL, 0))
        {
            ExitOnLastError(hr, "Failed to get shutdown privilege LUID.");
        }

        if (!::ExitWindowsEx(EWX_REBOOT, SHTDN_REASON_MAJOR_APPLICATION | SHTDN_REASON_MINOR_INSTALLATION | SHTDN_REASON_FLAG_PLANNED))
        {
            ExitOnLastError(hr, "Failed to schedule restart.");
        }

    LExit:
        return hr;
    }


public:
    //
    // Constructor - intitialize member variables.
    //
    CBurnInstallationAbstractionLayer()
    {
        m_hMsi = NULL;
        m_pfnMsiSetExternalUIRecord = NULL;
        m_pfnMsiBeginTransaction = NULL;
        m_pfnMsiEndTransaction = NULL;
        m_pfnMsiJoinTransaction = NULL;
    }

    //
    // Destructor - release member variables.
    //
    ~CBurnInstallationAbstractionLayer()
    {

        if (m_hMsi)
        {
            ::FreeLibrary(m_hMsi);
        }
    }

    //
    // Initialize - initializes abstraction layer.
    //
    HRESULT Initialize()
    {
        HRESULT hr = S_OK;

        m_hMsi = ::LoadLibraryW(L"MSI.DLL");
        ExitOnNullWithLastError(m_hMsi, hr, "Failed to load MSI.DLL");

        // Try to get ::MsiSetExternalUIRecord(), but ignore any failures if it isn't found.
        m_pfnMsiSetExternalUIRecord = reinterpret_cast<PFN_MSISETEXTERNALUIRECORD>(::GetProcAddress(m_hMsi, "MsiSetExternalUIRecord"));
        m_pfnMsiBeginTransaction = reinterpret_cast<PFN_MSIBEGINTRANSACTION>(::GetProcAddress(m_hMsi, "MsiBeginTransactionW"));
        m_pfnMsiEndTransaction = reinterpret_cast<PFN_MSIENDTRANSACTION>(::GetProcAddress(m_hMsi, "MsiEndTransaction"));
        m_pfnMsiJoinTransaction = reinterpret_cast<PFN_MSIJOINTRANSACTION>(::GetProcAddress(m_hMsi, "MsiJoinTransaction"));        

    LExit:
        return hr;
    }

private:
    HMODULE m_hMsi;
    PFN_MSISETEXTERNALUIRECORD m_pfnMsiSetExternalUIRecord;
    PFN_MSIBEGINTRANSACTION m_pfnMsiBeginTransaction;
    PFN_MSIENDTRANSACTION m_pfnMsiEndTransaction;
    PFN_MSIJOINTRANSACTION m_pfnMsiJoinTransaction;
};


//
// BurnCreateInstallationAbstractionLayer - creates a new IInstallationAbstractionLayer.
//
HRESULT BurnCreateInstallationAbstractionLayer(
    __out IInstallationAbstractionLayer **ppInstallationAbstractionLayer
    )
{
    HRESULT hr = S_OK;

    CBurnInstallationAbstractionLayer* pInstallationAbstractionLayer = new CBurnInstallationAbstractionLayer();
    ExitOnNull(pInstallationAbstractionLayer, hr, E_OUTOFMEMORY, "Failed to create new installation abstraction object.");

    hr = pInstallationAbstractionLayer->Initialize();
    ExitOnFailure(hr, "Failed to initialize abstraction layer.");

    *ppInstallationAbstractionLayer = pInstallationAbstractionLayer;
    pInstallationAbstractionLayer = NULL;

LExit:
    ReleaseObject(pInstallationAbstractionLayer);
    return hr;
}
