//-------------------------------------------------------------------------------------------------
// <copyright file="BurnMsiEngine.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

const DWORD BURN_MSI_PROGRESS_INVALID = 0xFFFFFFFF;

struct BURN_MSI_PROGRESS
{
    DWORD dwTotal;
    DWORD dwCompleted;
    DWORD dwStep;
    BOOL fMoveForward;
    BOOL fEnableActionData;
    BOOL fScriptInProgress;
};


//
// CBurnMsiEngine - MSI execute engine.
//
class CBurnMsiEngine : public CUnknownImpl<IBurnExecuteEngine>
{
public: // IBurnExecuteEngine
    //
    // Detect - gets the current state of a package.
    //
    STDMETHODIMP Detect(
        __in IBurnPackage* pPackage,
        __in IBurnProperties* /* pProperties */,
        __out PACKAGE_STATE* pInstalledState,
        __out DWORD64* pqwInstalledVersion
        )
    {
        Trace1(REPORT_STANDARD, "Detecting MSI package 0x%p", pPackage);

        HRESULT hr = S_OK;
        IBurnMsiPackage* pMsiPackage = NULL;
        LPWSTR sczProductCode = NULL;
        LPWSTR sczInstalledVersion = NULL;
        DWORD dwInstalledMajorVersion = 0;
        DWORD dwInstalledMinorVersion = 0;

        hr = pPackage->QueryInterface(__uuidof(IBurnMsiPackage), reinterpret_cast<void**>(&pMsiPackage));
        ExitOnFailure(hr, "Was not passed a valid MSI package interface.");

        hr = pMsiPackage->GetProductCode(&sczProductCode);
        ExitOnFailure(hr, "Failed to get ProductCode.");

        hr = m_pInstallationAbstractionLayer->MsiGetProductInfo(sczProductCode, L"VersionString", &sczInstalledVersion);
        if (SUCCEEDED(hr))
        {
            hr = FileVersionFromString(sczInstalledVersion, &dwInstalledMajorVersion, &dwInstalledMinorVersion);
            ExitOnFailure2(hr, "Failed to convert version: %S to DWORDs for ProductCode: %S", sczInstalledVersion, sczProductCode);

            *pInstalledState = PACKAGE_STATE_PRESENT; 
            *pqwInstalledVersion = static_cast<DWORD64>(dwInstalledMajorVersion) << 32 | dwInstalledMinorVersion;
        }
        else if (HRESULT_FROM_WIN32(ERROR_UNKNOWN_PRODUCT) == hr || HRESULT_FROM_WIN32(ERROR_UNKNOWN_PROPERTY) == hr) // package not present.
        {
            *pInstalledState = PACKAGE_STATE_ABSENT;
            *pqwInstalledVersion = 0;

            hr = S_OK;
        }
        else
        {
            ExitOnRootFailure1(hr, "Failed to get product information for ProductCode: %S", sczProductCode);
        }

    LExit:
        ReleaseStr(sczInstalledVersion);
        ReleaseStr(sczProductCode);
        ReleaseObject(pMsiPackage);
        return hr;
    }


    //
    // Plan - calculates the execute and rollback state for the requested package state.
    //
    STDMETHODIMP Plan(
        __in IBurnPackage* pPackage,
        __in IBurnProperties* /* pProperties */,
        __in REQUEST_STATE requested,
        __out ACTION_STATE* pExecute,
        __out ACTION_STATE* pRollback
        )
    {
        Trace1(REPORT_STANDARD, "Planning MSI package 0x%p", pPackage);

        HRESULT hr = S_OK;
        PACKAGE_STATE state = PACKAGE_STATE_UNKNOWN;
        DWORD64 qwInstalledVersion = 0;
        DWORD64 qwVersion = 0;

        pPackage->GetInstalledState(&state, &qwInstalledVersion);

        pPackage->GetVersion(&qwVersion);

        if (PACKAGE_STATE_PRESENT == state)
        {
            if (REQUEST_STATE_PRESENT == requested || REQUEST_STATE_REPAIR == requested)
            {
                //
                // Take a look at the version and determine if this is a potential
                // minor upgrade, a major upgrade (just install) otherwise, there
                // is a newer version so no work necessary.
                //
                // minor upgrade "10.3.2.0" > "10.3.1.0" or "10.3.2.1" > "10.3.2.0"
                if ((qwVersion >> 32) == (qwInstalledVersion >> 32) && qwVersion > qwInstalledVersion)
                {
                    *pExecute = ACTION_STATE_MINOR_UPGRADE;
                    *pRollback = ACTION_STATE_NONE;
                }
                else if (qwVersion > qwInstalledVersion) // major upgrade "11.X.X.X" > "10.X.X.X"
                {
                    *pExecute = ACTION_STATE_MAJOR_UPGRADE;
                    *pRollback = ACTION_STATE_NONE;
                }
                else if (qwVersion == qwInstalledVersion) // maintenance install "10.X.X.X" = "10.X.X.X"
                {
                    *pExecute = (REQUEST_STATE_REPAIR == requested) ? ACTION_STATE_RECACHE : ACTION_STATE_MAINTENANCE;
                    *pRollback = ACTION_STATE_NONE;
                }
                else // newer version present "14.X.X.X" < "15.X.X.X", skip
                {
                    *pExecute = ACTION_STATE_NONE; // TODO: inform about this state to enable warnings
                    *pRollback = ACTION_STATE_NONE;
                }
            }
            else if (PACKAGE_STATE_ABSENT == requested)
            {
                if (qwVersion == qwInstalledVersion) // install "10.X.X.X" = "10.X.X.X"
                {
                    *pExecute = ACTION_STATE_UNINSTALL;
                    *pRollback = ACTION_STATE_INSTALL;
                }
                else
                {
                    *pExecute = ACTION_STATE_NONE; // TODO: inform about this state to enable warnings
                    *pRollback = ACTION_STATE_NONE;
                }
            }
        }
        else // package not installed.
        {
            if (REQUEST_STATE_PRESENT == requested || REQUEST_STATE_REPAIR == requested)
            {
                *pExecute = ACTION_STATE_INSTALL;
                *pRollback = ACTION_STATE_UNINSTALL;
            }
            else
            {
                *pExecute = ACTION_STATE_NONE;
                *pRollback = ACTION_STATE_NONE;
            }
        }

        return hr;
    }


    //
    // Execute - execute the state changes for a package.
    //
    STDMETHODIMP Execute(
        __in IBurnPackage* pPackage,
        __in IBurnView* pView
        )
    {
        Trace1(REPORT_STANDARD, "Executing MSI package 0x%p", pPackage);

        HRESULT hr = S_OK;
        hr = SetExecutingPackage(pPackage, pView, FALSE);
        ExitOnFailure(hr, "Failed to set executing package.");

        hr = ConfigureExecutingMsi();
        ExitOnFailure(hr, "Failed to execute MSI package.");

    LExit:
        ResetExecutingPackage();
        return hr;
    }


    //
    // Rollback - rollback the state changes for a package.
    //
    STDMETHODIMP Rollback(
        __in IBurnPackage* pPackage,
        __in IBurnView* pView
        )
    {
        Trace1(REPORT_STANDARD, "Rolling back MSI package 0x%p", pPackage);

        HRESULT hr = S_OK;
        hr = SetExecutingPackage(pPackage, pView, TRUE);
        ExitOnFailure(hr, "Failed to set executing package for rollback.");

        hr = ConfigureExecutingMsi();
        ExitOnFailure(hr, "Failed to rollback MSI package.");

    LExit:
        ResetExecutingPackage();
        return hr;
    }

    STDMETHODIMP BeginTransaction(
        __out HANDLE* phChangeOfOwnerEvent
        )
    {
        HRESULT hr = S_OK;
        m_phTransactionHandle = NULL;

        hr = m_pInstallationAbstractionLayer->MsiBeginTransaction(L"Burn Installer Engine", 0, &m_phTransactionHandle, phChangeOfOwnerEvent);
        ExitOnFailure(hr, "Multi-package installations not supported.");

    LExit:
        return hr;
    }

    STDMETHODIMP JoinTransaction(
        __out HANDLE* phChangeOfOwnerEvent
        )
    {
        HRESULT hr = S_OK;
        hr = m_pInstallationAbstractionLayer->MsiJoinTransaction(m_phTransactionHandle, 0, phChangeOfOwnerEvent);
        ExitOnFailure(hr, "Multi-package installations not supported.");

    LExit:
        return hr;
    }

    STDMETHODIMP EndTransaction(
        __in INSTALL_TRANSACTION transactionState
        )
    {
        HRESULT hr = S_OK;
        hr = m_pInstallationAbstractionLayer->MsiEndTransaction(transactionState);
        ExitOnFailure(hr, "Multi-package installations not supported.");

    LExit:
        return hr;
    }

private:
    HRESULT SetExecutingPackage(
        __in IBurnPackage* pPackage,
        __in IBurnView* pView,
        __in BOOL fRollback
        )
    {
        HRESULT hr = S_OK;
        IBurnMsiPackage* pMsiPackage = NULL;
        ACTION_STATE execute = ACTION_STATE_NONE;
        ACTION_STATE rollback = ACTION_STATE_NONE;

        hr = pPackage->QueryInterface(__uuidof(IBurnMsiPackage), reinterpret_cast<void**>(&pMsiPackage));
        ExitOnFailure(hr, "Was not passed a valid MSI package interface.");

        pMsiPackage->GetActionState(&execute, &rollback);

        m_ExecutingAction = fRollback ? rollback : execute;
        m_fExecutingRollback = fRollback;

        m_pExecutingPackage = pMsiPackage;
        pMsiPackage = NULL;

        pView->AddRef();
        m_pExecutingView = pView;

    LExit:
        ReleaseObject(pMsiPackage);
        return hr;
    }


    void ResetExecutingPackage()
    {
        m_ExecutingAction = ACTION_STATE_NONE;
        m_fExecutingRollback = FALSE;
        ReleaseNullObject(m_pExecutingPackage);
        ReleaseNullObject(m_pExecutingView);
    }


    //
    // ConfigureExecutingMsi - configure the MSI package marked for executing.
    //
    HRESULT ConfigureExecutingMsi()
    {
        AssertSz(m_pExecutingView && m_pExecutingPackage, "Expected SetExecutingPackage() to be called before ConfigureExecutingMsi().");

        HRESULT hr = S_OK;

        LPWSTR sczProductCode = NULL;
        IBurnPayload* pPayload = NULL;
        LPWSTR sczMsiPath = NULL;

        DWORD dwUiLevel = INSTALLUILEVEL_NONE | INSTALLUILEVEL_SOURCERESONLY;
        DWORD dwMessageFilter = 0;
        INSTALLUI_HANDLERW pfnPreviousExternalUI = NULL;
        INSTALLUI_HANDLER_RECORD pfnPreviousExternalUIRecord = NULL;
        LPWSTR sczLogFile = NULL;
        BURN_LOGGING_LEVEL logLevel = BURN_LOGGING_LEVEL_UNKNOWN;
        LPWSTR sczProperties = NULL;

        hr = m_pExecutingPackage->GetProductCode(&sczProductCode);
        ExitOnFailure(hr, "Failed to get ProductCode.");

        // Always use the first payload as the installation package.
        hr = m_pExecutingPackage->GetPayload(1, &pPayload);
        ExitOnFailure(hr, "Failed to get installation payload.");

        hr = pPayload->GetCompletedPath(&sczMsiPath);
        ExitOnFailure(hr, "Failed to get completed path for installation payload.");

        /*
        switch (uiLevel)
        {
        case MSI_UI_LEVEL_DEFAULT:
            dwUiLevel = INSTALLUILEVEL_DEFAULT;
            break;
        case MSI_UI_LEVEL_NONE:
            dwUiLevel = INSTALLUILEVEL_NONE;
            if (pEngineInfo->fEnableSourceResPrompt)
            {
                dwUiLevel |= INSTALLUILEVEL_SOURCERESONLY;
            }
            break;
        case MSI_UI_LEVEL_BASIC:
            dwUiLevel = INSTALLUILEVEL_BASIC;
            break;
        case MSI_UI_LEVEL_REDUCED:
            dwUiLevel = INSTALLUILEVEL_REDUCED;
            break;
        case MSI_UI_LEVEL_FULL:
            dwUiLevel = INSTALLUILEVEL_FULL;
            break;
        default:
            ExitFunction1(hr = E_UNEXPECTED);
        }
        */

        // Wire up logging and the external UI handler.
        m_pInstallationAbstractionLayer->MsiSetInternalUI(static_cast<INSTALLUILEVEL>(dwUiLevel), NULL);

        dwMessageFilter = INSTALLLOGMODE_INITIALIZE | INSTALLLOGMODE_TERMINATE |
                          INSTALLLOGMODE_FATALEXIT | INSTALLLOGMODE_ERROR | INSTALLLOGMODE_WARNING |
                          INSTALLLOGMODE_RESOLVESOURCE | INSTALLLOGMODE_OUTOFDISKSPACE |
                          INSTALLLOGMODE_ACTIONSTART | INSTALLLOGMODE_ACTIONDATA | INSTALLLOGMODE_COMMONDATA | INSTALLLOGMODE_PROGRESS;

        // If the external UI record is available (MSI version >= 3.1) use it but fall back to the standard external
        // UI handler if necesary.
        hr = m_pInstallationAbstractionLayer->MsiSetExternalUIRecord(InstallEngineRecordCallback, dwMessageFilter, static_cast<LPVOID>(this), &pfnPreviousExternalUIRecord);
        if (FAILED(hr))
        {
            hr = m_pInstallationAbstractionLayer->MsiSetExternalUI(InstallEngineCallback, dwMessageFilter, static_cast<LPVOID>(this), &pfnPreviousExternalUI);
            ExitOnFailure(hr, "Failed to wire up external UI handler.");
        }

        // Enable logging based on the manifest settings or global log command-line switch.
        hr = m_pExecutingPackage->GetLogPath(&sczLogFile, &logLevel);
        if (SUCCEEDED(hr))
        {
            // default to "normal" logging
            DWORD dwLogMode = 
                INSTALLLOGMODE_FATALEXIT | INSTALLLOGMODE_ERROR | INSTALLLOGMODE_WARNING |
                INSTALLLOGMODE_USER | INSTALLLOGMODE_INFO | INSTALLLOGMODE_RESOLVESOURCE |
                INSTALLLOGMODE_OUTOFDISKSPACE | INSTALLLOGMODE_ACTIONSTART | INSTALLLOGMODE_ACTIONDATA |
                INSTALLLOGMODE_COMMONDATA | INSTALLLOGMODE_PROPERTYDUMP;
            if (BURN_LOGGING_LEVEL_VERBOSE == logLevel)
            {
                dwLogMode |= INSTALLLOGMODE_VERBOSE;
            }
            else if (BURN_LOGGING_LEVEL_DEBUG == logLevel)
            {
                dwLogMode |= INSTALLLOGMODE_VERBOSE | INSTALLLOGMODE_EXTRADEBUG;
            }

            // always append because InitializeLog truncated the log if overwrite was requested
            hr = m_pInstallationAbstractionLayer->MsiEnableLog(dwLogMode, sczLogFile, INSTALLLOGATTRIBUTES_APPEND);
            if (SUCCEEDED(hr))
            {
                LogStringLine(REPORT_STANDARD, "MSI log file for package '%ls' set to: %ls", m_pExecutingPackage->Id(), sczLogFile);
            }
        }

        // Ignore all failures setting up the UI handler and
        // logging, they aren't important enough to abort the
        // install attempt.
        hr = S_OK;

        // set up our properties
        hr = GetProperties(&sczProperties);
        ExitOnFailure(hr, "Failed to allocate string for MSI install arguments");

        //
        // Do the actual action.
        //
        switch (m_ExecutingAction)
        {
        case ACTION_STATE_ADMIN_INSTALL: __fallthrough;
            hr = StrAllocConcat(&sczProperties, L" ACTION=ADMIN", 0);
            ExitOnFailure(hr, "Failed to format property: ADMIN");
        case ACTION_STATE_MAJOR_UPGRADE: __fallthrough;
        case ACTION_STATE_INSTALL:
            hr = m_pInstallationAbstractionLayer->MsiInstallProduct(sczMsiPath, sczProperties);
            break;

        case ACTION_STATE_MINOR_UPGRADE:
            hr = StrAllocConcat(&sczProperties, L" REINSTALL=ALL REINSTALLMODE=\"vomus\"", 0);
            ExitOnFailure(hr, "failed to format properties: REINSTALL and REINSTALLMODE");

            hr = m_pInstallationAbstractionLayer->MsiInstallProduct(sczMsiPath, sczProperties);
            break;

        case ACTION_STATE_MAINTENANCE: __fallthrough;
        case ACTION_STATE_RECACHE:
            hr = StrAllocConcat(&sczProperties, L" REINSTALLMODE=\"vdmus\"", 0);
            ExitOnFailure(hr, "failed to format property: REINSTALLMODE");

            hr = m_pInstallationAbstractionLayer->MsiConfigureProductEx(sczProductCode, INSTALLLEVEL_DEFAULT, INSTALLSTATE_DEFAULT, sczProperties);
            break;

        case ACTION_STATE_UNINSTALL:
            hr = m_pInstallationAbstractionLayer->MsiConfigureProductEx(sczProductCode, INSTALLLEVEL_DEFAULT, INSTALLSTATE_ABSENT, sczProperties);
            break;
        }

    LExit:
        if (pfnPreviousExternalUI)  // unset the UI handler
        {
            m_pInstallationAbstractionLayer->MsiSetExternalUI(pfnPreviousExternalUI, 0, NULL, NULL);
        }

        if (pfnPreviousExternalUIRecord)  // unset the UI record handler
        {
            m_pInstallationAbstractionLayer->MsiSetExternalUIRecord(pfnPreviousExternalUIRecord, 0, NULL, NULL);
        }

        ReleaseStr(sczProperties);
        ReleaseStr(sczLogFile);

        ReleaseStr(sczMsiPath);
        ReleaseObject(pPayload);
        ReleaseStr(sczProductCode);

        return hr;
    }


    HRESULT GetProperties(
        __out_z LPWSTR* psczProperties
        )
    {
        return StrAllocString(psczProperties, L"", 0);
    //    HRESULT hr = S_OK;
    //    IBurnProperties* pProperties = NULL;
    //    DWORD cProperties = 0;

    //    LPWSTR sczName = NULL;
    //    LPWSTR sczValue = NULL;
    //    LPWSTR sczNameValue = NULL;

    //    hr = m_pExecutingPackage->GetProperties(&pProperties);
    //    ExitOnFailure(hr, "Failed to get package properties.");

    //    cProperties = pProperties->Count();

    //    for (DWORD i = 0; i < cProperties; ++i)
    //    {
    //        hr = pProperties->GetName(i, &sczName);
    //        ExitOnFailure(hr, "Failed to get property name.");

    //        hr = pProperties->GetStringValueByIndex(i, &sczValue);
    //        ExitOnFailure(hr, "Failed to get property value.");

    //        hr = StrAllocFormatted(&sczNameValue, L" %ls=\"%ls\"", sczName, sczValue);
    //        ExitOnFailure(hr, "Failed to allocate name value pair.");

    //        hr = StrAllocConcat(psczProperties, sczNameValue, 0);
    //        ExitOnFailure(hr, "Failed to add name value pair to properties.");

    //        ReleaseNullStr(sczNameValue);
    //        ReleaseNullStr(sczValue);
    //        ReleaseNullStr(sczName);
    //    }

    //LExit:
    //    ReleaseStr(sczNameValue);
    //    ReleaseStr(sczValue);
    //    ReleaseStr(sczName);
    //    ReleaseObject(pProperties);
    //    return hr;
    }


    static INT CALLBACK InstallEngineCallback(
        __in LPVOID pvContext,
        __in UINT uiMessage,
        __in LPCWSTR wzMessage
        )
    {
        INT nResult = IDERROR; // assume something will fail before the message is handled.
        CBurnMsiEngine* pEngine = static_cast<CBurnMsiEngine*>(pvContext);
        INSTALLMESSAGE mt = static_cast<INSTALLMESSAGE>(0xFF000000 & uiMessage);
        UINT uiFlags = 0x00FFFFFF & uiMessage;

        // Bail if we weren't passed the engine.
        if (!pEngine)
        {
            ExitFunction();
        }

        if (INSTALLMESSAGE_PROGRESS == mt)
        {
            nResult = pEngine->HandleInstallProgress(wzMessage, NULL);
        }
        else
        {
            nResult = pEngine->HandleInstallMessage(mt, uiFlags, wzMessage, NULL);
        }

    LExit:
        return nResult;
    }


    static INT CALLBACK InstallEngineRecordCallback(
        __in LPVOID pvContext,
        __in UINT uiMessage,
        __in MSIHANDLE hRecord
        )
    {
        INT nResult = IDERROR; // assume something will fail before the message is handled.
        HRESULT hr = S_OK;

        CBurnMsiEngine* pEngine = static_cast<CBurnMsiEngine*>(pvContext);
        INSTALLMESSAGE mt = static_cast<INSTALLMESSAGE>(0xFF000000 & uiMessage);
        UINT uiFlags = 0x00FFFFFF & uiMessage;
        LPWSTR sczMessage = NULL;
        DWORD cchMessage = 0;

        // Bail if we weren't passed the engine.
        if (!pEngine)
        {
            ExitFunction();
        }

        if (INSTALLMESSAGE_PROGRESS == mt)
        {
            nResult = pEngine->HandleInstallProgress(NULL, hRecord);
        }
        else
        {
            // create formated message string
            if (hRecord)
            {
#pragma prefast(push)
#pragma prefast(disable:6298) // docs explicitly say this is a valid option for getting the buffer size
                DWORD er = ::MsiFormatRecordW(NULL, hRecord, L"", &cchMessage);
#pragma prefast(pop)
                if (ERROR_MORE_DATA == er || ERROR_SUCCESS == er)
                {
                    hr = StrAlloc(&sczMessage, ++cchMessage);
                }
                else
                {
                    hr = HRESULT_FROM_WIN32(er);
                }
                ExitOnFailure(hr, "Failed to allocate string for formated message.");

                er = ::MsiFormatRecordW(NULL, hRecord, sczMessage, &cchMessage);
                hr = HRESULT_FROM_WIN32(er);
                ExitOnRootFailure(hr, "Failed to format message record.");
            }

            // Pass to handler including both the formated message and the original record.
            nResult = pEngine->HandleInstallMessage(mt, uiFlags, sczMessage, hRecord);
        }

    LExit:
        ReleaseStr(sczMessage);
        return nResult;
    }


    INT HandleInstallMessage(
        __in INSTALLMESSAGE mt,
        __in UINT uiFlags,
        __in LPCWSTR wzMessage,
        __in_opt MSIHANDLE hRecord
        )
    {
        INT nResult = IDOK;

Trace2(REPORT_STANDARD, "install[%x]: %ls", m_dwCurrentProgressIndex, wzMessage);

        switch (mt)
        {
        case INSTALLMESSAGE_INITIALIZE: // this message is received prior to internal UI initialization, no string data
            ResetProgress();
            break;

        case INSTALLMESSAGE_TERMINATE: // sent after UI termination, no string data
            break;

        case INSTALLMESSAGE_ACTIONSTART:
            if (BURN_MSI_PROGRESS_INVALID != m_dwCurrentProgressIndex && m_rgMsiProgress[m_dwCurrentProgressIndex].fEnableActionData)
            {
                m_rgMsiProgress[m_dwCurrentProgressIndex].fEnableActionData = FALSE;
            }

            nResult = m_pExecutingView->OnExecuteMsiMessage(m_pExecutingPackage, mt, uiFlags, wzMessage, hRecord);
            break;

        case INSTALLMESSAGE_ACTIONDATA:
            if (BURN_MSI_PROGRESS_INVALID != m_dwCurrentProgressIndex && m_rgMsiProgress[m_dwCurrentProgressIndex].fEnableActionData)
            {
                if (m_rgMsiProgress[m_dwCurrentProgressIndex].fMoveForward)
                {
                    m_rgMsiProgress[m_dwCurrentProgressIndex].dwCompleted += m_rgMsiProgress[m_dwCurrentProgressIndex].dwStep;
                }
                else
                {
                    m_rgMsiProgress[m_dwCurrentProgressIndex].dwCompleted -= m_rgMsiProgress[m_dwCurrentProgressIndex].dwStep;
                }
Trace3(REPORT_STANDARD, "progress[%x]: actiondata, progress: %u  forward: %d", m_dwCurrentProgressIndex, m_rgMsiProgress[m_dwCurrentProgressIndex].dwStep, m_rgMsiProgress[m_dwCurrentProgressIndex].fMoveForward);

                nResult = SendProgressUpdate();
            }
            else
            {
                nResult = m_pExecutingView->OnExecuteMsiMessage(m_pExecutingPackage, mt, uiFlags, wzMessage, hRecord);
            }
            break;

        case INSTALLMESSAGE_RESOLVESOURCE:
            m_pExecutingView->OnExecuteMsiMessage(m_pExecutingPackage, mt, uiFlags, wzMessage, hRecord);
            nResult = IDNOACTION; // always return no action (0) for resolve source.
            break;

        case INSTALLMESSAGE_OUTOFDISKSPACE: __fallthrough;
        case INSTALLMESSAGE_FATALEXIT: __fallthrough;
        case INSTALLMESSAGE_ERROR:
            {
            DWORD dwErrorCode = 0;
            if (hRecord)
            {
                dwErrorCode = ::MsiRecordGetInteger(hRecord, 1);
            }

            nResult = m_pExecutingView->OnExecuteError(m_pExecutingPackage, dwErrorCode, wzMessage, uiFlags);
            }
            break;

/*
        case INSTALLMESSAGE_WARNING:
        case INSTALLMESSAGE_USER:
        case INSTALLMESSAGE_INFO:
        case INSTALLMESSAGE_RMFILESINUSE:
        case INSTALLMESSAGE_FILESINUSE:
        case INSTALLMESSAGE_SHOWDIALOG: // sent prior to display of authored dialog or wizard

#if 0
        case INSTALLMESSAGE_COMMONDATA:
            if (L'1' == wzMessage[0] && L':' == wzMessage[1] && L' ' == wzMessage[2])
            {
                if (L'0' == wzMessage[3])
                {
                    // TODO: handle the language common data message.
                    lres = IDOK;
                    return lres;
                }
                else if (L'1' == wzMessage[3])
                {
                    // TODO: really handle sending the caption.
                    lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_SET_CAPTION, uiFlags, reinterpret_cast<LPARAM>(wzMessage + 3));
                    return lres;
                }
                else if (L'2' == wzMessage[3])
                {
                    // TODO: really handle sending the cancel button status.
                    lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_SET_CANCEL, uiFlags, reinterpret_cast<LPARAM>(wzMessage + 3));
                    return lres;
                }
            }
            break;
#endif
*/

        default:
            nResult = m_pExecutingView->OnExecuteMsiMessage(m_pExecutingPackage, mt, uiFlags, wzMessage, hRecord);
            break;
        }

        return nResult;
    }


    INT HandleInstallProgress(
        __in_z_opt LPCWSTR wzMessage,
        __in_opt MSIHANDLE hRecord
        )
    {
        HRESULT hr = S_OK;
        INT nResult = IDOK;
        INT iFields[4] = { 0 };
        INT cFields = 0;

        // get field values
        if (wzMessage)
        {
            // parse message string
            cFields = swscanf_s(wzMessage, L"1: %d 2: %d 3: %d 4: %d", &iFields[0], &iFields[1], &iFields[2], &iFields[3]);
        }
        else if (hRecord)
        {
            cFields = ::MsiRecordGetFieldCount(hRecord);
            cFields = min(cFields, countof(iFields)); // no buffer overrun if there are more fields than our buffer can hold
            for (INT i = 0; i < cFields; ++i)
            {
                iFields[i] = ::MsiRecordGetInteger(hRecord, i + 1);
            }
        }

#ifdef _DEBUG
WCHAR wz[256];
swprintf_s(wz, countof(wz), L"1: %d 2: %d 3: %d 4: %d", iFields[0], iFields[1], iFields[2], iFields[3]);
Trace2(REPORT_STANDARD, "progress[%x]: %ls", m_dwCurrentProgressIndex, wz);
#endif

        // verify that we have enought field values
        if (1 > cFields)
        {
            ExitFunction(); // unknown message, bail
        }

        // hande based on message type
        switch (iFields[0])
        {
        case 0: // master progress reset
            if (4 > cFields)
            {
                Trace2(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - Invalid field count %d, '%S'", cFields, wzMessage);
                ExitFunction();
            }
            //Trace3(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - MASTER RESET - %d, %d, %d", iFields[1], iFields[2], iFields[3]);

            // Update the index into progress array.
            if (BURN_MSI_PROGRESS_INVALID == m_dwCurrentProgressIndex)
            {
                m_dwCurrentProgressIndex = 0;
            }
            else if (m_dwCurrentProgressIndex + 1 < countof(m_rgMsiProgress))
            {
                ++m_dwCurrentProgressIndex;
            }
            else
            {
                hr = HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_BUFFER);
                ExitOnRootFailure(hr, "(Insufficient space to hold progress information.");
            }

            // we only care about the first stage after script execution has started
            //if (!pEngineInfo->fMsiProgressScriptInProgress && 1 != iFields[3])
            //{
            //    pEngineInfo->fMsiProgressFinished = TRUE;
            //}

            m_rgMsiProgress[m_dwCurrentProgressIndex].dwTotal = iFields[1];
            m_rgMsiProgress[m_dwCurrentProgressIndex].dwCompleted = 0 == iFields[2] ? 0 : iFields[1]; // if forward start at 0, if backwards start at max
            m_rgMsiProgress[m_dwCurrentProgressIndex].fMoveForward = (0 == iFields[2]);
            m_rgMsiProgress[m_dwCurrentProgressIndex].fEnableActionData = FALSE;
            m_rgMsiProgress[m_dwCurrentProgressIndex].fScriptInProgress = (1 == iFields[3]);

            if (0 == m_dwCurrentProgressIndex)
            {
                // HACK!!! this is a hack courtesy of the Windows Installer team. It seems the script planning phase
                // is always off by "about 50".  So we'll toss an extra 50 ticks on so that the standard progress
                // doesn't go over 100%.  If there are any custom actions, they may blow the total so we'll call this
                // "close" and deal with the rest.
                m_rgMsiProgress[m_dwCurrentProgressIndex].dwTotal += 50;
            }
            break;

        case 1: // action info
            if (3 > cFields)
            {
                Trace2(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - Invalid field count %d, '%S'", cFields, wzMessage);
                ExitFunction();
            }
            //Trace3(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - ACTION INFO - %d, %d, %d", iFields[1], iFields[2], iFields[3]);

            if (0 == iFields[2])
            {
                m_rgMsiProgress[m_dwCurrentProgressIndex].fEnableActionData = FALSE;
            }
            else
            {
                m_rgMsiProgress[m_dwCurrentProgressIndex].fEnableActionData = TRUE;
                m_rgMsiProgress[m_dwCurrentProgressIndex].dwStep = iFields[1];
            }
            break;

        case 2: // progress report
            if (2 > cFields)
            {
                Trace2(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - Invalid field count %d, '%S'", cFields, wzMessage);
                break;
            }
            //Trace3(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - PROGRESS REPORT - %d, %d, %d", iFields[1], iFields[2], iFields[3]);

            if (BURN_MSI_PROGRESS_INVALID == m_dwCurrentProgressIndex)
            {
                break;
            }
            else if (0 == m_rgMsiProgress[m_dwCurrentProgressIndex].dwTotal)
            {
                break;
            }

            // update progress
            if (m_rgMsiProgress[m_dwCurrentProgressIndex].fMoveForward)
            {
                m_rgMsiProgress[m_dwCurrentProgressIndex].dwCompleted += iFields[1];
            }
            else
            {
                m_rgMsiProgress[m_dwCurrentProgressIndex].dwCompleted -= iFields[1];
            }
            break;

        case 3:
            m_rgMsiProgress[m_dwCurrentProgressIndex].dwTotal += iFields[1];
            break;

        default:
            ExitFunction(); // unknown message, bail
        }

        // If we have a valid progress index, send an update.
        if (BURN_MSI_PROGRESS_INVALID != m_dwCurrentProgressIndex)
        {
            nResult = SendProgressUpdate();
        }

    LExit:
        return nResult;
    }


    //
    // SendProgressUpdate
    //
    INT SendProgressUpdate()
    {
        DWORD dwPercentage = 0; // number representing 0 - 100%
        int nResult = IDNOACTION;

        //DWORD dwMsiProgressTotal = pEngineInfo->dwMsiProgressTotal;
        //DWORD dwMsiProgressComplete = pEngineInfo->dwMsiProgressComplete; //min(dwMsiProgressTotal, pEngineInfo->dwMsiProgressComplete);
        //double dProgressGauge = 0;
        //double dProgressStageTotal = (double)pEngineInfo->qwProgressStageTotal;

        // Calculate progress for the phases of Windows Installer.
        // TODO: handle upgrade progress which would add another phase.
        dwPercentage += CalculatePhaseProgress(0, 15);
        dwPercentage += CalculatePhaseProgress(1, 80);
        dwPercentage += CalculatePhaseProgress(2, 5);

        //if (qwTotal) // avoid "divide by zero" if the MSI range is blank.
        //{
        //    // calculate gauge.
        //    double dProgressGauge = static_cast<double>(qwCompleted) / static_cast<double>(qwTotal);
        //    dProgressGauge = (1.0 / (1.0 + exp(3.7 - dProgressGauge * 7.5)) - 0.024127021417669196) / 0.975872978582330804;
        //    qwCompleted = (DWORD)(dProgressGauge * qwTotal);

        //    // calculate progress within range
        //    //qwProgressComplete = (DWORD64)(dwMsiProgressComplete * (dProgressStageTotal / dwMsiProgressTotal));
        //    //qwProgressComplete = min(qwProgressComplete, pEngineInfo->qwProgressStageTotal);
        //}

        dwPercentage = min(dwPercentage, 100);
        nResult = m_pExecutingView->OnExecuteProgress(m_pExecutingPackage, dwPercentage);

#ifdef _DEBUG
DWORD64 qwCompleted = m_rgMsiProgress[m_dwCurrentProgressIndex].dwCompleted;
DWORD64 qwTotal = m_rgMsiProgress[m_dwCurrentProgressIndex].dwTotal;
Trace3(REPORT_STANDARD, "progress: %I64u/%I64u (%u%%)", qwCompleted, qwTotal, dwPercentage);
//AssertSz(qwCompleted <= qwTotal, "Completed progress is larger than total progress.");
#endif

        return nResult;
    }


    //
    // ResetProgress - clears out any previous progress.
    //
    void ResetProgress()
    {
        ZeroMemory(m_rgMsiProgress, sizeof(m_rgMsiProgress));
        m_dwCurrentProgressIndex = BURN_MSI_PROGRESS_INVALID;
    }


    //
    // CalculatePhaseProgress - calculates the progress of a phase with a weighted percentage.
    //                          For example, 5 / 10 weighted 50% equals 2 ticks.
    //
    DWORD CalculatePhaseProgress(DWORD dwProgressIndex, DWORD dwWeightPercentage)
    {
        DWORD dwPhasePercentage = 0;

        // If we've already passed this progress index, return the maximum percentage possible (the weight)
        if (dwProgressIndex < m_dwCurrentProgressIndex)
        {
            dwPhasePercentage = dwWeightPercentage;
        }
        else if (dwProgressIndex == m_dwCurrentProgressIndex) // have to do the math for the current progress.
        {
            BURN_MSI_PROGRESS* pProgress = m_rgMsiProgress + dwProgressIndex;
            if (pProgress->dwTotal)
            {
                DWORD64 dw64Completed = m_fExecutingRollback ? pProgress->dwTotal - pProgress->dwCompleted : pProgress->dwCompleted;
                dwPhasePercentage = static_cast<DWORD>(dw64Completed * dwWeightPercentage / pProgress->dwTotal);
            }
        }
        // else we're not there yet so it has to be zero.

        return dwPhasePercentage;
    }


public:
    //
    // Constructor - intitialize member variables.
    //
    CBurnMsiEngine()
    {
        m_pExecutingView = NULL;
        m_pExecutingPackage = NULL;
        m_fExecutingRollback = FALSE;
        m_ExecutingAction = ACTION_STATE_NONE;
        m_phTransactionHandle = NULL;

        ZeroMemory(m_rgMsiProgress, sizeof(m_rgMsiProgress));
        m_dwCurrentProgressIndex = BURN_MSI_PROGRESS_INVALID;
    }

    //
    // Destructor - release member variables.
    //
    ~CBurnMsiEngine()
    {
        ReleaseNullObject(m_pExecutingPackage);
        ReleaseNullObject(m_pExecutingView);
        ReleaseNullObject(m_pInstallationAbstractionLayer);
    }


    //
    // Initialize - initializes engine member variables.
    //
    HRESULT Initialize(
        __in IInstallationAbstractionLayer* pInstallationAbstractionLayer
        )
    {
        HRESULT hr = S_OK;

        pInstallationAbstractionLayer->AddRef();
        m_pInstallationAbstractionLayer = pInstallationAbstractionLayer;

        return hr;
    }

private:
    IInstallationAbstractionLayer* m_pInstallationAbstractionLayer;

    IBurnView* m_pExecutingView;
    IBurnMsiPackage* m_pExecutingPackage;
    BOOL m_fExecutingRollback;
    ACTION_STATE m_ExecutingAction;
    MSIHANDLE m_phTransactionHandle;

    BURN_MSI_PROGRESS m_rgMsiProgress[64];
    DWORD m_dwCurrentProgressIndex;
};


//
// BurnCreateMsiEngine - creates a new IBurnExecuteEngine object.
//
HRESULT BurnCreateMsiEngine(
    __in IInstallationAbstractionLayer* pInstallationAbstractionLayer,
    __out IBurnExecuteEngine **ppMsiEngine
    )
{
    HRESULT hr = S_OK;

    CBurnMsiEngine* pMsiEngine = new CBurnMsiEngine();
    ExitOnNull(pMsiEngine, hr, E_OUTOFMEMORY, "Failed to create new MSI execute engine object.");

    hr = pMsiEngine->Initialize(pInstallationAbstractionLayer);
    ExitOnFailure(hr, "Failed to initialize MSI execute engine.");

    *ppMsiEngine = pMsiEngine;
    pMsiEngine = NULL;

LExit:
    ReleaseObject(pMsiEngine);
    return hr;
}
