//-------------------------------------------------------------------------------------------------
// <copyright file="BurnPackageBase.h" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#pragma once

template<class T1, class T2>
class CBurnPackageBase : public CUnknownImpl2<T1, T2>
{
public: // IBurnPackage
    STDMETHODIMP_(LPCWSTR) Id()
    {
        return m_sczId;
    }


    STDMETHODIMP_(int) CompareId(
        __in_z LPCWSTR wzId
        )
    {
        return ::CompareStringW(LOCALE_INVARIANT, 0, m_sczId, -1, wzId, -1);
    }


    STDMETHODIMP Detected(
        __in PACKAGE_STATE state,
        __in DWORD64 qwInstalledVersion
        )
    {
        HRESULT hr = S_OK;

        if (m_fDetected)
        {
            hr = HRESULT_FROM_WIN32(ERROR_ALREADY_INITIALIZED);
            ExitOnRootFailure(hr, "Package already detected.");
        }

        m_fDetected = TRUE;
        m_state = state;
        m_qwInstalledVersion = qwInstalledVersion;

    LExit:
        return hr;
    }


    STDMETHODIMP_(BOOL) IsDetected()
    {
        return m_fDetected;
    }


    STDMETHODIMP_(BOOL) IsPerMachine()
    {
        return m_fPerMachine;
    }


    STDMETHODIMP_(void) GetInstalledState(
        __out_opt PACKAGE_STATE* pState,
        __out_opt DWORD64* pqwVersion
        )
    {
        if (pState)
        {
            *pState = m_state;
        }

        if (pqwVersion)
        {
            *pqwVersion = m_qwInstalledVersion;
        }
    }


    STDMETHODIMP_(void) GetActionState(
        __out_opt ACTION_STATE* pExecute,
        __out_opt ACTION_STATE* pRollback
        )
    {
        if (pExecute)
        {
            *pExecute = m_execute;
        }

        if (pRollback)
        {
            *pRollback = m_rollback;
        }
    }


    STDMETHODIMP GetProperties(
        __out IBurnProperties** ppProperties
        )
    {
        HRESULT hr = S_OK;

        if (!m_pProperties)
        {
            hr = BurnCreateProperties(&m_pProperties);
            ExitOnFailure(hr, "Failed to create property collection for package.");
        }

        m_pProperties->AddRef();
        *ppProperties = m_pProperties;

    LExit:
        return hr;
    }


    STDMETHODIMP GetPayload(
        __in DWORD dwIndex,
        __out IBurnPayload** ppPayload
        )
    {
        HRESULT hr = S_OK;

        if (0 == dwIndex || dwIndex > m_cPayloads)
        {
            ExitOnFailure1(hr = E_INVALIDARG, "Payload index not in range 1..%d", m_cPayloads);
        }

        m_rgpPayloads[dwIndex - 1]->AddRef();
        *ppPayload = m_rgpPayloads[dwIndex - 1];

    LExit:
        return hr;
    }


    STDMETHODIMP_(DWORD) PayloadCount()
    {
        return m_cPayloads;
    }


    STDMETHODIMP_(void) GetVersion(
        __out DWORD64* pqwVersion
        )
    {
        *pqwVersion = m_qwVersion;
    }


    STDMETHODIMP_(void) SetActionState(
        __in ACTION_STATE execute,
        __in ACTION_STATE rollback
        )
    {
        m_execute = execute;
        m_rollback = rollback;
    }


    STDMETHODIMP InitializeLog(
        __in BURN_LOGGING_LEVEL logLevel,
        __in BURN_LOGGING_MODE logMode,
        __in LPCWSTR wzChainLogPath,
        __in_opt LPCWSTR wzLogPath
        )
    {
        Assert(wzChainLogPath);
        Assert(m_sczId);

        HRESULT hr = S_OK;
        LPWSTR sczNewLogPath = NULL;
        LPWSTR sczLogDirectory = NULL;
        LPWSTR sczSuffix = NULL;
        HANDLE hLogFile = INVALID_HANDLE_VALUE;

        if (wzLogPath)
        {
            // resolve %environment_variables%
            hr = FileResolvePath(wzLogPath, &sczNewLogPath);
            ExitOnFailure(hr, "Failed to resolve incoming log path.");
        }
        else
        {
            // the default package log is chainLogPath\chainLogBaseName.packageId.log
            hr = StrAllocFormatted(&sczSuffix, L".%ls", m_sczId);
            ExitOnFailure(hr, "Failed to allocate memory for package id suffix.");

            hr = FileAddSuffixToBaseName(wzChainLogPath, sczSuffix, &sczNewLogPath);
            ExitOnFailure1(hr, "Failed to add package id suffix to chain log path: %ls.", wzChainLogPath);
        }

        // create the directory if one was provided
        hr = PathGetDirectory(sczNewLogPath, &sczLogDirectory);
        ExitOnFailure1(hr, "Failed to get directory from log path: %ls", wzLogPath);

        if (sczLogDirectory)
        {
            hr = DirEnsureExists(sczLogDirectory, NULL);
            ExitOnFailure(hr, "Failed to create log directory.");
        }

        // handle rename-mode collision-avoidance using PathCreateTempFile's templating
        // (the PathCreateTempFile template is <logBaseName>.%04x.<logExtension>);
        // otherwise we're appending/overwriting so create a file with the right knobs.
        if (BURN_LOGGING_MODE_UNIQUE == logMode)
        {
            hr = FileAddSuffixToBaseName(FileFromPath(sczNewLogPath), L".%04x", &sczNewLogPath);
            ExitOnFailure(hr, "Failed to allocate renamed log template.");

            hr = PathCreateTempFile(sczLogDirectory, sczNewLogPath, SHRT_MAX, FILE_ATTRIBUTE_NORMAL, &sczNewLogPath, NULL);
            ExitOnFailure(hr, "Failed to create log file.");
        }
        else
        {
            hLogFile = ::CreateFileW(sczNewLogPath, GENERIC_WRITE, FILE_SHARE_READ, NULL, (BURN_LOGGING_MODE_APPEND == logMode) ? OPEN_ALWAYS : CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
            if (INVALID_HANDLE_VALUE == hLogFile)
            {
                ExitOnLastError1(hr, "Failed to create log file: %ls", sczNewLogPath);
            }
        }

        m_logLevel = logLevel;
        m_sczLogFile = sczNewLogPath;
        sczNewLogPath = NULL;

    LExit:
        ReleaseStr(sczLogDirectory);
        ReleaseStr(sczNewLogPath);
        ReleaseStr(sczSuffix);
        ReleaseFileHandle(hLogFile);

        return hr;
    }


    //
    // GetLogPath - gets the path to the log file. May return NULL.
    //
    STDMETHODIMP GetLogPath(
        __out LPWSTR* psczLogPath,
        __out_opt BURN_LOGGING_LEVEL* pLogLevel
        )
    {
        HRESULT hr = S_OK;

        if (m_sczLogFile)
        {
            hr = StrAllocString(psczLogPath, m_sczLogFile, 0);
            ExitOnFailure(hr, "Failed to copy log path.");
        }
        else
        {
            ReleaseNullStr(*psczLogPath);
        }

        if (pLogLevel)
        {
            *pLogLevel = m_logLevel;
        }

    LExit:
        return hr;
    }


public: // IBurnPackageTestable
    STDMETHODIMP AddPayloadForTestingPurposes(
        __in IBurnPayload* pPayload
        )
    {
        HRESULT hr = S_OK;
        if (!m_rgpPayloads)
        {
            m_rgpPayloads = static_cast<IBurnPayload**>(MemAlloc(sizeof(IBurnPayload*), FALSE));
            ExitOnNull(m_rgpPayloads, hr, E_OUTOFMEMORY, "Failed to allocate memory for payloads.");
        }
        else
        {
            LPVOID pv = MemReAlloc(m_rgpPayloads, sizeof(IBurnPayload*) * m_cPayloads + 1, FALSE);
            ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to allocate more memory for payloads.");

            m_rgpPayloads = static_cast<IBurnPayload**>(pv);
        }

        pPayload->AddRef();
        m_rgpPayloads[m_cPayloads] = pPayload;
        m_rgpPayloads[m_cPayloads]->SetIndex(m_cPayloads + 1);

        ++m_cPayloads;

    LExit:
        return hr;
    }


protected:
    //
    // Constructor - intitialize member variables.
    //
    CBurnPackageBase(
        __in_z LPCWSTR wzId,
        __in DWORD64 qwVersion,
        __in BOOL fPerMachine,
        __in DWORD cPayloads,
        __in_opt IBurnPayload** rgpPayloads,
        __out HRESULT* pHR
        )
    {
        HRESULT hr = S_OK;

        m_fDetected = FALSE;
        m_state = PACKAGE_STATE_UNKNOWN;
        m_qwInstalledVersion = 0;

        m_logLevel = BURN_LOGGING_LEVEL_UNKNOWN;
        m_sczLogFile = NULL;

        m_pProperties = NULL;

        m_execute = ACTION_STATE_NONE;
        m_rollback = ACTION_STATE_NONE;

        m_sczId = NULL;
        hr = StrAllocString(&m_sczId, wzId, 0);
        ExitOnFailure(hr, "Failed to copy package id.");

        m_fPerMachine = fPerMachine;
        m_qwVersion = qwVersion;
        m_cPayloads = cPayloads;
        m_rgpPayloads = rgpPayloads;

        if (m_rgpPayloads)
        {
            for (DWORD i = 0; i < m_cPayloads; ++i)
            {
                m_rgpPayloads[i]->SetIndex(i + 1);
            }
        }

    LExit:
        *pHR = hr;
    }


    //
    // Destructor - release member variables.
    //
    virtual ~CBurnPackageBase()
    {
        ReleaseObjectArray(m_rgpPayloads, m_cPayloads);
        ReleaseObject(m_pProperties);
        ReleaseStr(m_sczId);
        ReleaseStr(m_sczLogFile);
    }

protected:
    LPWSTR m_sczId;
    DWORD64 m_qwVersion;
    BURN_LOGGING_LEVEL m_logLevel;
    LPWSTR m_sczLogFile;

    BOOL m_fDetected;
    BOOL m_fPerMachine;
    PACKAGE_STATE m_state;
    DWORD64 m_qwInstalledVersion;

    IBurnProperties* m_pProperties;

    DWORD m_cPayloads;
    IBurnPayload** m_rgpPayloads;

    ACTION_STATE m_execute;
    ACTION_STATE m_rollback;
};
