//-------------------------------------------------------------------------------------------------
// <copyright file="BurnView.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


//
// CBurnView - implements IBurnView and IBurnCore.
//
class CBurnView : public CUnknownImpl2<IBurnCore, IBurnView>
{
public: // IBurnCore
    //
    // GetPackageCount - returns the current number of packages.
    //
    STDMETHODIMP GetPackageCount(
        __out DWORD* pcPackages
        )
    {
        HRESULT hr = S_OK;

        if (!pcPackages)
        {
            hr = E_INVALIDARG;
            ExitOnRootFailure(hr, "Invalid argument provided.");
        }

        *pcPackages = m_pModel->PackageCount();

    LExit:
        return hr;
    }


    //
    // GetCommandLineParameters - gets the command line parameters passed in with the exe
    //
    STDMETHODIMP GetCommandLineParameters(
        __out LPWSTR* psczCommandLine
        )
    {
        return m_pModel->GetCommandLineParameters(psczCommandLine);
    }


    //
    // SetPackageProperty - sets (or unsets) a property in the package.
    //
    STDMETHODIMP SetPackageProperty(
        __in_z LPCWSTR wzPackageId,
        __in_z LPCWSTR wzProperty,
        __in_z_opt LPCWSTR wzValue
        )
    {
        HRESULT hr = S_OK;
        DWORD dwPackageIndex = 0;
        IBurnPackage* pPackage = NULL;
        IBurnProperties* pProperties = NULL;

        hr = m_pModel->GetPackageIndex(wzPackageId, &dwPackageIndex);
        ExitOnFailure(hr, "Failed to get package index.");

        hr = m_pModel->GetPackage(dwPackageIndex, &pPackage);
        ExitOnFailure(hr, "Failed to get package by index.");

        hr = pPackage->GetProperties(&pProperties);
        ExitOnFailure(hr, "Failed to get package properties.");

        hr = pProperties->SetString(wzProperty, wzValue);
        ExitOnFailure(hr, "Failed to set property.");

    LExit:
        ReleaseObject(pProperties);
        ReleaseObject(pPackage);
        return hr;
    }


    //
    // Elevate - tells the controller to start the elvated process.
    //
    STDMETHODIMP Elevate(
        __in HWND hwndParent
        )
    {
        HRESULT hr = S_OK;

        hr = m_pController->Elevate(hwndParent);
        ExitOnFailure(hr, "Failed to elevate controller.");

    LExit:
        return hr;
    }


    //
    // Prepare - tells the controller to start preparing packages.
    //
    STDMETHODIMP Prepare(
        __in_z_opt LPCWSTR wzPackageId,
        __in REQUEST_STATE requested
        )
    {
        HRESULT hr = S_OK;
        DWORD dwPackageIndex = 0;

        if (wzPackageId)
        {
            hr = m_pModel->GetPackageIndex(wzPackageId, &dwPackageIndex);
            ExitOnFailure(hr, "Failed to get package index.");
        }

        hr = m_pController->Prepare(dwPackageIndex, requested);
        ExitOnFailure(hr, "Failed to have controller prepare packages.");

    LExit:
        return hr;
    }


    STDMETHODIMP Cache(
        __in_z_opt LPCWSTR wzPackageId
        )
    {
        HRESULT hr = S_OK;
        DWORD dwPackageIndex = 0;

        if (wzPackageId)
        {
            hr = m_pModel->GetPackageIndex(wzPackageId, &dwPackageIndex);
            ExitOnFailure(hr, "Failed to get package index.");
        }

        hr = m_pController->Cache(dwPackageIndex);
        ExitOnFailure(hr, "Failed to have controller cache packages.");

    LExit:
        return hr;
    }


    STDMETHODIMP Execute(
        __in REQUEST_STATE requested
        )
    {
        HRESULT hr = S_OK;

        hr = m_pController->Execute(requested);
        ExitOnFailure(hr, "Failed to execute.");

    LExit:
        return hr;
    }

public: // IBurnView
    //
    // Initialize - initialize the view.
    //
    STDMETHODIMP Initialize(
        __in IBurnModel* pModel,
        __in IBurnController* pController
        )
    {
        HRESULT hr = S_OK;
        IBurnUserExperience* pUX = NULL;

        pController->AddRef();
        m_pController = pController;

        pModel->AddRef();
        m_pModel = pModel;

        hr = pController->LoadUXInterface(&pUX);
        ExitOnFailure(hr, "Failed to load UX dll.");

        hr = pUX->Initialize(this, m_nCmdShow);
        ExitOnFailure(hr, "Failed to initialize UX dll.");

        m_pUX = pUX;
        pUX = NULL;

        m_fInitialized = TRUE;

    LExit:
        if (FAILED(hr))
        {
            Uninitialize();
        }

        ReleaseObject(pUX);
        return hr;
    }


    //
    // Uninitialize - ensure all the contained objects are freed.
    //
    STDMETHODIMP_(void) Uninitialize()
    {
        m_fInitialized = FALSE;

        if (m_pUX)
        {
            m_pUX->Uninitialize();
        }

        ReleaseNullObject(m_pUX);
        ReleaseNullObject(m_pController);
        ReleaseNullObject(m_pModel);
    }


    //
    // Run - run the view.
    //
    STDMETHODIMP Run()
    {
        if (!m_pUX)
        {
            return E_UNEXPECTED;
        }

        return m_pUX->Run();
    }


    STDMETHODIMP_(int) OnPrepareBegin(
        __in DWORD cPackages
        )
    {
        if (!m_fInitialized)
        {
            return IDNOACTION; // view silently continues without UX.
        }

        int nResult = m_pUX->OnPrepareBegin(cPackages);
        return nResult;
    }


    STDMETHODIMP_(int) OnPreparePackageBegin(
        __in IBurnPackage* pPackage
        )
    {
        if (!m_fInitialized)
        {
            return IDNOACTION; // view silently continues without UX.
        }

        LPCWSTR wzId = pPackage->Id();
        int nResult = m_pUX->OnPreparePackageBegin(wzId);
        return nResult;
    }


    STDMETHODIMP_(void) OnPreparePackageComplete(
        __in IBurnPackage* pPackage,
        __in HRESULT hrStatus,
        __in PACKAGE_STATE state,
        __in REQUEST_STATE requested,
        __in ACTION_STATE execute,
        __in ACTION_STATE rollback
        )
    {
        if (!m_fInitialized)
        {
            return;
        }

        LPCWSTR wzId = pPackage->Id();
        m_pUX->OnPreparePackageComplete(wzId, hrStatus, state, requested, execute, rollback);
    }


    STDMETHODIMP_(void) OnPrepareComplete(
        __in HRESULT hrStatus
        )
    {
        if (!m_fInitialized)
        {
            return;
        }

        Trace(REPORT_STANDARD, "CBurnView::OnPrepareComplete()");
        m_pUX->OnPrepareComplete(hrStatus);
    }


    STDMETHODIMP_(int) OnCacheBegin(
        __in DWORD cCachingPackages,
        __in DWORD cTotalPayloads,
        __in DWORD64 dw64TotalCacheSize
        )
    {
        if (!m_fInitialized)
        {
            return IDNOACTION; // view silently continues without UX.
        }

        m_cCachedPackages = 0;
        m_cCachingPackages = cCachingPackages;

        int nResult = IDNOACTION; // TODO: m_pUX->OnCacheBegin(cCachingPackages, cTotalPayloads, dw64TotalCacheSize);
        return nResult;
    }


    STDMETHODIMP_(int) OnCachePackageBegin(
        __in IBurnPackage* pPackage,
        __in DWORD cPayloads,
        __in DWORD64 dw64PackageCacheSize
        )
    {
        if (!m_fInitialized)
        {
            return IDNOACTION; // view silently continues without UX.
        }

        LPCWSTR wzId = pPackage->Id();
        int nResult = IDNOACTION; // TODO: m_pUX->OnCachePackageBegin(wzId, cPayloads, dw64PackageCacheSize);
        return nResult;
    }


    STDMETHODIMP_(int) OnCachePackagePayloadBegin(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in SOURCE_TYPE sourceType,
        __in_z LPCWSTR wzSourcePath
        )
    {
        if (!m_fInitialized)
        {
            return IDNOACTION; // view silently continues without UX.
        }

        LPCWSTR wzId = pPackage->Id();
        DWORD dwIndex = pPayload->Index();
        int nResult = IDNOACTION; // TODO: m_pUX->OnCachePackagePayloadBegin(wzId, dwIndex, sourceType, wzSourcePath);
        return nResult;
    }


    STDMETHODIMP_(int) OnCacheProgress(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in DWORD64 dw64PayloadCached,
        __in DWORD64 dw64PayloadSize
        )
    {
        if (!m_fInitialized)
        {
            return IDNOACTION; // view silently continues without UX.
        }

        DWORD dwPercentage = static_cast<DWORD>(dw64PayloadCached * 100 / dw64PayloadSize); // TODO: use a formula that doesn't have overflow issues with packages over 10.2 GB.
        DWORD dwOverallProgress = 0;

        // If we know the count of packages that will be cached, calculate the "overall percentage".
        if (m_cCachingPackages)
        {
            dwOverallProgress = (m_cCachedPackages * 100 / m_cCachingPackages) + (dwPercentage / m_cCachingPackages);
        }

        int nResult = IDNOACTION; // TODO: m_pUX->OnCacheProgress(dw64Cached, dw64Total, dwPercentage, dwOverallProgress);
        return nResult;
    }


    STDMETHODIMP_(void) OnCachePackagePayloadComplete(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in HRESULT hrExitCode
        )
    {
        if (!m_fInitialized)
        {
            return;
        }

        LPCWSTR wzId = pPackage->Id();
        DWORD dwIndex = pPayload->Index();
        // TODO: m_pUX->OnCachePackagePayloadComplete(wzId, dwIndex, hrExitCode);
    }


    STDMETHODIMP_(void) OnCachePackageComplete(
        __in IBurnPackage* pPackage,
        __in HRESULT hrExitCode
        )
    {
        if (!m_fInitialized)
        {
            return;
        }

        ++m_cCachedPackages;

        LPCWSTR wzId = pPackage->Id();
        // TODO: m_pUX->OnCachePackageComplete(wzid, hrExitCode);
    }


    STDMETHODIMP_(void) OnCacheComplete(
        __in HRESULT hrStatus
        )
    {
        if (!m_fInitialized)
        {
            return;
        }

        m_cCachedPackages = 0;
        m_cCachingPackages = 0;

        m_pUX->OnCacheComplete(hrStatus);
    }


    STDMETHODIMP_(int) OnExecuteBegin(
        __in DWORD cExecutingPackages
        )
    {
        if (!m_fInitialized)
        {
            return IDNOACTION; // view silently continues without UX.
        }

        m_fExecutingRollback = FALSE;
        m_cExecutedPackages = 0;
        m_cExecutingPackages = cExecutingPackages;

        int nResult = m_pUX->OnExecuteBegin(cExecutingPackages);
        return nResult;
    }


    STDMETHODIMP_(int) OnExecutePackageBegin(
        __in IBurnPackage* pPackage,
        __in BOOL fExecute
        )
    {
        LPCWSTR wzId = pPackage->Id();
        int nResult = m_pUX->OnExecutePackageBegin(wzId, fExecute);
        return nResult;
    }


    STDMETHODIMP_(int) OnExecuteError(
        __in IBurnPackage* pPackage,
        __in DWORD dwCode,
        __in_z LPCWSTR wzMessage,
        __in UINT uiFlags
        )
    {
        if (!m_fInitialized)
        {
            return IDNOACTION;
        }

        int nResult = m_pUX->OnError(dwCode, wzMessage, uiFlags);
        return nResult;
    }


    STDMETHODIMP_(int) OnExecuteProgress(
        __in IBurnPackage* pPackage,
        __in DWORD dwPercentage
        )
    {
        if (!m_fInitialized)
        {
            return IDNOACTION; // view silently continues without UX.
        }

        DWORD dwOverallProgress = 0;

        // If we recieved the count of packages that will be executing,
        // calculate the "overall percentage".
        if (m_cExecutingPackages)
        {
            dwOverallProgress = (m_cExecutedPackages * 100 / m_cExecutingPackages) + (dwPercentage / m_cExecutingPackages);
        }

        int nResult = m_pUX->OnProgress(dwPercentage, dwOverallProgress);
        return nResult;
    }


    STDMETHODIMP_(int) OnExecuteMsiMessage(
        __in IBurnMsiPackage* pPackage,
        __in INSTALLMESSAGE mt,
        __in UINT uiFlags,
        __in_z LPCWSTR wzMessage,
        __in_opt MSIHANDLE hRecord
        )
    {
        if (!m_fInitialized)
        {
            return IDNOACTION;
        }

        int nResult = IDNOACTION;
        return nResult;
    }


    STDMETHODIMP_(void) OnExecutePackageComplete(
        __in IBurnPackage* pPackage,
        __in HRESULT hrExitCode
        )
    {
        if (!m_fInitialized)
        {
            return;
        }

        LPCWSTR wzId = pPackage->Id();

        if (SUCCEEDED(hrExitCode) || m_fExecutingRollback)
        {
            m_cExecutedPackages += m_fExecutingRollback ? - 1 : 1;
        }
        else
        {
            m_fExecutingRollback = TRUE;
        }

        m_pUX->OnExecutePackageComplete(wzId, hrExitCode);
        return;
    }


    STDMETHODIMP_(void) OnExecuteComplete(
        __in HRESULT hrStatus
        )
    {
        if (!m_fInitialized)
        {
            return;
        }

        m_pUX->OnExecuteComplete(hrStatus);

        m_cExecutingPackages = 0;
        m_cExecutedPackages = 0;
        m_fExecutingRollback = FALSE;
    }


    STDMETHODIMP_(BOOL) OnRestartRequired()
    {
        if (!m_fInitialized)
        {
            return FALSE;
        }

        BOOL fRestartAllowed = m_pUX->OnRestartRequired();
        return fRestartAllowed;
    }


    STDMETHODIMP GetControlText(
        __in DWORD dwControlID,
        __out LPWSTR* psczText
        )
    {
        HRESULT hr = S_OK;

        if (!m_fInitialized)
        {
            ExitFunction1(hr = S_FALSE);
        }

        hr = m_pUX->GetControlText(dwControlID, psczText);

    LExit:
        return hr;
    }


    STDMETHODIMP WaitIfPaused()
    {
        return E_NOTIMPL;
    }

public: // internals


public:
    //
    // Constructor - intitialize member variables.
    //
    CBurnView(
        __in int nCmdShow
        )
    {
        m_fInitialized = FALSE;

        m_pModel = NULL;
        m_pController = NULL;
        m_pUX = NULL;

        m_nCmdShow = nCmdShow;
        m_cCachedPackages = 0;
        m_cCachingPackages = 0;
        m_fExecutingRollback = FALSE;
        m_cExecutedPackages = 0;
        m_cExecutingPackages = 0;
    }

    //
    // Destructor - release member variables.
    //
    ~CBurnView()
    {
        AssertSz(!m_fInitialized, "Should not still be initialized in CBurnView::~CBurnView().");

        AssertSz(!m_pUX, "UX should have been freed via CBurnView::Uninitialize().");
        AssertSz(!m_pController, "Controller should have been freed via CBurnView::Uninitialize().");
        AssertSz(!m_pModel, "Model should have been freed via CBurnView::Uninitialize().");
    }

private:
    BOOL m_fInitialized;

    IBurnModel* m_pModel;
    IBurnController* m_pController;
    IBurnUserExperience* m_pUX;

    int m_nCmdShow;
    DWORD m_cCachedPackages;
    DWORD m_cCachingPackages;
    BOOL m_fExecutingRollback;
    DWORD m_cExecutedPackages;
    DWORD m_cExecutingPackages;
};


//
// BurnCreateView - creates a new IBurnView object.
//
HRESULT BurnCreateView(
    __in int nCmdShow,
    __out IBurnView **ppView
    )
{
    HRESULT hr = S_OK;

    CBurnView* pView = new CBurnView(nCmdShow);
    ExitOnNull(pView, hr, E_OUTOFMEMORY, "Failed to create new view object.");

    *ppView = pView;
    pView = NULL;

LExit:
    ReleaseObject(pView);
    return hr;
}
