//-------------------------------------------------------------------------------------------------
// <copyright file="core.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//
// <summary>
//    Setup chainer/bootstrapper core for WiX toolset.
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

#define HRESULT_FROM_RPC(x)      ((HRESULT) ((x) | FACILITY_RPC))


// globals


// prototypes

static HRESULT OpenWorkerMessagePipe(
    __in LPCWSTR wzPipeName,
    __out HANDLE* phWorkerMessagePipe
    );
static void FreeProperties(
    __in DWORD cProperties,
    __in SETUP_PROPERTY* rgProperties
    );
static void FreeRegistrySearches(
    __in DWORD cRegistrySearches,
    __in SETUP_REGISTRY_SEARCH* rgRegistrySearches
    );
static void FreePrerequisiteChecks(
    __in DWORD cPrerequisiteChecks,
    __in SETUP_PREREQUISITE_CHECK* rgPrerequisiteChecks
    );
static void FreeDiskCosts(
    __in DWORD cDiskCosts,
    __in SETUP_DISK_COST* rgDiskCosts
    );
static void FreeSourcePaths(
    __in DWORD cSourcePaths,
    __in SETUP_SOURCE_PATH* rgSourcePaths
    );
static HRESULT SetKnownFolderProperty(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in int nFolder,
    __in LPCWSTR wzProperty
    );


// function definitions

HRESULT WIXAPI InitializeSetupEngine(
    __in SETUP_COMMAND_LINE* pSetupCommandLine,
    __out SETUP_ENGINE_INFORMATION** ppEngineInfo
    )
{
    HRESULT hr = S_OK;
    RPC_STATUS rs = RPC_S_OK;
    SETUP_ENGINE_INFORMATION* pEngineInfo = NULL;
    SETUP_WORKER_INFORMATION SetupWorkerInformation = { 0 };
    UUID uuid = { 0 };
    RPC_WSTR pwzUuid = NULL;
    HANDLE hInitEvent = NULL;
    HANDLE rgHandles[2];
    DLLGETVERSIONPROC pfnMsiDllGetVersion = NULL;

    pEngineInfo = static_cast<SETUP_ENGINE_INFORMATION*>(MemAlloc(sizeof(SETUP_ENGINE_INFORMATION), TRUE));
    ExitOnNull(pEngineInfo, hr, E_OUTOFMEMORY, "Failed to allocate Setup UX information.");

    pEngineInfo->pSetupCommandLine = pSetupCommandLine;

    // load msi.dll
    pEngineInfo->hMsi = ::LoadLibraryW(L"msi.dll"); // TODO: user GetModuleHandle() instead?
    ExitOnNullWithLastError(pEngineInfo->hMsi, hr, "Failed to load msi.dll.");

    pfnMsiDllGetVersion = (DLLGETVERSIONPROC)::GetProcAddress(pEngineInfo->hMsi, "DllGetVersion");
    ExitOnNullWithLastError(pfnMsiDllGetVersion, hr, "Failed to find DllGetVersion entry point in msi.dll.");

    pEngineInfo->MsiVersionInfo.cbSize = sizeof(DLLVERSIONINFO);
    hr = pfnMsiDllGetVersion(&pEngineInfo->MsiVersionInfo);
    ExitOnFailure(hr, "Failed to get msi.dll version info.");

    pEngineInfo->dwVersionMsi = (pEngineInfo->MsiVersionInfo.dwMajorVersion * 100) + (pEngineInfo->MsiVersionInfo.dwMinorVersion * 10);

    // get msi 3.1 function entry points
    if (310 <= pEngineInfo->dwVersionMsi)
    {
        // MsiSetExternalUIRecord
        pEngineInfo->pfnMsiSetExternalUIRecord = (PFN_MSISETEXTERNALUIRECORD)::GetProcAddress(pEngineInfo->hMsi, "MsiSetExternalUIRecord");
        ExitOnNullWithLastError(pEngineInfo->pfnMsiSetExternalUIRecord, hr, "Failed to find MsiSetExternalUIRecord entry point in msi.dll.");
    }

    // create log file
    if (pSetupCommandLine->wzLog)
    {
        hr = StrAllocString(&pEngineInfo->pwzLogFile, pSetupCommandLine->wzLog, 0);
    }
    else
    {
        hr = PathCreateTempFile(NULL, L"setup.wix.%00000x.log", SHRT_MAX, FILE_ATTRIBUTE_NORMAL, &pEngineInfo->pwzLogFile, NULL);
    }
    if (SUCCEEDED(hr))
    {
        hr = LogInitialize(::GetModuleHandle(NULL), pEngineInfo->pwzLogFile, NULL, FALSE, TRUE);
    }
    if (FAILED(hr))
    {
        TraceError1(hr, "Failed to initialize log. Path = '%S'", pSetupCommandLine->wzLog);
    }
    else
    {
        pEngineInfo->fLogInitialized = TRUE;
    }

    // create worker message pipe name
    rs = ::UuidCreate(&uuid);
    hr = HRESULT_FROM_RPC(rs);
    ExitOnFailure(hr, "Failed to get uuid.");

    rs = ::UuidToStringW(&uuid, &pwzUuid);
    hr = HRESULT_FROM_RPC(rs);
    ExitOnFailure(hr, "Failed to convert uuid to string.");

    hr = StrAllocFormatted(&pEngineInfo->pwzPipe, L"\\\\.\\pipe\\%s", pwzUuid);
    ExitOnFailure(hr, "Failed to allocate pipe name.");

    // start worker thread
    hInitEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);
    ExitOnNullWithLastError(hInitEvent, hr, "Failed to create worker init event.");

    SetupWorkerInformation.pEngineInfo = pEngineInfo;
    SetupWorkerInformation.hInitEvent = hInitEvent;
    pEngineInfo->hThread = ::CreateThread(NULL, 0, WorkerThread, static_cast<LPVOID>(&SetupWorkerInformation), 0, NULL);
    ExitOnNullWithLastError(pEngineInfo->hThread, hr, "Failed to create worker thread.");

    rgHandles[0] = hInitEvent;
    rgHandles[1] = pEngineInfo->hThread;
    switch (::WaitForMultipleObjects(2, rgHandles, FALSE, INFINITE))
    {
    case WAIT_OBJECT_0:
        break; // success
    case WAIT_OBJECT_0 + 1:
        //::GetExitCodeThread(pEngineInfo->hThread, &hr);
        ExitOnFailure(hr = E_UNEXPECTED, "Worker thread terminated abnormaly.");
        break;
    case WAIT_FAILED:
        ExitWithLastError(hr, "Failed wait for thread init completion.");
        // here it is actually not safe to leave scope as the worker thread
        // might still attempt to read the SetupWorkerInformation struct
    }

    // open message pipe
    pEngineInfo->hPipe = INVALID_HANDLE_VALUE;

    hr = OpenWorkerMessagePipe(pEngineInfo->pwzPipe, &pEngineInfo->hPipe);
    ExitOnFailure(hr, "Failed to open pipe for client.");

    // return values
    *ppEngineInfo = pEngineInfo;
    pEngineInfo = NULL;

LExit:
    if (hInitEvent)
    {
        ::CloseHandle(hInitEvent);
    }
    if (pwzUuid)
    {
        ::RpcStringFreeW(&pwzUuid);
    }
    if (pEngineInfo)
    {
        UninitializeSetupEngine(pEngineInfo);
    }

    return hr;
}

void WIXAPI UninitializeSetupEngine(
    __in_opt SETUP_ENGINE_INFORMATION* pEngineInfo
    )
{
    if (pEngineInfo)
    {
        // terminate worker thread
        PostWorkerMessage(pEngineInfo, SWM_QUIT, NULL, 0);

        if (pEngineInfo->hThread)
        {
            ::WaitForSingleObject(pEngineInfo->hThread, INFINITE);
            ::CloseHandle(pEngineInfo->hThread);
        }

        // close pipe
        if (INVALID_HANDLE_VALUE != pEngineInfo->hPipe)
        {
            ::CloseHandle(pEngineInfo->hPipe);
        }

        ReleaseStr(pEngineInfo->pwzPipe);

        // setup packages
        FreeSetupPackages(pEngineInfo->pSetupPackages);

        // ux resources
        FreeUXResources(pEngineInfo->pUXResources);
        // ignore failure to delete the UX directory; nothing to be done about it anyway
        DirEnsureDelete(pEngineInfo->pwzUxDirectory, TRUE, TRUE);

        // source paths
        if (pEngineInfo->rgSourcePaths)
        {
            FreeSourcePaths(pEngineInfo->cSourcePaths, pEngineInfo->rgSourcePaths);
        }

        // log
        if (pEngineInfo->fLogInitialized)
        {
            LogUninitialize(TRUE);
        }

        ReleaseStr(pEngineInfo->pwzUxDirectory);
        ReleaseStr(pEngineInfo->pwzSetupUxDll);
        ReleaseStr(pEngineInfo->pwzLogFile);

        // msi.dll
        if (pEngineInfo->hMsi)
        {
            ::FreeLibrary(pEngineInfo->hMsi);
        }

        ReleaseStr(pEngineInfo->pwzCachePath);

        // release memory
        MemFree(pEngineInfo);
    }
}


#ifdef DEBUG
void WIXAPI SetEngineDebugHooks(
    __in SETUP_DEBUG_HOOKS* pDebugHooks
    )
{
    ::CopyMemory(&EngineDebugHooks, pDebugHooks, sizeof(SETUP_DEBUG_HOOKS));
}
#endif


HRESULT WIXAPI SetEngineMessageHandler(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in PFN_SETUP_MESSAGE_HANDLER pfnMessageHandler,
    __in LPVOID pvContext
    )
{
    pEngineInfo->pfnMessageHandler = pfnMessageHandler;
    pEngineInfo->pvHandlerContext = pvContext;
    return S_OK;
}


HRESULT WIXAPI SetMsiSourceResolutionPrompt(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in BOOL fEnableSourceResPrompt
    )
{
    pEngineInfo->fEnableSourceResPrompt = fEnableSourceResPrompt;
    return S_OK;
}


HRESULT WIXAPI AddSourcePath(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in LPCWSTR wzSourcePath
    )
{
    Assert(wzSourcePath);

    HRESULT hr = S_OK;
    DWORD cSourcePaths = 0;
    SETUP_SOURCE_PATH* pSourcePaths = NULL;
    SETUP_SOURCE_PATH* pSourcePath = NULL;
    int cch = 0;

    // create new source path array
    cSourcePaths = pEngineInfo->cSourcePaths + 1;
    pSourcePaths = (SETUP_SOURCE_PATH*)MemAlloc(cSourcePaths * sizeof(SETUP_SOURCE_PATH), TRUE);
    ExitOnNull(pSourcePaths, hr, E_OUTOFMEMORY, "Failed to allocate new source path array.");

    // init new source path
    pSourcePath = &pSourcePaths[pEngineInfo->cSourcePaths];

    cch = lstrlenW(wzSourcePath);
    if (0 == cch || L'\\' == wzSourcePath[cch - 1])
    {
        // the path is terminated by a '\' character
        hr = StrAllocString(&pSourcePath->wzPath, wzSourcePath, cch);
        ExitOnFailure(hr, "Failed to copy source path string.");
    }
    else
    {
        // add a terminating '\' character
        hr = StrAllocFormatted(&pSourcePath->wzPath, L"%s\\", wzSourcePath);
        ExitOnFailure(hr, "Failed to copy source path string and add terminationg '\\'.");
    }

    // copy old values
    if (0 != pEngineInfo->cSourcePaths)
    {
        ::CopyMemory(pSourcePaths, pEngineInfo->rgSourcePaths, pEngineInfo->cSourcePaths * sizeof(SETUP_SOURCE_PATH));
    }

    // update array
    ReleaseMem(pEngineInfo->rgSourcePaths);
    pEngineInfo->rgSourcePaths = pSourcePaths;
    pSourcePaths = NULL;
    pEngineInfo->cSourcePaths = cSourcePaths;

    hr = S_OK;

LExit:
    if (pSourcePaths)
    {
        FreeSourcePaths(cSourcePaths, pSourcePaths);
    }
    return hr;
}


HRESULT WIXAPI InitializeSetupUX(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SETUP_COMMAND_LINE* pSetupCommandLine,
    __out SETUP_UX_INFORMATION** ppSetupUXInformation
    )
{
    HRESULT hr = S_OK;
    SETUP_UX_INFORMATION* pSetupUXInformation = NULL;

    pSetupUXInformation = static_cast<SETUP_UX_INFORMATION*>(MemAlloc(sizeof(SETUP_UX_INFORMATION), TRUE));
    ExitOnNull(pSetupUXInformation, hr, E_OUTOFMEMORY, "Failed to allocate Setup UX information.");

    pSetupUXInformation->hModule = ::LoadLibraryW(pEngineInfo->pwzSetupUxDll ? pEngineInfo->pwzSetupUxDll : L"stdux.dll");
    ExitOnNullWithLastError(pSetupUXInformation->hModule, hr, "Failed to load Setup UX DLL.");

    pSetupUXInformation->pfnInitialize = reinterpret_cast<PFN_SETUPUX_INITIALIZE>(::GetProcAddress(pSetupUXInformation->hModule, "SetupUXInitialize"));
    ExitOnNullWithLastError(pSetupUXInformation->pfnInitialize, hr, "Failed to find SetupUXInitialize entry point in Setup UX DLL.");

    pSetupUXInformation->pfnMessageLoop = reinterpret_cast<PFN_SETUPUX_MESSAGE_LOOP>(::GetProcAddress(pSetupUXInformation->hModule, "SetupUXMessageLoop"));
    ExitOnNullWithLastError(pSetupUXInformation->pfnMessageLoop, hr, "Failed to find SetupUXMessageLoop entry point in Setup UX DLL.");

    hr = (*pSetupUXInformation->pfnInitialize)(pSetupCommandLine, static_cast<HANDLE>(pSetupUXInformation), &pSetupUXInformation->hWnd);
    ExitOnFailure(hr, "Failed to initialize Setup UX DLL.");

    *ppSetupUXInformation = pSetupUXInformation;
    pSetupUXInformation = NULL;

LExit:
    if (pSetupUXInformation)
    {
        UninitializeSetupUX(pSetupUXInformation);
    }

    return hr;
}


void WIXAPI UninitializeSetupUX(
    __in_opt SETUP_UX_INFORMATION* pSetupUXInformation
    )
{
    if (pSetupUXInformation)
    {
        if (pSetupUXInformation->hModule)
        {
            ::FreeLibrary(pSetupUXInformation->hModule);
        }

        MemFree(pSetupUXInformation);
    }
}


HRESULT WIXAPI SetupInitializeDefaultProperties(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo
    )
{
    HRESULT hr = S_OK;
    DWORD cch = 0;
    LPWSTR pwz = NULL;
    OSVERSIONINFOEXW ovix = { 0 };


    //
    // set OS related properties
    //

    // set version properties
    ovix.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEXW);
    if (!::GetVersionExW((LPOSVERSIONINFOW)&ovix))
    {
        ExitWithLastError(hr, "Failed to get OS info.");
    }

    hr = SetIntProperty(pEngineInfo, L"WIX_VersionMsi", pEngineInfo->MsiVersionInfo.dwMajorVersion * 100 + pEngineInfo->MsiVersionInfo.dwMinorVersion);
    ExitOnFailure(hr, "Failed to set WIX_VersionMsi property.");

    hr = SetIntProperty(pEngineInfo, L"WIX_VersionNT", ovix.dwMajorVersion * 100 + ovix.dwMinorVersion);
    ExitOnFailure(hr, "Failed to set WIX_VersionNT property.");

    if (ovix.wServicePackMajor)
    {
        hr = SetIntProperty(pEngineInfo, L"WIX_ServicePackLevel", ovix.wServicePackMajor);
        ExitOnFailure(hr, "Failed to set WIX_ServicePackLevel property.");

        hr = SetIntProperty(pEngineInfo, L"WIX_ServicePackLevelMinor", ovix.wServicePackMinor);
        ExitOnFailure(hr, "Failed to set WIX_ServicePackLevelMinor property.");
    }

#if !defined(_WIN64)
    // TODO: tidy up
    BOOL fIsWow64 = FALSE;
    typedef BOOL (WINAPI *LPFN_ISWOW64PROCESS) (HANDLE, PBOOL);
    LPFN_ISWOW64PROCESS fnIsWow64Process = (LPFN_ISWOW64PROCESS)::GetProcAddress(::GetModuleHandleW(L"kernel32"), "IsWow64Process");
    if (NULL != fnIsWow64Process)
    {
        if (!fnIsWow64Process(::GetCurrentProcess(), &fIsWow64))
        {
            ExitWithLastError(hr, "Failed to check WOW64 process.");
        }
    }
    if (fIsWow64)
#endif
    {
        hr = SetIntProperty(pEngineInfo, L"WIX_VersionNT64", ovix.dwMajorVersion * 100 + ovix.dwMinorVersion);
        ExitOnFailure(hr, "Failed to set WIX_VersionNT property.");
    }

    hr = SetIntProperty(pEngineInfo, L"WIX_WindowsBuild", ovix.dwBuildNumber);
    ExitOnFailure(hr, "Failed to set WIX_WindowsBuild property.");

    hr = SetIntProperty(pEngineInfo, L"WIX_ServicePackLevel", ovix.wServicePackMajor);
    ExitOnFailure(hr, "Failed to set WIX_ServicePackLevel property.");

    hr = SetIntProperty(pEngineInfo, L"WIX_ServicePackLevelMinor", ovix.wServicePackMinor);
    ExitOnFailure(hr, "Failed to set WIX_ServicePackLevelMinor property.");

    // set computer name property
    cch = 0;
    if (!::GetComputerNameW(NULL, &cch) && ERROR_BUFFER_OVERFLOW != ::GetLastError())
    {
        ExitWithLastError(hr, "Failed to get computer name length.");
    }

    hr = StrAlloc(&pwz, ++cch);
    ExitOnFailure(hr, "Failed to allocate computer name string.");

    if (!::GetComputerNameW(pwz, &cch))
    {
        ExitWithLastError(hr, "Failed to get computer name.");
    }

    hr = SetStrPropertyEscaped(pEngineInfo, L"WIX_ComputerName", pwz);
    ExitOnFailure(hr, "Failed to set WIX_ComputerName property.");


    //
    // set known folders
    //

    //hr = SetKnownFolderProperty(pEngineInfo, CSIDL_ADMINTOOLS, L"WIX_AdminToolsFolder");
    //ExitOnFailure(hr, "Failed to set WIX_AdminToolsFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_APPDATA, L"WIX_AppDataFolder");
    ExitOnFailure(hr, "Failed to set WIX_AppDataFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_COMMON_APPDATA, L"WIX_CommonAppDataFolder");
    ExitOnFailure(hr, "Failed to set WIX_CommonAppDataFolder property.");

    // WIX_CommonFiles64Folder

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_PROGRAM_FILES_COMMON, L"WIX_CommonFilesFolder");
    ExitOnFailure(hr, "Failed to set WIX_CommonFilesFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_DESKTOP, L"WIX_DesktopFolder");
    ExitOnFailure(hr, "Failed to set WIX_DesktopFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_FAVORITES, L"WIX_FavoritesFolder");
    ExitOnFailure(hr, "Failed to set WIX_FavoritesFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_FONTS, L"WIX_FontsFolder");
    ExitOnFailure(hr, "Failed to set WIX_FontsFolder property.");

    // WIX_LocalAppDataFolder

    //if (6 <= pEngineInfo->MsiVersionInfo.dwMajorVersion)
    //{
    //    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_MYPICTURES, L"WIX_MyPicturesFolder");
    //    ExitOnFailure(hr, "Failed to set WIX_MyPicturesFolder property.");
    //}

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_PERSONAL, L"WIX_PersonalFolder");
    ExitOnFailure(hr, "Failed to set WIX_MyPicturesFolder property.");

    // WIX_ProgramFiles64Folder

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_PERSONAL, L"WIX_PersonalFolder");
    ExitOnFailure(hr, "Failed to set WIX_MyPicturesFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_PROGRAM_FILES, L"WIX_ProgramFilesFolder");
    ExitOnFailure(hr, "Failed to set WIX_ProgramFilesFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_PROGRAMS, L"WIX_ProgramMenuFolder");
    ExitOnFailure(hr, "Failed to set WIX_ProgramMenuFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_SENDTO, L"WIX_SendToFolder");
    ExitOnFailure(hr, "Failed to set WIX_SendToFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_STARTMENU, L"WIX_StartMenuFolder");
    ExitOnFailure(hr, "Failed to set WIX_StartMenuFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_STARTUP, L"WIX_StartupFolder");
    ExitOnFailure(hr, "Failed to set WIX_StartupFolder property.");

    // WIX_System16Folder
    // WIX_System64Folder

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_SYSTEM, L"WIX_SystemFolder");
    ExitOnFailure(hr, "Failed to set WIX_SystemFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_TEMPLATES, L"WIX_TemplateFolder");
    ExitOnFailure(hr, "Failed to set WIX_TemplateFolder property.");

    hr = SetKnownFolderProperty(pEngineInfo, CSIDL_WINDOWS, L"WIX_WindowsFolder");
    ExitOnFailure(hr, "Failed to set WIX_WindowsFolder property.");

    // WIX_WindowsVolume

    // get temp folder
    cch = ::GetTempPathW(0, NULL);
    if (0 == cch)
    {
        ExitWithLastError(hr, "Failed to get temp path length.");
    }
    hr = StrAlloc(&pwz, cch);
    ExitOnFailure(hr, "Failed to allocate temp path string.");

    if (::GetTempPathW(cch, pwz) != cch - 1)
    {
        ExitWithLastError(hr, "Failed to get temp path.");
    }
    pwz[cch - 2] = L'\0'; // remove trailing '\'

    hr = SetStrPropertyEscaped(pEngineInfo, L"WIX_TempFolder", pwz);
    ExitOnFailure(hr, "Failed to set WIX_TempFolder property.");

LExit:
    ReleaseStr(pwz);
    return hr;
}


HRESULT WIXAPI LoadManifest(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo
    )
{
    HRESULT hr = S_OK;
    DWORD dwPayloadsOffset = 0;

    ExitOnNull(pEngineInfo, hr, E_INVALIDARG, "Invalid null argument: pEngineInfo");
    ExitOnNull(pEngineInfo->pSetupCommandLine, hr, E_INVALIDARG, "Invalid null argument: pSetupCommandLine");

    hr = ManifestInitialize();
    ExitOnFailure(hr, "Failed to initialize manifest system.");

    // if there's a manifest specified in the command line, use it in preference to one that's embedded
    if (pEngineInfo->pSetupCommandLine->wzManifest)
    {
        hr = ManifestParseFromFile(pEngineInfo->pSetupCommandLine->wzManifest, &pEngineInfo->pSetupPackages, &pEngineInfo->pUXResources);
        ExitOnFailure1(hr, "Failed to parse external manifest: %S", pEngineInfo->pSetupCommandLine->wzManifest);
    }
    else
    {
        //hr = ManifestParseFromEmbedded(&pEngineInfo->pSetupPackages, &pEngineInfo->pUXResources, &dwPayloadsOffset);
        ExitOnFailure(hr, "Failed to parse embedded manifest.");

        hr = ExtractUXPayloads(dwPayloadsOffset, pEngineInfo->pUXResources, &pEngineInfo->pwzUxDirectory);
        ExitOnFailure(hr, "Failed to extract embedded payloads.");

        hr = StrAllocString(&pEngineInfo->pwzSetupUxDll, pEngineInfo->pUXResources->rgUXResources[0].wzPath, 0);
        ExitOnFailure(hr, "Failed to update setup UX in engine.");
    }

LExit:
    ManifestUninitialize();

    return hr;
}


void WIXAPI FreeSetupPackages(
    __in_opt SETUP_PACKAGES* pSetupPackages
    )
{
    if (pSetupPackages)
    {
        FreeProperties(pSetupPackages->cProperties, pSetupPackages->rgProperties);
        FreeRegistrySearches(pSetupPackages->cRegistrySearches, pSetupPackages->rgRegistrySearches);

        for (DWORD i = 0; i < pSetupPackages->cPackages; ++i)
        {
            SETUP_PACKAGE* pSetupPackage = pSetupPackages->rgPackages + i;
            ReleaseStr(pSetupPackage->wzId);
            ReleaseStr(pSetupPackage->wzSource);
            ReleaseStr(pSetupPackage->wzExecutePath);
            //ReleaseStr(pSetupPackage->wzDownloadPath);
            ReleaseStr(pSetupPackage->pwzCondition);
            FreeDiskCosts(pSetupPackage->cDiskCosts, pSetupPackage->rgDiskCosts);
            switch (pSetupPackage->PackageType)
            {
            case SETUP_PACKAGE_TYPE_MSI:
                for (DWORD fi = 0; fi < pSetupPackage->Msi.cFeatures; fi++)
                {
                    SETUP_PACKAGE_MSI_FEATURE* pFeature = &pSetupPackage->Msi.rgFeatures[fi];
                    ReleaseStr(pFeature->wzId);
                    FreeDiskCosts(pFeature->cDiskCosts, pFeature->rgDiskCosts);
                }
                ReleaseMem(pSetupPackage->Msi.rgFeatures);

                for (DWORD pi = 0; pi < pSetupPackage->Msi.cProperties; pi++)
                {
                    SETUP_PACKAGE_MSI_PROPERTY* pProperty = &pSetupPackage->Msi.rgProperties[pi];
                    ReleaseStr(pProperty->wzId);
                    ReleaseStr(pProperty->wzRequestedValue);
                    ReleaseStr(pProperty->wzExpectedValue);
                }
                ReleaseMem(pSetupPackage->Msi.rgProperties);
                break;
            case SETUP_PACKAGE_TYPE_EXE:
                ReleaseStr(pSetupPackage->Exe.wzArguments);
                ReleaseStr(pSetupPackage->Exe.wzRepairArguments);
                ReleaseStr(pSetupPackage->Exe.wzUninstallArguments);

                FreePrerequisiteChecks(pSetupPackage->cPrereqChecks, pSetupPackage->rgPrereqChecks);
                ReleaseStr(pSetupPackage->Exe.wzInstallCheck);
                break;
            }
        }

        MemFree(pSetupPackages);
    }
}


HRESULT WIXAPI EnterSetupUXMessageLoop(
    __in SETUP_UX_INFORMATION* pSetupUXInformation
    )
{
    HRESULT hr = (*pSetupUXInformation->pfnMessageLoop)(reinterpret_cast<HANDLE>(pSetupUXInformation));
    return hr;
}


HRESULT WIXAPI PostWorkerMessage(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in DWORD dwMessage,
    __in_opt LPVOID pvData,
    __in DWORD cbData
    )
{
    HRESULT hr = S_OK;
    SETUP_COMM_MESSAGE msg = { 0 };
    DWORD cbWritten = 0;

    msg.dwMessage = dwMessage;
    msg.cbData = cbData;
    msg.pvData = pvData;

    if (!::WriteFile(pEngineInfo->hPipe, &msg, sizeof(DWORD) * 2, &cbWritten, NULL))
    {
        ExitWithLastError(hr, "Failed to write message.");
    }

    if (!::WriteFile(pEngineInfo->hPipe, msg.pvData, cbData, &cbWritten, NULL))
    {
        ExitWithLastError(hr, "Failed to write message data.");
    }

LExit:
    return hr;
}


HRESULT WIXAPI GetLogPath(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __out LPWSTR* ppwzLogPath
    )
{
    return StrAllocString(ppwzLogPath, pEngineInfo->pwzLogFile, 0);
}


HRESULT WIXAPI AddRegistrySearch(
    __in SETUP_PACKAGES* pSetupPackages,
    __in LPCWSTR wzId,
    __in REGISTRY_SEARCH_TYPE eType,
    __in HKEY hRoot,
    __in LPCWSTR wzKey,
    __in LPCWSTR wzValue,
    __in LPCWSTR wzProperty
    )
{
    HRESULT hr = S_OK;
    DWORD cRegistrySearches = 0;
    SETUP_REGISTRY_SEARCH* pRegistrySearches = NULL;
    SETUP_REGISTRY_SEARCH* pRegistrySearch = NULL;

    // create new property array
    cRegistrySearches = pSetupPackages->cRegistrySearches + 1;
    pRegistrySearches = (SETUP_REGISTRY_SEARCH*)MemAlloc(cRegistrySearches * sizeof(SETUP_REGISTRY_SEARCH), TRUE);
    ExitOnNull(pRegistrySearches, hr, E_OUTOFMEMORY, "Failed to allocate new registry search array.");

    // init new registry search
    pRegistrySearch = &pRegistrySearches[pSetupPackages->cRegistrySearches];

    hr = StrAllocString(&pRegistrySearch->wzId, wzId, 0);
    ExitOnFailure(hr, "Failed to copy property id string.");

    pRegistrySearch->Type = eType;
    pRegistrySearch->hRoot = hRoot;

    hr = StrAllocString(&pRegistrySearch->wzKey, wzKey, 0);
    ExitOnFailure(hr, "Failed to copy property id string.");

    if (wzValue)
    {
        hr = StrAllocString(&pRegistrySearch->wzValue, wzValue, 0);
        ExitOnFailure(hr, "Failed to copy property id string.");
    }

    hr = StrAllocString(&pRegistrySearch->wzProperty, wzProperty, 0);
    ExitOnFailure(hr, "Failed to copy property id string.");

    // copy old values
    if (0 != pSetupPackages->cRegistrySearches)
    {
        ::CopyMemory(pRegistrySearches, pSetupPackages->rgRegistrySearches, pSetupPackages->cRegistrySearches * sizeof(SETUP_REGISTRY_SEARCH));
    }

    // update array
    ReleaseMem(pSetupPackages->rgRegistrySearches);
    pSetupPackages->rgRegistrySearches = pRegistrySearches;
    pRegistrySearches = NULL;
    pSetupPackages->cRegistrySearches = cRegistrySearches;

    hr = S_OK;

LExit:
    if (pRegistrySearches)
    {
        FreeRegistrySearches(cRegistrySearches, pRegistrySearches);
    }
    return hr;
}


HRESULT WIXAPI SetPackageProperty(
    __in SETUP_PACKAGE* pPackage,
    __in LPCWSTR wzProperty,
    __in_opt LPCWSTR wzRequestedValue,
    __in_opt LPCWSTR wzExpectedValue
    )
{
    HRESULT hr = S_OK;
    SETUP_PACKAGE_MSI_PROPERTY* pProperty = NULL;

    // find existing property
    for (DWORD i = 0; i < pPackage->Msi.cProperties; ++i)
    {
        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, pPackage->Msi.rgProperties[i].wzId, -1, wzProperty, -1))
        {
            pProperty = &pPackage->Msi.rgProperties[i];
            break;
        }
    }

    // add property if not found
    if (!pProperty)
    {
        if (pPackage->Msi.rgProperties)
        {
            LPVOID pv = MemReAlloc(pPackage->Msi.rgProperties, sizeof(SETUP_PACKAGE_MSI_PROPERTY) * (pPackage->Msi.cProperties + 1), TRUE);
            ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to reallocate property array.");
            pPackage->Msi.rgProperties = (SETUP_PACKAGE_MSI_PROPERTY*)pv;
        }
        else
        {
            pPackage->Msi.rgProperties = (SETUP_PACKAGE_MSI_PROPERTY*)MemAlloc(sizeof(SETUP_PACKAGE_MSI_PROPERTY) * (pPackage->Msi.cProperties + 1), TRUE);
            ExitOnNull(pPackage->Msi.rgProperties, hr, E_OUTOFMEMORY, "Failed to allocate property array.");
        }

        pProperty = &pPackage->Msi.rgProperties[pPackage->Msi.cProperties];

        hr = StrAllocString(&pProperty->wzId, wzProperty, 0);
        ExitOnFailure(hr, "Failed to get allocate property id string.");

        pPackage->Msi.cProperties++;
    }

    // set requested value
    if (wzRequestedValue)
    {
        hr = StrAllocString(&pProperty->wzRequestedValue, wzRequestedValue, 0);
        ExitOnFailure(hr, "Failed to copy requested value string.");
    }
    else
    {
        ReleaseNullStr(pProperty->wzRequestedValue);
    }

    // set expected value
    if (wzExpectedValue)
    {
        hr = StrAllocString(&pProperty->wzExpectedValue, wzExpectedValue, 0);
        ExitOnFailure(hr, "Failed to copy expected value string.");
    }
    else
    {
        ReleaseNullStr(pProperty->wzExpectedValue);
    }

LExit:
    return hr;
}


void WIXAPI FreeSetupCommandLine(
    __in SETUP_COMMAND_LINE* pSetupCommandLine
    )
{
    ReleaseStr(pSetupCommandLine->wzLog);
    ReleaseStr(pSetupCommandLine->wzManifest);
    ReleaseStr(pSetupCommandLine->wzProperties);
}


// internal functions

static HRESULT OpenWorkerMessagePipe(
    __in LPCWSTR wzPipeName,
    __out HANDLE* phWorkerMessagePipe
    )
{
    HRESULT hr = S_OK;
    HANDLE hPipe = INVALID_HANDLE_VALUE;

    hPipe = ::CreateFileW(wzPipeName, GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hPipe)
    {
        ExitWithLastError1(hr, "Failed to open worker message pipe: %S", wzPipeName)
    }

    *phWorkerMessagePipe = hPipe;
    hPipe = INVALID_HANDLE_VALUE;

LExit:
    if (INVALID_HANDLE_VALUE != hPipe)
    {
        ::CloseHandle(hPipe);
    }

    return hr;
}


static void FreeProperties(
    __in DWORD cProperties,
    __in SETUP_PROPERTY* rgProperties
    )
{
    for (DWORD i = 0; i < cProperties; i++)
    {
        SETUP_PROPERTY* pProperty = &rgProperties[i];

        ReleaseStr(pProperty->wzId);
        ::VariantClear(&pProperty->vtValue);
    }
    ReleaseMem(rgProperties);
}


static void FreeRegistrySearches(
    __in DWORD cRegistrySearches,
    __in SETUP_REGISTRY_SEARCH* rgRegistrySearches
    )
{
    for (DWORD i = 0; i < cRegistrySearches; i++)
    {
        SETUP_REGISTRY_SEARCH* pRegistrySearch = &rgRegistrySearches[i];

        ReleaseStr(pRegistrySearch->wzId);
        ReleaseStr(pRegistrySearch->wzKey);
        ReleaseStr(pRegistrySearch->wzValue);
        ReleaseStr(pRegistrySearch->wzProperty);
    }
    ReleaseMem(rgRegistrySearches);
}


static void FreePrerequisiteChecks(
    __in DWORD cPrerequisiteChecks,
    __in SETUP_PREREQUISITE_CHECK* rgPrerequisiteChecks
    )
{
    for (DWORD i = 0; i < cPrerequisiteChecks; i++)
    {
        SETUP_PREREQUISITE_CHECK* pPrerequisiteChecks = &rgPrerequisiteChecks[i];

        ReleaseStr(pPrerequisiteChecks->wzId);
        ReleaseStr(pPrerequisiteChecks->pwzCondition);
    }
    ReleaseMem(rgPrerequisiteChecks);
}


static void FreeDiskCosts(
    __in DWORD cDiskCosts,
    __in SETUP_DISK_COST* rgDiskCosts
    )
{
    for (DWORD i = 0; i < cDiskCosts; ++i)
    {
        ReleaseStr(rgDiskCosts[i].wzPath);
    }
    ReleaseMem(rgDiskCosts);
}


static void FreeSourcePaths(
    __in DWORD cSourcePaths,
    __in SETUP_SOURCE_PATH* rgSourcePaths
    )
{
    for (DWORD i = 0; i < cSourcePaths; i++)
    {
        ReleaseStr(rgSourcePaths[i].wzPath);
    }
    ReleaseMem(rgSourcePaths);
}


void WIXAPI FreeUXResources(
    __in_opt SETUP_UX_RESOURCES* pUXResources
    )
{
    if (pUXResources)
    {
        for (DWORD i = 0; i < pUXResources->cUXResources; i++)
        {
            ReleaseStr(pUXResources->rgUXResources[i].wzFileName);
            ReleaseStr(pUXResources->rgUXResources[i].wzDestination);
            ReleaseStr(pUXResources->rgUXResources[i].wzPath);
        }
        ReleaseMem(pUXResources->rgUXResources);
    }
}


static HRESULT SetKnownFolderProperty(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in int nFolder,
    __in LPCWSTR wzProperty
    )
{
    HRESULT hr = S_OK;
    LPWSTR pwzPath = NULL;

    pwzPath = (LPWSTR)::CoTaskMemAlloc(MAX_PATH * sizeof(WCHAR));
    ExitOnNull(pwzPath, hr, E_OUTOFMEMORY, "Failed to allocate path buffer.");

    hr = ::SHGetFolderPathW(NULL, nFolder, NULL, 0, pwzPath);
    ExitOnFailure(hr, "Failed to get folder path.");

    hr = SetStrPropertyEscaped(pEngineInfo, wzProperty, pwzPath);
    ExitOnFailure(hr, "Failed to set property.");

LExit:
    if (pwzPath)
    {
        ::CoTaskMemFree(pwzPath);
    }
    return hr;
}
