//-------------------------------------------------------------------------------------------------
// <copyright file="exework.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//
// <summary>
// EXE Package operations for setup stub executable.
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


// function definitions

HRESULT DetectExe(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SETUP_PACKAGE_EXE* pExePackage,
    __in BOOL fRepair,
    __in PACKAGE_STATE requested,
    __in PACKAGE_STATE expected,
    __out PACKAGE_STATE* pCurrent,
    __out ACTION_STATE* pExecuteAction,
    __out ACTION_STATE* pRollbackAction
    )
{
    HRESULT hr = S_OK;
    BOOL fInstalled = FALSE; // a blank install check should always consider the package not installed

    // install check
    if (pExePackage->wzInstallCheck && *pExePackage->wzInstallCheck)
    {
        hr = EvaluateCondition(pEngineInfo, pExePackage->wzInstallCheck, &fInstalled);
        ExitOnFailure1(hr, "Failed to evaluate install check: '%S'", pExePackage->wzInstallCheck);
    }

    if (fInstalled)
    {
        // installed
        *pCurrent = PACKAGE_STATE_PRESENT;
        switch (requested)
        {
        case PACKAGE_STATE_PRESENT:
            *pExecuteAction = (fRepair && pExePackage->fEnableRepair) ? ACTION_STATE_RECACHE : ACTION_STATE_NONE;
            break;
        case PACKAGE_STATE_ABSENT:
            *pExecuteAction = ACTION_STATE_UNINSTALL;
            break;
        default:
            ExitFunction1(hr = E_INVALIDARG);
        }
    }
    else
    {
        // not installed
        *pCurrent = PACKAGE_STATE_ABSENT;
        switch (requested)
        {
        case PACKAGE_STATE_PRESENT:
            *pExecuteAction = ACTION_STATE_INSTALL;
            break;
        case PACKAGE_STATE_ABSENT:
            *pExecuteAction = ACTION_STATE_NONE;
            break;
        default:
            ExitFunction1(hr = E_INVALIDARG);
        }
    }

    // rollback action
    switch (PACKAGE_STATE_UNKNOWN != expected ? expected : *pCurrent)
    {
    case PACKAGE_STATE_PRESENT:
        switch (requested)
        {
        case PACKAGE_STATE_PRESENT:
            *pRollbackAction = ACTION_STATE_NONE;
            break;
        case PACKAGE_STATE_ABSENT:
            *pRollbackAction = ACTION_STATE_INSTALL;
            break;
        default:
            ExitFunction1(hr = E_UNEXPECTED);
        }
        break;
    case PACKAGE_STATE_ABSENT:
        switch (requested)
        {
        case PACKAGE_STATE_PRESENT:
            *pRollbackAction = ACTION_STATE_UNINSTALL;
            break;
        case PACKAGE_STATE_ABSENT:
            *pRollbackAction = ACTION_STATE_NONE;
            break;
        default:
            ExitFunction1(hr = E_UNEXPECTED);
        }
        break;
    default:
        ExitFunction1(hr = E_UNEXPECTED);
    }

LExit:
    return hr;
}


HRESULT ConfigureExe(
    __in LPCWSTR wzExecutablePath,
    __in LPCWSTR wzExecutableArgs,
    __out LPDWORD pdwExitCode
    )
{
    HRESULT hr = S_OK;

    STARTUPINFOW startupInfo = { 0 };
    PROCESS_INFORMATION procInfo = { 0 };
    LPWSTR pwzArguments = NULL;

    if (wzExecutableArgs && *wzExecutableArgs)
    {
        hr = StrAllocFormatted(&pwzArguments, L"\"%s\" %s", wzExecutablePath, wzExecutableArgs);
        ExitOnFailure(hr, "Failed to format arguments string.");
    }
    else
    {
        hr = StrAllocFormatted(&pwzArguments, L"\"%s\"", wzExecutablePath);
        ExitOnFailure(hr, "Failed to format arguments string.");
    }

    startupInfo.cb = sizeof(startupInfo);
    if (!HOOKABLE_CreateProcessW(wzExecutablePath, pwzArguments, NULL, NULL, FALSE, 0, NULL, NULL, &startupInfo, &procInfo))
    {
        ExitOnLastError1(hr, "Failed to CreateProcess on path: %S", wzExecutablePath);
    }

    if (WAIT_OBJECT_0 != ::WaitForSingleObject(procInfo.hProcess, INFINITE))
    {
        ExitOnFailure1(hr = E_UNEXPECTED, "Unexpected terminated process: %S", wzExecutablePath);
    }

    if (!HOOKABLE_GetExitCodeProcess(procInfo.hProcess, pdwExitCode))
    {
        ExitOnLastError1(hr, "Failed to get process exit code: %S", wzExecutablePath);
    }

LExit:
    ReleaseStr(pwzArguments);

    if (procInfo.hProcess)
    {
        ::CloseHandle(procInfo.hProcess);
    }

    if (procInfo.hThread)
    {
        ::CloseHandle(procInfo.hThread);
    }

    return hr;
}
