//-------------------------------------------------------------------------------------------------
// <copyright file="expression.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//
// <summary>
//    Expression parser for Burn.
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


// constants

#define NONTERMINAL (1 << 16)

enum SYMBOL_TYPE
{
    // terminals
    SYMBOL_TYPE_NONE       =  0,
    SYMBOL_TYPE_END        =  1,
    SYMBOL_TYPE_OR         =  2,  // OR
    SYMBOL_TYPE_AND        =  3,  // AND
    SYMBOL_TYPE_NOT        =  4,  // NOT
    SYMBOL_TYPE_LT         =  5,  // <
    SYMBOL_TYPE_GT         =  6,  // >
    SYMBOL_TYPE_LE         =  7,  // <=
    SYMBOL_TYPE_GE         =  8,  // >=
    SYMBOL_TYPE_EQ         =  9,  // =
    SYMBOL_TYPE_NE         = 10,  // <>
    SYMBOL_TYPE_LPAREN     = 11,  // (
    SYMBOL_TYPE_RPAREN     = 12,  // )
    SYMBOL_TYPE_NUMBER     = 13,
    SYMBOL_TYPE_IDENTIFIER = 14,
    SYMBOL_TYPE_LITERAL    = 15,
    // nonterminals
    SYMBOL_TYPE_EXPR    = 16 | NONTERMINAL,
    SYMBOL_TYPE_BTERM   = 17 | NONTERMINAL,
    SYMBOL_TYPE_BFACTOR = 18 | NONTERMINAL,
    SYMBOL_TYPE_TERM    = 19 | NONTERMINAL,
    SYMBOL_TYPE_VALUE   = 20 | NONTERMINAL,
};

enum ACTION_TYPE
{
    ACTION_TYPE_NONE         = 0,
    ACTION_TYPE_SHIFT        = 1 << 0,
    ACTION_TYPE_REDUCE       = 1 << 1,
    ACTION_TYPE_ACCEPT       = 1 << 2,
    ACTION_TYPE_ERROR        = 1 << 3,
    ACTION_TYPE_SHIFT_REDUCE  = ACTION_TYPE_SHIFT | ACTION_TYPE_REDUCE
};


// structs

typedef struct _SYMBOL
{
    SYMBOL_TYPE Type;
    VARIANT vtValue;
    DWORD iPosition;
} SYMBOL;

typedef struct _STATE
{
    SYMBOL* pSymbol;
    int state;
    _STATE* pNext;
} STATE;

typedef HRESULT (*PFN_REDUCE)(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );

typedef struct _ACTION
{
    SYMBOL_TYPE SymbolType;
    ACTION_TYPE ActionType;
    int state;
    PFN_REDUCE pfnReduce;
    DWORD nReduce;
} ACTION;


// internal function declarations

static HRESULT NextToken(
    __in LPCWSTR wz,
    __inout DWORD* pi,
    __out SYMBOL** ppSymbol
    );
static HRESULT PushStack(
    __inout STATE** ppStack,
    __in int state,
    __in SYMBOL* pSymbol
    );
static HRESULT PopStack(
    __inout STATE** ppStack,
    __inout int* piState,
    __in DWORD nLevels
    );
static void FreeState(
    __in STATE* pState,
    __in BOOL fRecursive
    );
static void FreeSymbol(
    __in SYMBOL* pSymbol
    );

static HRESULT Reduce_Expr_BTerm(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );
static HRESULT Reduce_Expr_BTermOrExpr(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );
static HRESULT Reduce_BTerm_BFactor(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );
static HRESULT Reduce_BFactor_Term(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );
static HRESULT Reduce_BFactor_NotTerm(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );
static HRESULT Reduce_BTerm_BTermAndBFactor(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );
static HRESULT Reduce_Term_Value(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );
static HRESULT Reduce_Term_ValueCompValue(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );
static HRESULT Reduce_Term_LParenExprRParen(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );
static HRESULT Reduce_Value(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );
static HRESULT Reduce_Value_Identifier(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    );


//
// parse rules
//
// value                property | literal | integer
// comparison-operator  < | > | <= | >= | = | <>
// term                 value | value comparison-operator value | ( expression )
// boolean-factor       term | NOT term
// boolean-term         boolean-factor | boolean-term AND boolean-factor
// expression           boolean-term | boolean-term OR expression
//

// the maximum number of symbols any rule can reduce
#define MAX_REDUCE 3

//
// parser state table
//
// * Noneterminal symbol types are not allowed to have a shift or reduce operation.
// * Symbols are matched in order and terminal symbols should be listed before noneterminal symbols.
// * Error and reduce actions can not have goto states.
// * This table is both an action table and a goto table combined.
//
static ACTION vActionTable[10][11] =
{
    { // state 0
        { SYMBOL_TYPE_NOT,        ACTION_TYPE_SHIFT,         1, NULL,                         0 },
        { SYMBOL_TYPE_LPAREN,     ACTION_TYPE_SHIFT,         2, NULL,                         0 },
        { SYMBOL_TYPE_NUMBER,     ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_IDENTIFIER, ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value_Identifier,      1 },
        { SYMBOL_TYPE_LITERAL,    ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_EXPR,       ACTION_TYPE_NONE,          3, NULL,                         0 },
        { SYMBOL_TYPE_BTERM,      ACTION_TYPE_NONE,          4, NULL,                         0 },
        { SYMBOL_TYPE_BFACTOR,    ACTION_TYPE_REDUCE,       -1, Reduce_BTerm_BFactor,         1 },
        { SYMBOL_TYPE_TERM,       ACTION_TYPE_REDUCE,       -1, Reduce_BFactor_Term,          1 },
        { SYMBOL_TYPE_VALUE,      ACTION_TYPE_NONE,          5, NULL,                         0 },
        { SYMBOL_TYPE_NONE,       ACTION_TYPE_ERROR,        -1, NULL,                         0 }
    },
    { // state 1
        { SYMBOL_TYPE_LPAREN,     ACTION_TYPE_SHIFT,         2, NULL,                         0 },
        { SYMBOL_TYPE_NUMBER,     ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_IDENTIFIER, ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value_Identifier,      1 },
        { SYMBOL_TYPE_LITERAL,    ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_TERM,       ACTION_TYPE_REDUCE,       -1, Reduce_BFactor_NotTerm,       2 },
        { SYMBOL_TYPE_VALUE,      ACTION_TYPE_NONE,          5, NULL,                         0 },
        { SYMBOL_TYPE_NONE,       ACTION_TYPE_ERROR,        -1, NULL,                         0 }
    },
    { // state 2
        { SYMBOL_TYPE_NOT,        ACTION_TYPE_SHIFT,         1, NULL,                         0 },
        { SYMBOL_TYPE_LPAREN,     ACTION_TYPE_SHIFT,         2, NULL,                         0 },
        { SYMBOL_TYPE_NUMBER,     ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_IDENTIFIER, ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value_Identifier,      1 },
        { SYMBOL_TYPE_LITERAL,    ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_EXPR,       ACTION_TYPE_NONE,          6, NULL,                         0 },
        { SYMBOL_TYPE_BTERM,      ACTION_TYPE_NONE,          4, NULL,                         0 },
        { SYMBOL_TYPE_BFACTOR,    ACTION_TYPE_REDUCE,       -1, Reduce_BTerm_BFactor,         1 },
        { SYMBOL_TYPE_TERM,       ACTION_TYPE_REDUCE,       -1, Reduce_BFactor_Term,          1 },
        { SYMBOL_TYPE_VALUE,      ACTION_TYPE_NONE,          5, NULL,                         0 },
        { SYMBOL_TYPE_NONE,       ACTION_TYPE_ERROR,        -1, NULL,                         0 }
    },
    { // state 3
        { SYMBOL_TYPE_END,        ACTION_TYPE_ACCEPT,        0, NULL,                         0 },
        { SYMBOL_TYPE_NONE,       ACTION_TYPE_ERROR,        -1, NULL,                         0 }
    },
    { // state 4
        { SYMBOL_TYPE_OR,         ACTION_TYPE_SHIFT,         7, NULL,                         0 },
        { SYMBOL_TYPE_AND,        ACTION_TYPE_SHIFT,         8, NULL,                         0 },
        { SYMBOL_TYPE_NONE,       ACTION_TYPE_REDUCE,       -1, Reduce_Expr_BTerm,            1 }
    },
    { // state 5
        { SYMBOL_TYPE_LT,         ACTION_TYPE_SHIFT,         9, NULL,                         0 },
        { SYMBOL_TYPE_GT,         ACTION_TYPE_SHIFT,         9, NULL,                         0 },
        { SYMBOL_TYPE_LE,         ACTION_TYPE_SHIFT,         9, NULL,                         0 },
        { SYMBOL_TYPE_GE,         ACTION_TYPE_SHIFT,         9, NULL,                         0 },
        { SYMBOL_TYPE_EQ,         ACTION_TYPE_SHIFT,         9, NULL,                         0 },
        { SYMBOL_TYPE_NE,         ACTION_TYPE_SHIFT,         9, NULL,                         0 },
        { SYMBOL_TYPE_NONE,       ACTION_TYPE_REDUCE,       -1, Reduce_Term_Value,            1 }
    },
    { // state 6
        { SYMBOL_TYPE_RPAREN,     ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Term_LParenExprRParen, 3 },
        { SYMBOL_TYPE_NONE,       ACTION_TYPE_ERROR,        -1, NULL,                         0 }
    },
    { // state 7
        { SYMBOL_TYPE_NOT,        ACTION_TYPE_SHIFT,         1, NULL,                         0 },
        { SYMBOL_TYPE_LPAREN,     ACTION_TYPE_SHIFT,         2, NULL,                         0 },
        { SYMBOL_TYPE_NUMBER,     ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_IDENTIFIER, ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value_Identifier,      1 },
        { SYMBOL_TYPE_LITERAL,    ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_EXPR,       ACTION_TYPE_REDUCE,       -1, Reduce_Expr_BTermOrExpr,      3 },
        { SYMBOL_TYPE_BTERM,      ACTION_TYPE_NONE,          4, NULL,                         0 },
        { SYMBOL_TYPE_BFACTOR,    ACTION_TYPE_REDUCE,       -1, Reduce_BTerm_BFactor,         1 },
        { SYMBOL_TYPE_TERM,       ACTION_TYPE_REDUCE,       -1, Reduce_BFactor_Term,          1 },
        { SYMBOL_TYPE_VALUE,      ACTION_TYPE_NONE,          5, NULL,                         0 },
        { SYMBOL_TYPE_NONE,       ACTION_TYPE_ERROR,        -1, NULL,                         0 }
    },
    { // state 8
        { SYMBOL_TYPE_NOT,        ACTION_TYPE_SHIFT,         1, NULL,                         0 },
        { SYMBOL_TYPE_LPAREN,     ACTION_TYPE_SHIFT,         2, NULL,                         0 },
        { SYMBOL_TYPE_NUMBER,     ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_IDENTIFIER, ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value_Identifier,      1 },
        { SYMBOL_TYPE_LITERAL,    ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_BFACTOR,    ACTION_TYPE_REDUCE,       -1, Reduce_BTerm_BTermAndBFactor, 3 },
        { SYMBOL_TYPE_TERM,       ACTION_TYPE_REDUCE,       -1, Reduce_BFactor_Term,          1 },
        { SYMBOL_TYPE_VALUE,      ACTION_TYPE_NONE,          5, NULL,                         0 },
        { SYMBOL_TYPE_NONE,       ACTION_TYPE_ERROR,        -1, NULL,                         0 }
    },
    { // state 9
        { SYMBOL_TYPE_NUMBER,     ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_IDENTIFIER, ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value_Identifier,      1 },
        { SYMBOL_TYPE_LITERAL,    ACTION_TYPE_SHIFT_REDUCE, -1, Reduce_Value,                 1 },
        { SYMBOL_TYPE_VALUE,      ACTION_TYPE_REDUCE,       -1, Reduce_Term_ValueCompValue,   3 },
        { SYMBOL_TYPE_NONE,       ACTION_TYPE_ERROR,        -1, NULL,                         0 }
    }
};


// function definitions

HRESULT WIXAPI EvaluateCondition(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in LPCWSTR wzExpr,
    __out BOOL* pf
    )
{
    HRESULT hr = S_OK;
    DWORD iExpr = 0;
    int state = 0; // initial state is 0
    STATE* pStack = NULL;
    SYMBOL* pSymbol = NULL;
    SYMBOL* pLookAhead = NULL;
    ACTION* pAction = NULL;
    SYMBOL* rgReduceSymbols[MAX_REDUCE] = { 0 };

    for (;;)
    {
        // match symbol with an action for the current state
        for (pAction = vActionTable[state]; SYMBOL_TYPE_NONE != pAction->SymbolType; ++pAction)
        {
            if (NONTERMINAL & pAction->SymbolType)
            {
                // compare with top item on stack
                if (pStack && pStack->pSymbol->Type == pAction->SymbolType)
                {
                    break;
                }
            }
            else
            {
                // get look ahead token if we don't have one
                if (!pLookAhead)
                {
                    hr = NextToken(wzExpr, &iExpr, &pLookAhead);
                    ExitOnFailure(hr, "Failed to get next token.");
                }

                // compare with look ahead tooken
                if (pLookAhead->Type == pAction->SymbolType)
                {
                    break;
                }
            }
        }
        // at this point pAction is either an explicit action
        // for the symbol or the default action

        // error
        if (ACTION_TYPE_ERROR & pAction->ActionType)
        {
            ExitOnFailure1(hr = E_FAIL, "Unexpected symbol at %d.", pLookAhead ? pLookAhead->iPosition : 0);
        }

        // accept
        if (ACTION_TYPE_ACCEPT & pAction->ActionType)
        {
            break;
        }

        // shift
        if (ACTION_TYPE_SHIFT & pAction->ActionType)
        {
            // push parse stack
            hr = PushStack(&pStack, state, pLookAhead);
            ExitOnFailure(hr, "Failed to push parse stack.");
            pLookAhead = NULL;
        }

        // reduce
        if (ACTION_TYPE_REDUCE & pAction->ActionType)
        {
            // build array of symbols to reduce
            STATE* pState = pStack;
            for (int i = pAction->nReduce - 1; 0 <= i; --i)
            {
                rgReduceSymbols[i] = pState->pSymbol;
                pState = pState->pNext;
            }

            // allocate memory for new symbol
            pSymbol = (SYMBOL*)MemAlloc(sizeof(SYMBOL), TRUE);
            ExitOnNull(pSymbol, hr, E_OUTOFMEMORY, "Failed to allocate memory for symbol.");

            // call reduce function
            hr = pAction->pfnReduce(pEngineInfo, rgReduceSymbols, pSymbol);
            ExitOnFailure(hr, "Failed to reduce.");

            // pop stack and make then symbol resulting form the reduction the current symbol
            pSymbol->iPosition = rgReduceSymbols[0]->iPosition;
            hr = PopStack(&pStack, &state, pAction->nReduce);
            ExitOnFailure(hr, "Failed to pop parse stack.");
            hr = PushStack(&pStack, state, pSymbol);
            ExitOnFailure(hr, "Failed to push parse stack.");

            continue;
        }

        // if we did not performa a reduce we update the parser state
        state = pAction->state;
    }

    // return value
    *pf = VARIANT_FALSE != pStack->pSymbol->vtValue.boolVal;

LExit:
    if (pLookAhead)
    {
        FreeSymbol(pLookAhead);
    }
    if (pStack)
    {
        FreeState(pStack, TRUE);
    }
    return hr;
}


// internal function definitions

static HRESULT NextToken(
    __in LPCWSTR wz,
    __inout DWORD* pi,
    __out SYMBOL** ppSymbol
    )
{
    HRESULT hr = S_OK;
    WORD charType = 0;
    DWORD n = 0;
    SYMBOL* pSymbol = NULL;

    // allocate symbol
    pSymbol = (SYMBOL*)MemAlloc(sizeof(SYMBOL), TRUE);
    ExitOnNull(pSymbol, hr, E_OUTOFMEMORY, "Failed to allocate memory for symbol.");

    // skip past blanks
    for (;;)
    {
        ::GetStringTypeW(CT_CTYPE1, &wz[*pi], 1, &charType);
        if (0 == (C1_BLANK & charType))
        {
            break; // no blank, done
        }
        ++*pi;
    }
    pSymbol->iPosition = *pi;

    // read depending on first character type
    switch (wz[*pi])
    {
    case L'\0':
        pSymbol->Type = SYMBOL_TYPE_END;
        n = 1;
        break;
    case L'>':
        if (L'=' == wz[*pi + 1])
        {
            pSymbol->Type = SYMBOL_TYPE_GE;
            n = 2;
        }
        else
        {
            pSymbol->Type = SYMBOL_TYPE_GT;
            n = 1;
        }
        break;
    case L'<':
        switch (wz[*pi + 1])
        {
        case L'=':
            pSymbol->Type = SYMBOL_TYPE_LE;
            n = 2;
            break;
        case L'>':
            pSymbol->Type = SYMBOL_TYPE_NE;
            n = 2;
            break;
        default:
            pSymbol->Type = SYMBOL_TYPE_LT;
            n = 1;
        }
        break;
    case L'=':
        pSymbol->Type = SYMBOL_TYPE_EQ;
        n = 1;
        break;
    case L'(':
        pSymbol->Type = SYMBOL_TYPE_LPAREN;
        n = 1;
        break;
    case L')':
        pSymbol->Type = SYMBOL_TYPE_RPAREN;
        n = 1;
        break;
    case L'"': // literal
        do {
            ++n;
            if (L'\0' == wz[*pi + n])
            {
                // error
                ExitFunction1(hr = E_FAIL);
            }
        } while (L'"' != wz[*pi + n]);
        ++n; // terminating '"'

        pSymbol->Type = SYMBOL_TYPE_LITERAL;
        pSymbol->vtValue.vt = VT_BSTR;
        pSymbol->vtValue.bstrVal = ::SysAllocStringLen(&wz[*pi + 1], n - 2);
        ExitOnNull(pSymbol->vtValue.bstrVal, hr, E_OUTOFMEMORY, "Failed to allocate BSTR.");
        break;
    default:
        if (C1_DIGIT & charType)
        {
            do {
                ++n;
                ::GetStringTypeW(CT_CTYPE1, &wz[*pi + n], 1, &charType);
                if (C1_ALPHA & charType || L'_' == wz[*pi + n])
                {
                    // error (identifier can not start with a digit)
                    ExitFunction1(hr = E_FAIL);
                }
            } while (C1_DIGIT & charType);

            // number
            pSymbol->Type = SYMBOL_TYPE_NUMBER;
            pSymbol->vtValue.vt = VT_I8;
            pSymbol->vtValue.llVal = _wtoi64(&wz[*pi]); // TODO: handle overflow
        }
        else if (C1_ALPHA & charType || L'_' == wz[*pi])
        {
            do {
                ++n;
                ::GetStringTypeW(CT_CTYPE1, &wz[*pi + n], 1, &charType);
            } while (C1_ALPHA & charType || C1_DIGIT & charType || L'_' == wz[*pi + n]);

            if (2 == n && CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &wz[*pi], 2, L"OR", 2))
            {
                // OR
                pSymbol->Type = SYMBOL_TYPE_OR;
            }
            else if (3 == n && CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &wz[*pi], 3, L"AND", 3))
            {
                // AND
                pSymbol->Type = SYMBOL_TYPE_AND;
            }
            else if (3 == n && CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &wz[*pi], 3, L"NOT", 3))
            {
                // NOT
                pSymbol->Type = SYMBOL_TYPE_NOT;
            }
            else
            {
                // identifier
                pSymbol->Type = SYMBOL_TYPE_IDENTIFIER;
                pSymbol->vtValue.vt = VT_BSTR;
                pSymbol->vtValue.bstrVal = ::SysAllocStringLen(&wz[*pi], n);
                ExitOnNull(pSymbol->vtValue.bstrVal, hr, E_OUTOFMEMORY, "Failed to allocate BSTR.");
            }
        }
        else
        {
            // error
            ExitFunction1(hr = E_FAIL);
        }
    }
    *pi += n;

    // return value
    *ppSymbol = pSymbol;
    pSymbol = NULL;

LExit:
    if (pSymbol)
    {
        FreeSymbol(pSymbol);
    }
    return hr;
}


static HRESULT PushStack(
    __inout STATE** ppStack,
    __in int state,
    __in SYMBOL* pSymbol
    )
{
    HRESULT hr = S_OK;
    STATE* pState = NULL;

    // create state structure
    pState = (STATE*)MemAlloc(sizeof(STATE), TRUE);
    ExitOnNull(pState, hr, E_OUTOFMEMORY, "Failed to allocate memory for state struct.");

    pState->state = state;
    pState->pSymbol = pSymbol;

    // push stack
    pState->pNext = *ppStack;
    *ppStack = pState;

LExit:
    return hr;
}


static HRESULT PopStack(
    __inout STATE** ppStack,
    __inout int* piState,
    __in DWORD nLevels
    )
{
    HRESULT hr = S_OK;
    STATE* pState = NULL;
    int state = *piState;

    for (DWORD i = nLevels; 0 < i; --i)
    {
        // check that we haven't reached the bottom of the stack
        if (!*ppStack)
        {
            ExitOnFailure(hr = E_FAIL, "Stack empty.");
        }

        // remove top state
        pState = *ppStack;
        *ppStack = (*ppStack)->pNext;
        state = pState->state;

        // free state
        FreeState(pState, FALSE);
    }

    // return values
    *piState = state;

LExit:
    return hr;
}


static void FreeState(
    __in STATE* pState,
    __in BOOL fRecursive
    )
{
    if (fRecursive && pState->pNext)
    {
        FreeState(pState->pNext, TRUE);
    }
    if (pState->pSymbol)
    {
        FreeSymbol(pState->pSymbol);
    }
    MemFree(pState);
}


static void FreeSymbol(
    __in SYMBOL* pSymbol
    )
{
    ::VariantClear(&pSymbol->vtValue);
    MemFree(pSymbol);
}


static HRESULT Reduce_Expr_BTerm(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    UNREFERENCED_PARAMETER(pEngineInfo);
    Assert(VT_BOOL == rgSymbols[0]->vtValue.vt);

    pSymbol->Type = SYMBOL_TYPE_EXPR;
    return ::VariantCopy(&pSymbol->vtValue, &rgSymbols[0]->vtValue);
}


static HRESULT Reduce_Expr_BTermOrExpr(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    UNREFERENCED_PARAMETER(pEngineInfo);
    Assert(VT_BOOL == rgSymbols[0]->vtValue.vt);
    Assert(VT_BOOL == rgSymbols[2]->vtValue.vt);

    // compare
    BOOL f = (VARIANT_TRUE == rgSymbols[0]->vtValue.boolVal) || VARIANT_TRUE == (rgSymbols[2]->vtValue.boolVal);

    // reduced symbol
    pSymbol->Type = SYMBOL_TYPE_EXPR;
    pSymbol->vtValue.vt = VT_BOOL;
    pSymbol->vtValue.boolVal = f ? VARIANT_TRUE : VARIANT_FALSE;

    return S_OK;
}


static HRESULT Reduce_BTerm_BFactor(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    UNREFERENCED_PARAMETER(pEngineInfo);
    Assert(VT_BOOL == rgSymbols[0]->vtValue.vt);

    pSymbol->Type = SYMBOL_TYPE_BTERM;
    return ::VariantCopy(&pSymbol->vtValue, &rgSymbols[0]->vtValue);
}


static HRESULT Reduce_BFactor_Term(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    UNREFERENCED_PARAMETER(pEngineInfo);
    Assert(VT_BOOL == rgSymbols[0]->vtValue.vt);

    pSymbol->Type = SYMBOL_TYPE_BFACTOR;
    return ::VariantCopy(&pSymbol->vtValue, &rgSymbols[0]->vtValue);
}


static HRESULT Reduce_BFactor_NotTerm(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    UNREFERENCED_PARAMETER(pEngineInfo);
    Assert(VT_BOOL == rgSymbols[1]->vtValue.vt);

    // invert
    BOOL f = (VARIANT_FALSE == rgSymbols[1]->vtValue.boolVal);

    // reduced symbol
    pSymbol->Type = SYMBOL_TYPE_BFACTOR;
    pSymbol->vtValue.vt = VT_BOOL;
    pSymbol->vtValue.boolVal = f ? VARIANT_TRUE : VARIANT_FALSE;

    return S_OK;
}


static HRESULT Reduce_BTerm_BTermAndBFactor(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    UNREFERENCED_PARAMETER(pEngineInfo);
    Assert(VT_BOOL == rgSymbols[0]->vtValue.vt);
    Assert(VT_BOOL == rgSymbols[2]->vtValue.vt);

    // compare
    BOOL f = (VARIANT_TRUE == rgSymbols[0]->vtValue.boolVal) && VARIANT_TRUE == (rgSymbols[2]->vtValue.boolVal);

    // reduced symbol
    pSymbol->Type = SYMBOL_TYPE_BTERM;
    pSymbol->vtValue.vt = VT_BOOL;
    pSymbol->vtValue.boolVal = f ? VARIANT_TRUE : VARIANT_FALSE;

    return S_OK;
}


static HRESULT Reduce_Term_Value(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    UNREFERENCED_PARAMETER(pEngineInfo);
    Assert(VT_I8 == rgSymbols[0]->vtValue.vt || VT_BSTR == rgSymbols[0]->vtValue.vt);

    SYMBOL* pValue = rgSymbols[0];
    BOOL f = FALSE;

    // convert to boolean
    switch (pValue->vtValue.vt)
    {
    case VT_I8:
        f = 0 != pValue->vtValue.llVal;
        break;
    case VT_BSTR:
        f = pValue->vtValue.bstrVal && pValue->vtValue.bstrVal[0];
        break;
    }

    // reduced symbol
    pSymbol->Type = SYMBOL_TYPE_TERM;
    pSymbol->vtValue.vt = VT_BOOL;
    pSymbol->vtValue.boolVal = f ? VARIANT_TRUE : VARIANT_FALSE;

    return S_OK;
}


static HRESULT Reduce_Term_ValueCompValue(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    UNREFERENCED_PARAMETER(pEngineInfo);
    Assert(VT_I8 == rgSymbols[0]->vtValue.vt || VT_BSTR == rgSymbols[0]->vtValue.vt);
    Assert(VT_I8 == rgSymbols[2]->vtValue.vt || VT_BSTR == rgSymbols[2]->vtValue.vt);

    HRESULT hr = S_OK;
    SYMBOL* pLeftValue = rgSymbols[0];
    SYMBOL* pRightValue = rgSymbols[2];
    LONGLONG llLeftCompare = 0;
    LONGLONG llRightCompare = 0;
    BOOL f = FALSE;
    VARIANT vt = { 0 };

    // get values to compare based on type
    if (VT_I8 == pLeftValue->vtValue.vt && VT_I8 == pRightValue->vtValue.vt)
    {
        llLeftCompare = pLeftValue->vtValue.llVal;
        llRightCompare = pRightValue->vtValue.llVal;
    }
    else if (VT_BSTR == pLeftValue->vtValue.vt && VT_BSTR == pRightValue->vtValue.vt)
    {
        llLeftCompare = ::CompareStringW(LOCALE_INVARIANT, 0, pLeftValue->vtValue.bstrVal, -1, pRightValue->vtValue.bstrVal, -1);
        llRightCompare = CSTR_EQUAL;
    }
    else if (VT_I8 == pLeftValue->vtValue.vt && VT_BSTR == pRightValue->vtValue.vt)
    {
        hr = ::VariantChangeType(&vt, &pRightValue->vtValue, 0, VT_I8);
        if (DISP_E_TYPEMISMATCH != hr)
        {
            ExitOnFailure(hr, "Failed to change variant type.");

            llLeftCompare = pLeftValue->vtValue.llVal;
            llRightCompare = vt.llVal;
        }
    }
    else if (VT_BSTR == pLeftValue->vtValue.vt && VT_I8 == pRightValue->vtValue.vt)
    {
        hr = ::VariantChangeType(&vt, &pLeftValue->vtValue, 0, VT_I8);
        if (DISP_E_TYPEMISMATCH != hr)
        {
            ExitOnFailure(hr, "Failed to change variant type.");

            llLeftCompare = vt.llVal;
            llRightCompare = pRightValue->vtValue.llVal;
        }
    }

    // compare values
    if (DISP_E_TYPEMISMATCH == hr)
    {
        // if we failed to convert to comparable types
        f = SYMBOL_TYPE_NE == rgSymbols[1]->Type;
    }
    else
    {
        switch (rgSymbols[1]->Type)
        {
        case SYMBOL_TYPE_LT: f = llLeftCompare <  llRightCompare; break;
        case SYMBOL_TYPE_GT: f = llLeftCompare >  llRightCompare; break;
        case SYMBOL_TYPE_LE: f = llLeftCompare <= llRightCompare; break;
        case SYMBOL_TYPE_GE: f = llLeftCompare >= llRightCompare; break;
        case SYMBOL_TYPE_EQ: f = llLeftCompare == llRightCompare; break;
        case SYMBOL_TYPE_NE: f = llLeftCompare != llRightCompare; break;
        }
    }

    // reduced symbol
    pSymbol->Type = SYMBOL_TYPE_TERM;
    pSymbol->vtValue.vt = VT_BOOL;
    pSymbol->vtValue.boolVal = f ? VARIANT_TRUE : VARIANT_FALSE;

    hr = S_OK;

LExit:
    return hr;
}


static HRESULT Reduce_Term_LParenExprRParen(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    UNREFERENCED_PARAMETER(pEngineInfo);
    Assert(VT_BOOL == rgSymbols[1]->vtValue.vt);

    pSymbol->Type = SYMBOL_TYPE_TERM;
    return ::VariantCopy(&pSymbol->vtValue, &rgSymbols[1]->vtValue);
}


static HRESULT Reduce_Value(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    UNREFERENCED_PARAMETER(pEngineInfo);
    Assert(VT_I8 == rgSymbols[0]->vtValue.vt || VT_BSTR == rgSymbols[0]->vtValue.vt);

    pSymbol->Type = SYMBOL_TYPE_VALUE;
    return ::VariantCopy(&pSymbol->vtValue, &rgSymbols[0]->vtValue);
}


static HRESULT Reduce_Value_Identifier(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SYMBOL** rgSymbols,
    __out SYMBOL* pSymbol
    )
{
    Assert(VT_BSTR == rgSymbols[0]->vtValue.vt);

    pSymbol->Type = SYMBOL_TYPE_VALUE;
    return GetVarProperty(pEngineInfo, rgSymbols[0]->vtValue.bstrVal, &pSymbol->vtValue);
}
