//-------------------------------------------------------------------------------------------------
// <copyright file="search.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
// </copyright>
//
// <summary>
//
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


// structs

struct DEPENDENDY_CHAIN_LINK
{
    DEPENDENDY_CHAIN_LINK* pPrevious;
    LPCWSTR sczKey;
};


// internal function declarations


static HRESULT CopySearchWithDependencies(
    __in DEPENDENDY_CHAIN_LINK* pPreviousLink,
    __in BURN_SEARCHES* pSearches,
    __in DWORD iCurrent,
    __in BOOL* rgfCopied,
    __in BURN_SEARCH* rgTarget,
    __inout DWORD* piTarget
    );
static HRESULT DirectorySearchExists(
    __in BURN_SEARCH* pSearch,
    __in BURN_PROPERTIES* pProperties
    );
static HRESULT FileSearchExists(
    __in BURN_SEARCH* pSearch,
    __in BURN_PROPERTIES* pProperties
    );
static HRESULT FileSearchVersion(
    __in BURN_SEARCH* pSearch,
    __in BURN_PROPERTIES* pProperties
    );
static HRESULT RegistrySearchExists(
    __in BURN_SEARCH* pSearch,
    __in BURN_PROPERTIES* pProperties
    );
static HRESULT RegistrySearchValue(
    __in BURN_SEARCH* pSearch,
    __in BURN_PROPERTIES* pProperties
    );


// function definitions

extern "C" HRESULT SearchAdd(
    __in BURN_SEARCHES* pSearches,
    __out BURN_SEARCH** ppSearch
    )
{
    HRESULT hr = S_OK;

    if (pSearches->rgSearches)
    {
        LPVOID pv = MemReAlloc(pSearches->rgSearches, sizeof(BURN_SEARCH) * (pSearches->cSearches + 1), TRUE);
        ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to reallocate search array.");
        pSearches->rgSearches = (BURN_SEARCH*)pv;
    }
    else
    {
        pSearches->rgSearches = (BURN_SEARCH*)MemAlloc(sizeof(BURN_SEARCH) * (pSearches->cSearches + 1), TRUE);
        ExitOnNull(pSearches->rgSearches, hr, E_OUTOFMEMORY, "Failed to allocate search array.");
    }
    *ppSearch = &pSearches->rgSearches[pSearches->cSearches];
    ++pSearches->cSearches;

LExit:
    return hr;
}

extern "C" HRESULT SearchAddDependency(
    __in BURN_SEARCH* pSearch,
    __out LPWSTR** ppsczSearchDependency
    )
{
    HRESULT hr = S_OK;

    if (pSearch->rgsczDependencies)
    {
        LPVOID pv = MemReAlloc(pSearch->rgsczDependencies, sizeof(LPWSTR) * (pSearch->cDependencies + 1), TRUE);
        ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to reallocate search dependency array.");
        pSearch->rgsczDependencies = (LPWSTR*)pv;
    }
    else
    {
        pSearch->rgsczDependencies = (LPWSTR*)MemAlloc(sizeof(LPWSTR) * (pSearch->cDependencies + 1), TRUE);
        ExitOnNull(pSearch->rgsczDependencies, hr, E_OUTOFMEMORY, "Failed to allocate search dependency array.");
    }
    *ppsczSearchDependency = &pSearch->rgsczDependencies[pSearch->cDependencies];
    ++pSearch->cDependencies;

LExit:
    return hr;
}

extern "C" HRESULT SearchesSort(
    __in BURN_SEARCHES* pSearches
    )
{
    HRESULT hr = S_OK;
    BOOL* rgfCopied = NULL;
    BURN_SEARCH* rgTarget = NULL;
    DWORD iTarget = 0;

    // create copied flag array to track items that has already been copied
    rgfCopied = (BOOL*)MemAlloc(pSearches->cSearches * sizeof(BOOL), TRUE);
    ExitOnNull(rgfCopied, hr, E_OUTOFMEMORY, "Failed to allocate copied flag array.");

    // create target array for sorted items
    rgTarget = (BURN_SEARCH*)MemAlloc(pSearches->cSearches * sizeof(BURN_SEARCH), TRUE);
    ExitOnNull(rgTarget, hr, E_OUTOFMEMORY, "Failed to allocate target search array.");

    // copy items one by one, including dependencies
    for (DWORD i = 0; i < pSearches->cSearches; ++i)
    {
        hr = CopySearchWithDependencies(NULL, pSearches, i, rgfCopied, rgTarget, &iTarget);
        ExitOnFailure1(hr, "Failed to copy search element. Key: %S", pSearches->rgSearches[i].sczKey);
    }

    // replace array
    ReleaseMem(pSearches->rgSearches);
    pSearches->rgSearches = rgTarget;
    rgTarget = NULL;

LExit:
    ReleaseMem(rgfCopied);
    ReleaseMem(rgTarget);
    return hr;
}

extern "C" HRESULT SearchesExecute(
    __in BURN_SEARCHES* pSearches,
    __in BURN_PROPERTIES* pProperties,
    __in BURN_SEARCH_SCHEDULING scheduling
    )
{
    HRESULT hr = S_OK;
    BOOL f = FALSE;

    for (DWORD i = 0; i < pSearches->cSearches; ++i)
    {
        BURN_SEARCH* pSearch = &pSearches->rgSearches[i];

        // exclued if not the requested scheduling
        if (scheduling != pSearch->Scheduling)
        {
            continue;
        }

        // evaluate condition
        if (pSearch->sczCondition && *pSearch->sczCondition)
        {
            hr = ConditionEvaluate(pProperties, pSearch->sczCondition, &f);
            if (S_FALSE == hr)
            {
                TraceError2(hr, "Failed to parse search condition. Id = '%S', Condition = '%S'", pSearch->sczKey, pSearch->sczCondition);
                continue;
            }
            ExitOnFailure2(hr, "Failed to evaluate search condition. Id = '%S', Condition = '%S'", pSearch->sczKey, pSearch->sczCondition);

            if (!f)
            {
                continue; // condition evaluated to false, skip
            }
        }

        switch (pSearch->Type)
        {
        case BURN_SEARCH_TYPE_DIRECTORY:
            switch (pSearch->DirectorySearch.Type)
            {
            case BURN_DIRECTORY_SEARCH_TYPE_EXISTS:
                hr = DirectorySearchExists(pSearch, pProperties);
                break;
            default:
                hr = E_UNEXPECTED;
            }
            break;
        case BURN_SEARCH_TYPE_FILE:
            switch (pSearch->FileSearch.Type)
            {
            case BURN_FILE_SEARCH_TYPE_EXISTS:
                hr = FileSearchExists(pSearch, pProperties);
                break;
            case BURN_FILE_SEARCH_TYPE_VERSION:
                hr = FileSearchVersion(pSearch, pProperties);
                break;
            default:
                hr = E_UNEXPECTED;
            }
            break;
        case BURN_SEARCH_TYPE_REGISTRY:
            switch (pSearch->RegistrySearch.Type)
            {
            case BURN_REGISTRY_SEARCH_TYPE_EXISTS:
                hr = RegistrySearchExists(pSearch, pProperties);
                break;
            case BURN_REGISTRY_SEARCH_TYPE_VALUE:
                hr = RegistrySearchValue(pSearch, pProperties);
                break;
            default:
                hr = E_UNEXPECTED;
            }
            break;
        default:
            hr = E_UNEXPECTED;
        }
        if (FAILED(hr))
        {
            TraceError1(hr, "Search failed. Id = '%S'", pSearch->sczKey);
            continue;
        }
    }

    hr = S_OK;

LExit:
    return hr;
}

extern "C" void SearchesUninitialize(
    __in BURN_SEARCHES* pSearches
    )
{
    if (pSearches->rgSearches)
    {
        for (DWORD i = 0; i < pSearches->cSearches; ++i)
        {
            SearchFree(&pSearches->rgSearches[i]);
        }
        MemFree(pSearches->rgSearches);
    }
}

extern "C" void SearchFree(
    __in BURN_SEARCH* pSearch
    )
{
    ReleaseStr(pSearch->sczKey);
    ReleaseStr(pSearch->sczProperty);
    ReleaseStr(pSearch->sczCondition);

    for (DWORD iDependency = 0; iDependency < pSearch->cDependencies; ++iDependency)
    {
        ReleaseStr(pSearch->rgsczDependencies[iDependency]);
    }

    switch (pSearch->Type)
    {
    case BURN_SEARCH_TYPE_DIRECTORY:
        ReleaseStr(pSearch->DirectorySearch.sczPath);
        break;
    case BURN_SEARCH_TYPE_FILE:
        ReleaseStr(pSearch->FileSearch.sczPath);
        break;
    case BURN_SEARCH_TYPE_REGISTRY:
        ReleaseStr(pSearch->RegistrySearch.sczKey);
        ReleaseStr(pSearch->RegistrySearch.sczValue);
        break;
    }
}


// internal function definitions

static HRESULT CopySearchWithDependencies(
    __in DEPENDENDY_CHAIN_LINK* pPreviousLink,
    __in BURN_SEARCHES* pSearches,
    __in DWORD iCurrent,
    __in BOOL* rgfCopied,
    __in BURN_SEARCH* rgTarget,
    __inout DWORD* piTarget
    )
{
    HRESULT hr = S_OK;
    BURN_SEARCH* pSearch = &pSearches->rgSearches[iCurrent];
    DEPENDENDY_CHAIN_LINK link = {pPreviousLink, pSearch->sczKey};
    DWORD iDependency = 0;

    // check if search has already been copied
    if (rgfCopied[iCurrent])
    {
        ExitFunction1(hr = S_OK); // search already in target list, nothing to do
    }

    // copy dependencies before current element
    for (DWORD i = 0; i < pSearch->cDependencies; ++i)
    {
        LPCWSTR wzDependency = pSearch->rgsczDependencies[i];

        // check for circular dependencies
        for (DEPENDENDY_CHAIN_LINK* pCurrent = &link; pCurrent; pCurrent = pCurrent->pPrevious)
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, wzDependency, -1, pCurrent->sczKey, -1))
            {
                hr = E_FAIL;
                ExitOnFailure1(hr, "Circular search dependency found. Key = %S", wzDependency);
            }
        }

        // find search
        hr = HRESULT_FROM_WIN32(ERROR_NOT_FOUND);
        for (iDependency = 0; i < pSearches->cSearches; ++i)
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, wzDependency, -1, pSearches->rgSearches[iDependency].sczKey, -1))
            {
                hr = S_OK;
                break;
            }
        }
        ExitOnFailure1(hr, "Failed to find search dependency. Key = %S", wzDependency);

        // copy dependency
        hr = CopySearchWithDependencies(&link, pSearches, iDependency, rgfCopied, rgTarget, piTarget);
        ExitOnFailure1(hr, "Failed to move dependent searches. Key = %S", wzDependency);
    }

    // copy dependent search
    memcpy_s(&rgTarget[*piTarget], sizeof(BURN_SEARCH), pSearch, sizeof(BURN_SEARCH));
    ++*piTarget;
    rgfCopied[iCurrent] = TRUE;

LExit:
    return hr;
}

static HRESULT DirectorySearchExists(
    __in BURN_SEARCH* pSearch,
    __in BURN_PROPERTIES* pProperties
    )
{
    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;

    DWORD dwAttributes = ::GetFileAttributesW(pSearch->DirectorySearch.sczPath);
    if (INVALID_FILE_ATTRIBUTES == dwAttributes)
    {
        er = ::GetLastError();
        if (ERROR_FILE_NOT_FOUND != er)
        {
            hr = HRESULT_FROM_WIN32(er);
            ExitOnFailure1(hr, "Failed get to file attributes. '%S'", pSearch->DirectorySearch.sczPath);
        }
    }
    else if (dwAttributes & FILE_ATTRIBUTE_DIRECTORY)
    {
        hr = PropertySetNumeric(pProperties, pSearch->sczProperty, TRUE);
    }

LExit:
    return hr;
}

static HRESULT FileSearchExists(
    __in BURN_SEARCH* pSearch,
    __in BURN_PROPERTIES* pProperties
    )
{
    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;
    LPWSTR sczPath = NULL;

    // format path
    hr = PropertyFormatString(pProperties, pSearch->DirectorySearch.sczPath, &sczPath);
    ExitOnFailure(hr, "Failed to format property string.");

    DWORD dwAttributes = ::GetFileAttributesW(sczPath);
    if (INVALID_FILE_ATTRIBUTES == dwAttributes)
    {
        er = ::GetLastError();
        if (ERROR_FILE_NOT_FOUND != er)
        {
            hr = HRESULT_FROM_WIN32(er);
            ExitOnFailure1(hr, "Failed get to file attributes. '%S'", pSearch->DirectorySearch.sczPath);
        }
    }
    else if (FILE_ATTRIBUTE_DIRECTORY != (dwAttributes & FILE_ATTRIBUTE_DIRECTORY))
    {
        hr = PropertySetNumeric(pProperties, pSearch->sczProperty, TRUE);
    }

LExit:
    ReleaseStr(sczPath);
    return hr;
}

static HRESULT FileSearchVersion(
    __in BURN_SEARCH* pSearch,
    __in BURN_PROPERTIES* pProperties
    )
{
    HRESULT hr = S_OK;
    ULARGE_INTEGER uliVersion = { 0 };
    LPWSTR sczPath = NULL;

    // format path
    hr = PropertyFormatString(pProperties, pSearch->FileSearch.sczPath, &sczPath);
    ExitOnFailure(hr, "Failed to format path string.");

    // get file version
    hr = FileVersion(sczPath, &uliVersion.HighPart, &uliVersion.LowPart);
    ExitOnFailure(hr, "Failed get file version.");

    // set property
    hr = PropertySetVersion(pProperties, pSearch->sczProperty, uliVersion.QuadPart);

LExit:
    ReleaseStr(sczPath);
    return hr;
}

static HRESULT RegistrySearchExists(
    __in BURN_SEARCH* pSearch,
    __in BURN_PROPERTIES* pProperties
    )
{
    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;
    LPWSTR sczKey = NULL;
    HKEY hKey = NULL;
    BOOL fExists = FALSE;

    // format key string
    hr = PropertyFormatString(pProperties, pSearch->RegistrySearch.sczKey, &sczKey);
    ExitOnFailure(hr, "Failed to format key string.");

    // open key
    er = ::RegOpenKeyExW(pSearch->RegistrySearch.hRoot, sczKey, 0, KEY_QUERY_VALUE, &hKey);
    switch (er)
    {
    case ERROR_SUCCESS:
        fExists = TRUE;
        break;
    case ERROR_FILE_NOT_FOUND:
        break;
    default:
        hr = HRESULT_FROM_WIN32(er);
        ExitOnFailure(hr, "Failed to open registry key.");
    }

    hr = PropertySetNumeric(pProperties, pSearch->sczProperty, fExists);

LExit:
    ReleaseStr(sczKey);
    if (hKey)
    {
        ::RegCloseKey(hKey);
    }
    return hr;
}

static HRESULT RegistrySearchValue(
    __in BURN_SEARCH* pSearch,
    __in BURN_PROPERTIES* pProperties
    )
{
    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;
    LPWSTR sczKey = NULL;
    LPWSTR sczValue = NULL;
    HKEY hKey = NULL;
    DWORD dwType = 0;
    DWORD cbData = 0;
    LPBYTE pData = NULL;
    DWORD cch = 0;
    BURN_VARIANT value = { 0 };

    // format key string
    hr = PropertyFormatString(pProperties, pSearch->RegistrySearch.sczKey, &sczKey);
    ExitOnFailure(hr, "Failed to format key string.");

    // format value string
    hr = PropertyFormatString(pProperties, pSearch->RegistrySearch.sczValue, &sczValue);
    ExitOnFailure(hr, "Failed to format value string.");

    // open key
    er = ::RegOpenKeyExW(pSearch->RegistrySearch.hRoot, sczKey, 0, KEY_QUERY_VALUE, &hKey);
    if (ERROR_FILE_NOT_FOUND == er)
    {
        LogStringLine(REPORT_STANDARD, "Registry key not found. Key = '%S'", sczKey);
        ExitFunction1(hr = S_OK);
    }
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to open registry key.");

    // get value
    er = ::RegQueryValueExW(hKey, sczValue, NULL, &dwType, NULL, &cbData);
    if (ERROR_FILE_NOT_FOUND == er)
    {
        LogStringLine(REPORT_STANDARD, "Registry value not found. Key = '%S', Value = '%S'", sczKey, sczValue);
        ExitFunction1(hr = S_OK);
    }
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to query registry key value size.");

    pData = (LPBYTE)MemAlloc(cbData + sizeof(WCHAR), TRUE); // + sizeof(WCHAR) here to ensure that we always have a null terminator for REG_SZ
    ExitOnNull(pData, hr, E_OUTOFMEMORY, "Failed to allocate memory registry value.");

    er = ::RegQueryValueExW(hKey, sczValue, NULL, &dwType, pData, &cbData);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to query registry key value.");

    switch (dwType)
    {
    case REG_DWORD:
        if (sizeof(LONG) != cbData)
        {
            ExitFunction1(hr = E_UNEXPECTED);
        }
        hr = BVariantSetNumeric(&value, *((LONG*)pData));
        break;
    case REG_QWORD:
        if (sizeof(LONGLONG) != cbData)
        {
            ExitFunction1(hr = E_UNEXPECTED);
        }
        hr = BVariantSetNumeric(&value, *((LONGLONG*)pData));
        break;
    case REG_EXPAND_SZ:
        if (pSearch->RegistrySearch.fExpandEnvironment)
        {
            hr = StrAlloc(&value.wzValue, cbData);
            ExitOnFailure(hr, "Failed to allocate string buffer.");
            value.Type = BURN_VARIANT_TYPE_STRING;

            cch = ExpandEnvironmentStringsW((LPCWSTR)pData, value.wzValue, cbData);
            if (cch > cbData)
            {
                hr = StrAlloc(&value.wzValue, cch);
                ExitOnFailure(hr, "Failed to allocate string buffer.");

                if (cch != ExpandEnvironmentStringsW((LPCWSTR)pData, value.wzValue, cch))
                {
                    ExitWithLastError(hr, "Failed to get expand environment string.");
                }
            }
            break;
        }
    case REG_SZ:
        hr = BVariantSetString(&value, (LPCWSTR)pData, 0);
        break;
    default:
        ExitOnFailure1(hr = E_NOTIMPL, "Unsupported registry key value type. Type = '%u'", dwType);
    }
    ExitOnFailure(hr, "Failed to read registry value.");

    // change value to requested type
    hr = BVariantChangeType(&value, pSearch->RegistrySearch.ValueType);
    ExitOnFailure(hr, "Failed to set property.");

    // set property
    hr = PropertySetVariant(pProperties, pSearch->sczProperty, &value);
    ExitOnFailure(hr, "Failed to set property.");

LExit:
    ReleaseStr(sczKey);
    ReleaseStr(sczValue);
    if (hKey)
    {
        ::RegCloseKey(hKey);
    }
    ReleaseMem(pData);
    BVariantUninitialize(&value);
    return hr;
}
