﻿//-------------------------------------------------------------------------------------------------
// <copyright file="BurnTask.cs" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
// 
// <summary>
// WiX Burn bootstrapper build task runs the Burn bundle bunder given a Burn chainer manifest.
// </summary>
//-------------------------------------------------------------------------------------------------

namespace Microsoft.Tools.WindowsInstallerXml.Build.Tasks
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Xml;
    using Microsoft.Build.Framework;
    using Microsoft.Build.Utilities;

    /// <summary>
    /// An MSBuild task to generate chainer manifests and run the Burn.exe bundle builder.
    /// </summary>
    public class BurnTask : WixToolTask
    {
        private const string BurnToolName = "burn.exe";

        #region Burn.exe switch equivalent properties
        /// <summary>
        /// Gets or sets the base input paths for finding files on disk. Equivalent to burn.exe -b switch.
        /// </summary>
        /// <value>A list of directories.</value>
        public string[] BaseInputPaths { get; set; }

        /// <summary>
        /// Gets or sets the bundle's default compression level. Equivalent to burn.exe -dcl switch.
        /// </summary>
        /// <value>One of: none, low, medium, high, mszip.</value>
        public string DefaultCompressionLevel { get; set; }

        /// <summary>
        /// Gets or sets preprocessor variable definitions. Equivalent to burn.exe -d switch.
        /// </summary>
        /// <value>Name=Value pairs for each preprocessor variable.</value>
        public string[] DefineConstants { get; set; }

        /// <summary>
        /// Gets or sets search path for files included in generated Burn manifest. Equivalent to burn.exe -I switch.
        /// </summary>
        /// <value>A list of directories.</value>
        public string[] IncludeSearchPaths { get; set; }

        /// <summary>
        /// Gets or sets the path to the Burn manifest to build.
        /// </summary>
        /// <value>Path to the input Burn manifest.</value>
        [Required]
        public ITaskItem Manifest { get; set; }

        /// <summary>
        /// Gets or sets the path to the output bundle. Equivalent to burn.exe -out switch.
        /// </summary>
        /// <value>Path to the output bundle.</value>
        [Required]
        [Output]
        public ITaskItem OutputFile { get; set; }
        #endregion

        #region ToolTask and Task overrides
        /// <summary>
        /// Gets the name of the executable.
        /// </summary>
        /// <remarks>The ToolName is used with the ToolPath to get the location of burn.exe.</remarks>
        /// <value>The name of the executable.</value>
        protected override string ToolName
        {
            get { return BurnToolName; }
        }

        /// <summary>
        /// Get the path to the executable. 
        /// </summary>
        /// <remarks>GenerateFullPathToTool is only called when the ToolPath property is not set (see the ToolName remarks above).</remarks>
        /// <returns>The full path to the executable or simply burn.exe if it's expected to be in the system path.</returns>
        protected override string GenerateFullPathToTool()
        {
            // If there's not a ToolPath specified, it has to be in the system path.
            if (String.IsNullOrEmpty(this.ToolPath))
            {
                return BurnToolName;
            }

            return Path.Combine(Path.GetFullPath(this.ToolPath), BurnToolName);
        }
        #endregion

        /// <summary>
        /// Builds a command line from options in this and derivative tasks.
        /// </summary>
        /// <param name="commandLineBuilder">Command-line builder object from the base class.</param>
        protected override void BuildCommandLine(WixCommandLineBuilder commandLineBuilder)
        {
            commandLineBuilder.AppendArrayIfNotNull("-b ", this.BaseInputPaths);
            commandLineBuilder.AppendArrayIfNotNull("-d", this.DefineConstants);
            commandLineBuilder.AppendSwitchIfNotNull("-dcl:", this.DefaultCompressionLevel);
            commandLineBuilder.AppendArrayIfNotNull("-I", this.IncludeSearchPaths);
            commandLineBuilder.AppendIfTrue("-nologo", this.NoLogo);
            commandLineBuilder.AppendSwitchIfNotNull("-out ", this.OutputFile);
            commandLineBuilder.AppendIfTrue("-v", this.VerboseOutput);
            commandLineBuilder.AppendFileNameIfNotNull(this.Manifest);
        }
    }
}
