﻿// <copyright file="ControllerTest.cs" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

namespace MS.Tools.WindowsInstallerXml.Test.Bootstrapper
{
    using System;
    using System.IO;
    using System.Threading;
    using System.Windows.Forms;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using System.IO.Pipes;
    using System.Text;

    /// <summary>
    /// Unit test for Burn controller.
    /// </summary>
    [TestClass]
    public class ControllerTest
    {
        private readonly string RootWorkingFolder;
        private readonly string CompletedFolder;

        public ControllerTest()
        {
            this.RootWorkingFolder = Path.Combine(Path.GetTempPath(), Guid.Empty.ToString("B"));
            this.CompletedFolder = Path.Combine(this.RootWorkingFolder, "completed");
        }

        [TestInitialize]
        public void InitializeTest()
        {
            // Always clean up the working directory.
            try
            {
                if (Directory.Exists(this.RootWorkingFolder))
                {
                    Directory.Delete(this.RootWorkingFolder, true);
                }
            }
            catch
            {
            }
        }

        [TestMethod]
        public void ControllerCacheTest()
        {
            string payloadFileName = "test.txt";
            using (BurnPayloadTestShim payload = new BurnPayloadTestShim(PayloadSourceType.Download, Guid.Empty, "http://localhost/burntest/test.txt", payloadFileName, 8, new byte[256], false))
            {
                payload.CompletedRoot = this.CompletedFolder;

                using (BurnMsiPackageTestShim msiPackage = new BurnMsiPackageTestShim("download", Guid.Empty.ToString("B"), new Version(0, 9, 0, 0), false))
                {
                    msiPackage.AddPayload(payload);

                    using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe"))
                    {
                        model.AddPackage(msiPackage);

                        using (MockBurnView view = new MockBurnView())
                        {
                            using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
                            {
                                Assert.IsFalse(File.Exists(Path.Combine(this.CompletedFolder, payloadFileName)));

                                controller.OnCachePackages(0);

                                Assert.AreEqual(TestConstants.Success, view.Result);
                                Assert.IsTrue(File.Exists(Path.Combine(this.CompletedFolder, payloadFileName)));
                            }
                        }
                    }
                }
            }

            //using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe"))
            //{
            //    BurnViewMock view = new BurnViewMock();
            //    using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
            //    {
            //        model.Load("setup.exe");

            //        AsyncControllerContext context = new AsyncControllerContext { Controller = controller, View = view };
            //        Thread t = ControllerTest.AsyncControllerStart(context);

            //        controller.Cache(0);
            //        view.WaitEvent.WaitOne();
            //        view.Done();

            //        context.CompleteEvent.WaitOne();

            //        Assert.AreEqual(TestConstants.Success, view.Result);
            //    }
            //}
        }

        [TestMethod]
        public void ControllerElevateTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe"))
            {
                using (MockBurnView view = new MockBurnView())
                {
                    using (MockInstallationAbstractionLayer ial = new MockInstallationAbstractionLayer())
                    {
                        using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view, ial))
                        {
                            controller.Elevate();
                            ial.JoinMockElevatedProcess();

                            Assert.AreEqual(ial.ExpectedToken, ial.Result.Token);
                        }
                    }
                }
            }
        }

        [TestMethod]
        public void ControllerElevatedTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe -burn.elevated pipeName token"))
            {
                Assert.AreEqual(BurnMode.Elevated, model.Mode);

                MockViewElevated view = new MockViewElevated();
                using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
                {
                    controller.Run();

                    Assert.IsTrue(view.Initialized);
                    Assert.IsTrue(view.Ran);
                    Assert.IsTrue(view.Uninitialized);
                }
            }
        }

        [TestMethod]
        [DeploymentItem(@"data\setup_external_ux.exe")]
        [DeploymentItem("data\\stdux.dll")]
        [DeploymentItem("data\\thm.xml")]
        [DeploymentItem("data\\en-us.wxl")]
        [DeploymentItem("data\\ps-ps.wxl")]
        [DeploymentItem("data\\stdux.png")]
        [DeploymentItem("data\\tou.htm")]
        public void ControllerPassiveInstallTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe -passive"))
            {
                using (BurnViewTestShim view = new BurnViewTestShim())
                {
                    using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
                    {
                        model.Load("setup_external_ux.exe");
                        controller.Run();
                    }
                }
            }
        }

        [TestMethod]
        [DeploymentItem(@"data\setup_external_ux.exe")]
        [DeploymentItem("data\\stdux.dll")]
        [DeploymentItem("data\\thm.xml")]
        [DeploymentItem("data\\en-us.wxl")]
        [DeploymentItem("data\\ps-ps.wxl")]
        [DeploymentItem("data\\stdux.png")]
        [DeploymentItem("data\\tou.htm")]
        public void ControllerPassiveUninstallTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe -passive -uninstall"))
            {
                BurnViewTestShim view = new BurnViewTestShim();
                using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
                {
                    model.Load("setup_external_ux.exe");
                    controller.Run();
                }
            }
        }

        [TestMethod]
        [DeploymentItem(@"data\setup.exe")]
        public void ControllerRollbackTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe"))
            {
                MockBurnView view = new MockBurnView();
                using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
                {
                    model.Load("setup.exe");

                    // Ensure the 1st MSI package fails when attempting to install.
                    using (BurnPackageTestShim package = model.GetPackage(2))
                    {
                        package.Properties.SetString("FAIL", "1");
                    }

                    controller.Run();
                    Assert.AreEqual(TestConstants.InstallFailure, view.Result);
                }
            }
        }

        [TestMethod]
        [DeploymentItem(@"data\setup.exe")]
        public void ControllerMockInstallUninstallTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe"))
            {
                using (MockBurnView view = new MockBurnView())
                {
                    using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
                    {
                        model.Load("setup.exe");
                        controller.Run();
                    }
                }
            }

            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe -uninstall"))
            {
                using (MockBurnView view = new MockBurnView())
                {
                    using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
                    {
                        model.Load("setup.exe");
                        controller.Run();
                    }
                }
            }
        }

        private static Thread AsyncControllerStart(AsyncControllerContext context)
        {
            context.View.RunForever = true;

            Thread t = new Thread(new ParameterizedThreadStart(ControllerTest.RunControllerThread));
            t.Start(context);

            WaitHandle[] waitAny = new WaitHandle[] { context.CompleteEvent, context.View.WaitEvent };
            if (0 == ManualResetEvent.WaitAny(waitAny))
            {
                throw new InvalidOperationException("Failed to start controller async thread.");
            }

            return t;
        }

        private static void RunControllerThread(object obj)
        {
            AsyncControllerContext context = (AsyncControllerContext)obj;

            context.CompleteEvent.Reset(); // not done
            context.Controller.Run();
            context.CompleteEvent.Set(); // done
        }

        /// <summary>
        /// Mock view implementation.
        /// </summary>
        private class MockBurnView : BurnViewMockBase
        {
            private AutoResetEvent wait = new AutoResetEvent(false);
            private AutoResetEvent waitForever = new AutoResetEvent(false);
            private BurnModelTestShim model;
            private BurnControllerTestShim controller;

            private bool runForever = false;
            private int packageCount = 0;
            private bool execute = true;
            private int currentPackage = 0;
            private uint currentPercentage = 0;
            private uint totalPercentage = 0;
            private int result = TestConstants.Success;

            public int Result
            {
                get { return this.result; }
            }

            public bool RunForever
            {
                get { return this.runForever; }
                set { this.runForever = value; }
            }

            public AutoResetEvent WaitEvent
            {
                get { return wait; }
            }

            public void Done()
            {
                this.waitForever.Set();
            }

            public override int Initialize(BurnModelTestShim model, BurnControllerTestShim controller)
            {
                this.model = model;
                this.controller = controller;

                return TestConstants.Success;
            }

            public override int Run()
            {
                if (this.runForever)
                {
                    this.wait.Set();
                    this.waitForever.WaitOne();
                }
                else
                {
                    this.controller.Prepare(0, BurnAction.Uninstall == this.model.Command.action ? RequestState.Absent : RequestState.Present);
                    wait.WaitOne();

                    this.controller.Cache(0);
                    wait.WaitOne();

                    this.controller.Execute(RequestState.None);
                    wait.WaitOne();
                }

                return TestConstants.Success;
            }

            public override bool OnPreparePackageBegin(BurnPackageTestShim package)
            {
                return true;
            }

            public override void OnPreparePackageComplete(BurnPackageTestShim package, int status)
            {
                this.RememberFirstErrorResult(status);
                ++this.packageCount;
            }

            public override void OnPrepareComplete(int status)
            {
                this.RememberFirstErrorResult(status);
                wait.Set();
            }

            public override bool OnCacheBegin(uint packages, uint payloads, UInt64 totalBytes)
            {
                return true;
            }

            public override void OnCacheComplete(int status)
            {
                this.RememberFirstErrorResult(status);
                wait.Set();
            }

            public override bool OnExecutePackageBegin(BurnPackageTestShim package, bool execute)
            {
                this.execute = execute;
                this.currentPercentage = 0;
                ++this.currentPackage;
                return true;
            }

            public override bool OnProgress(uint percentage, uint overallPercentage)
            {
                if (this.execute)
                {
                    Assert.IsTrue(percentage <= 100, "Progress percentage should never go over 100%.");
                    Assert.IsTrue(overallPercentage <= 100, "Overall progress percentage should never go over 100%.");

                    if (this.result == TestConstants.Success)
                    {
                        //Assert.IsTrue(this.currentPercentage <= overallPercentage, "Progress should never go backwards as long as the install was successful.");
                    }
                }
                else // rollback
                {
                    Assert.IsTrue(percentage <= 100, "Progress percentage should never go over 100%.");
                    Assert.IsTrue(overallPercentage <= 100, "Overall progress percentage should never go over 100%.");
                    Assert.IsTrue(this.currentPercentage >= percentage, "Progress should only go backwards during rollback.");
                }
                this.currentPercentage = percentage;

                return true;
            }

            public override void OnExecutePackageComplete(BurnPackageTestShim package, int status)
            {
                this.RememberFirstErrorResult(status);

                if (this.execute)
                {
                    if (this.result == TestConstants.Success)
                    {
                        Assert.IsTrue((this.totalPercentage + this.currentPercentage) / this.currentPackage <= 100, "Total progress percentage should never go over 100%.");
                        this.currentPercentage = 100;
                        this.totalPercentage += 100;
                    }
                    else
                    {
                        Assert.AreEqual(this.totalPercentage, (uint)(this.currentPackage - 1) * 100, "Total progress percentage should be equal to current package progress.");
                        this.currentPercentage = 0;
                    }
                }
                else
                {
                    this.currentPercentage = 0;
                    this.totalPercentage -= 100;
                }
            }

            public override void OnExecuteComplete(int status)
            {
                this.RememberFirstErrorResult(status);

                if (this.result == TestConstants.Success)
                {
                    if (this.execute)
                    {
                        Assert.IsTrue(this.totalPercentage == this.packageCount * 100, "Total progress should be 100% when execute is completed successfully.");
                    }
                    else
                    {
                        Assert.IsTrue(this.totalPercentage == 0, "Total progress should be 0% when rollback is completed.");
                    }
                }

                wait.Set();
            }

            public override DialogResult OnError(uint code, string message, MessageBoxButtons buttons, MessageBoxIcon icon)
            {
                this.RememberFirstErrorResult(TestConstants.InstallFailure);
                return DialogResult.Cancel;
            }

            private void RememberFirstErrorResult(int status)
            {
                if (TestConstants.Success == this.result)
                {
                    this.result = status;
                }
            }
        }

        /// <summary>
        /// Mock installation abstraction layer to capture elevated pipe name and token.
        /// </summary>
        private class MockInstallationAbstractionLayer : InstallationAbstractionLayerMockBase
        {
            private Thread mockElevatedProcess;
            private MockElevatedProcessResult result = new MockElevatedProcessResult();

            public string ExpectedToken { get; set; }
            public MockElevatedProcessResult Result { get { return this.result; } }

            /// <summary>
            /// Always say we could elevate since we want to mock the elevation with
            /// a thread below.
            /// </summary>
            /// <returns>S_FALSE</returns>
            public override int IsElevated()
            {
                return TestConstants.FalseSuccess;
            }

            /// <summary>
            /// Abstraction layer to execute elevated process.
            /// </summary>
            /// <param name="file">Path of file to launch.</param>
            /// <param name="parameters">Paramters passed to executable.</param>
            /// <returns>HRESULT</returns>
            public override int ShellExecuteW(string file, string parameters)
            {
                string[] args = parameters.Split(new string[] { " " }, StringSplitOptions.RemoveEmptyEntries);
                if (args.Length != 3)
                {
                    return TestConstants.InvalidData;
                }

                this.ExpectedToken = args[2];
                MockElevatedProcessContext context = new MockElevatedProcessContext() { PipeName = args[1], Result = this.result };

                this.mockElevatedProcess = new Thread(new ParameterizedThreadStart(MockInstallationAbstractionLayer.MockElevatedProcessThread));
                this.mockElevatedProcess.Start(context);

                return TestConstants.Success;
            }

            /// <summary>
            /// Wait for mock elevated process to complete.
            /// </summary>
            public void JoinMockElevatedProcess()
            {
                if (null != this.mockElevatedProcess)
                {
                    this.mockElevatedProcess.Join();
                    this.mockElevatedProcess = null;
                }
            }

            /// <summary>
            /// Thread to act as the elevated process started by ShellExecute.
            /// </summary>
            /// <param name="obj">Context for mock process.</param>
            private static void MockElevatedProcessThread(object obj)
            {
                MockElevatedProcessContext context = (MockElevatedProcessContext)obj;
                using (NamedPipeClientStream pipe = new NamedPipeClientStream(context.PipeName))
                {
                    pipe.Connect();

                    // First read the token size from the pipe.
                    byte[] data = new byte[4];
                    pipe.Read(data, 0, data.Length);

                    uint tokenSize = BitConverter.ToUInt32(data, 0);

                    // Read the token from the pipe and remember it in the result.
                    data = new byte[tokenSize];
                    pipe.Read(data, 0, data.Length);

                    context.Result.Token = Encoding.Unicode.GetString(data);

                    // Now send back the ACK.
                    uint ack = 1;
                    data = BitConverter.GetBytes(ack);
                    pipe.Write(data, 0, data.Length);
                }
            }

            /// <summary>
            /// Contained class to hold the results of the test.
            /// </summary>
            public class MockElevatedProcessResult
            {
                public string Token { get; set; }
            }

            /// <summary>
            /// Thread context for mock elevated process.
            /// </summary>
            private class MockElevatedProcessContext
            {
                public string PipeName { get; set; }
                public MockElevatedProcessResult Result { get; set; }
            }
        }

        /// <summary>
        /// Mock "elevated" view implementation.
        /// </summary>
        private class MockViewElevated : BurnViewMockBase
        {
            public bool Initialized { get; set; }
            public bool Ran { get; set; }
            public bool Uninitialized { get; set; }

            public override int Initialize(BurnModelTestShim model, BurnControllerTestShim controller)
            {
                this.Initialized = true;
                return TestConstants.Success;
            }

            public override int Run()
            {
                this.Ran = true;
                return TestConstants.Success;
            }

            public override void Uninitialize()
            {
                this.Uninitialized = true;
            }
        }

        /// <summary>
        /// Context passed to asynchronous controller start.
        /// </summary>
        private class AsyncControllerContext
        {
            private ManualResetEvent complete = new ManualResetEvent(false);

            public ManualResetEvent CompleteEvent { get { return this.complete; } }
            public BurnControllerTestShim Controller { get; set; }
            public MockBurnView View { get; set; }
        }
    }
}
