﻿// <copyright file="PropertiesTest.cs" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

namespace MS.Tools.WindowsInstallerXml.Test.Bootstrapper
{
    using System;

    using Microsoft.VisualStudio.TestTools.UnitTesting;

    /// <summary>
    /// Unit test for Burn properties.
    /// </summary>
    [TestClass]
    public class PropertiesTest
    {
        [TestMethod]
        public void PropertiesBasicTest()
        {
            using (BurnPropertiesTestShim properties = new BurnPropertiesTestShim())
            {
                properties.SetString("PROP1", "VAL1");

                properties.SetNumeric("PROP2", 2);
                properties.SetString("PROP5", "VAL5");
                properties.SetString("PROP3", "VAL3");
                properties.SetString("PROP4", "VAL4");
                properties.SetString("PROP6", "VAL6");
                properties.SetString("PROP7", "7");

                Assert.AreEqual("VAL1", properties.GetStringValueByName("PROP1"));
                Assert.AreEqual(2, properties.GetNumericValueByName("PROP2"));
                Assert.AreEqual("2", properties.GetStringValueByName("PROP2"));
                Assert.AreEqual("VAL3", properties.GetStringValueByName("PROP3"));
                Assert.AreEqual("VAL4", properties.GetStringValueByName("PROP4"));
                Assert.AreEqual("VAL5", properties.GetStringValueByName("PROP5"));
                Assert.AreEqual("VAL6", properties.GetStringValueByName("PROP6"));
                Assert.AreEqual(7, properties.GetNumericValueByName("PROP7"));
            }
        }

        [TestMethod]
        public void PropertiesFormatTest()
        {
            using (BurnPropertiesTestShim properties = new BurnPropertiesTestShim())
            {
                properties.SetString("PROP1", "VAL1");
                properties.SetString("PROP2", "VAL2");
                properties.SetNumeric("PROP3", 3);

                Assert.AreEqual("NOPROP", properties.FormatString("NOPROP"));
                Assert.AreEqual("VAL1", properties.FormatString("[PROP1]"));
                Assert.AreEqual(" VAL1 ", properties.FormatString(" [PROP1] "));
                Assert.AreEqual("PRE VAL1", properties.FormatString("PRE [PROP1]"));
                Assert.AreEqual("VAL1 POST", properties.FormatString("[PROP1] POST"));
                Assert.AreEqual("PRE VAL1 POST", properties.FormatString("PRE [PROP1] POST"));
                Assert.AreEqual("VAL1 MID VAL2", properties.FormatString("[PROP1] MID [PROP2]"));
                Assert.AreEqual("", properties.FormatString("[NONE]"));
                Assert.AreEqual("", properties.FormatString("[prop1]"));
                Assert.AreEqual("[", properties.FormatString(@"[\[]"));
                Assert.AreEqual("]", properties.FormatString(@"[\]]"));
                Assert.AreEqual("[]", properties.FormatString("[]"));
                Assert.AreEqual("[NONE", properties.FormatString("[NONE"));
                Assert.AreEqual("VAL2", properties.GetFormattedValueByName("PROP2"));
                Assert.AreEqual("3", properties.GetFormattedValueByName("PROP3"));
            }
        }

        [TestMethod]
        public void PropertiesEscapeTest()
        {
            using (BurnPropertiesTestShim properties = new BurnPropertiesTestShim())
            {
                properties.SetString("PROP1", "VAL1");
                properties.SetString("PROP2", "VAL2");

                Assert.AreEqual(@"[\[]", properties.EscapeString("["));
                Assert.AreEqual(@"[\]]", properties.EscapeString("]"));
                Assert.AreEqual(@" [\[]TEXT[\]] ", properties.EscapeString(" [TEXT] "));
            }
        }

        [TestMethod]
        public void PropertiesConditionTest()
        {
            using (BurnPropertiesTestShim properties = new BurnPropertiesTestShim())
            {
                properties.SetString("PROP1", "VAL1");
                properties.SetString("PROP2", "VAL2");
                properties.SetString("PROP3", "VAL3");
                properties.SetString("PROP4", "BEGIN MID END");
                properties.SetNumeric("PROP5", 5);
                properties.SetNumeric("PROP6", 6);
                properties.SetString("PROP7", "");
                properties.SetNumeric("PROP8", 0);
                properties.SetString("_PROP9", "VAL9");
                properties.SetNumeric("PROP10", -10);
                properties.SetNumeric("PROP11", 9223372036854775807);
                properties.SetNumeric("PROP12", -9223372036854775808);
                properties.SetNumeric("PROP13", 65536); // 0x000F0000
                properties.SetNumeric("PROP14", 1);// 0x0000000F
                properties.SetNumeric("PROP15", 65537); // 0x000F000F
                properties.SetVersion("PROP16", 0ul); // 0.0.0.0
                properties.SetVersion("PROP17", 1ul << 48); // 1.0.0.0
                properties.SetVersion("PROP18", (1ul << 48) | (1ul << 32)); // 1.1.0.0
                properties.SetVersion("PROP19", (1ul << 48) | (1ul << 32) | (1ul << 16)); // 1.1.1.0
                properties.SetVersion("PROP20", (1ul << 48) | (1ul << 32) | (1ul << 16) | 1ul); // 1.1.1.1
                properties.SetNumeric("vPROP21", 1);
                properties.SetVersion("PROP22", (65535ul << 48) | (65535ul << 32) | (65535ul << 16) | 65535ul); // 65535.65535.65535.65535

                Assert.IsTrue(properties.EvaluateCondition("PROP1"));
                Assert.IsTrue(properties.EvaluateCondition("PROP5"));
                Assert.IsFalse(properties.EvaluateCondition("PROP7"));
                Assert.IsFalse(properties.EvaluateCondition("PROP8"));
                Assert.IsTrue(properties.EvaluateCondition("_PROP9"));

                Assert.IsTrue(properties.EvaluateCondition("PROP1 = \"VAL1\""));
                Assert.IsFalse(properties.EvaluateCondition("NONE = \"NOT\""));
                Assert.IsFalse(properties.EvaluateCondition("PROP1 <> \"VAL1\""));
                Assert.IsTrue(properties.EvaluateCondition("NONE <> \"NOT\""));

                Assert.IsTrue(properties.EvaluateCondition("PROP5 = 5"));
                Assert.IsFalse(properties.EvaluateCondition("PROP5 = 0"));
                Assert.IsFalse(properties.EvaluateCondition("PROP5 <> 5"));
                Assert.IsTrue(properties.EvaluateCondition("PROP5 <> 0"));

                Assert.IsTrue(properties.EvaluateCondition("PROP10 = -10"));
                Assert.IsFalse(properties.EvaluateCondition("PROP10 <> -10"));

                Assert.IsTrue(properties.EvaluateCondition("PROP17 = v1"));
                Assert.IsFalse(properties.EvaluateCondition("PROP17 = v0"));
                Assert.IsFalse(properties.EvaluateCondition("PROP17 <> v1"));
                Assert.IsTrue(properties.EvaluateCondition("PROP17 <> v0"));

                Assert.IsTrue(properties.EvaluateCondition("PROP16 = v0"));
                Assert.IsTrue(properties.EvaluateCondition("PROP17 = v1"));
                Assert.IsTrue(properties.EvaluateCondition("PROP18 = v1.1"));
                Assert.IsTrue(properties.EvaluateCondition("PROP19 = v1.1.1"));
                Assert.IsTrue(properties.EvaluateCondition("PROP20 = v1.1.1.1"));
                Assert.IsFalse(properties.EvaluateCondition("PROP20 = v1.1.1.1.0"));
                Assert.IsFalse(properties.EvaluateCondition("PROP20 = v1.1.1.1.1"));
                Assert.IsTrue(properties.EvaluateCondition("vPROP21 = 1"));

                Assert.IsFalse(properties.EvaluateCondition("PROP11 = 9223372036854775806"));
                Assert.IsTrue(properties.EvaluateCondition("PROP11 = 9223372036854775807"));
                Assert.IsFalse(properties.EvaluateCondition("PROP11 = 9223372036854775808"));
                Assert.IsFalse(properties.EvaluateCondition("PROP11 = 92233720368547758070000"));

                Assert.IsFalse(properties.EvaluateCondition("PROP12 = -9223372036854775807"));
                Assert.IsTrue(properties.EvaluateCondition("PROP12 = -9223372036854775808"));
                Assert.IsFalse(properties.EvaluateCondition("PROP12 = -9223372036854775809"));
                Assert.IsFalse(properties.EvaluateCondition("PROP12 = -92233720368547758080000"));

                Assert.IsTrue(properties.EvaluateCondition("PROP22 = v65535.65535.65535.65535"));
                Assert.IsFalse(properties.EvaluateCondition("PROP22 = v65536.65535.65535.65535"));
                Assert.IsFalse(properties.EvaluateCondition("PROP22 = v65535.655350000.65535.65535"));

                Assert.IsTrue(properties.EvaluateCondition("PROP5 < 6"));
                Assert.IsFalse(properties.EvaluateCondition("PROP5 < 5"));
                Assert.IsTrue(properties.EvaluateCondition("PROP5 > 4"));
                Assert.IsFalse(properties.EvaluateCondition("PROP5 > 5"));
                Assert.IsTrue(properties.EvaluateCondition("PROP5 <= 6"));
                Assert.IsTrue(properties.EvaluateCondition("PROP5 <= 5"));
                Assert.IsFalse(properties.EvaluateCondition("PROP5 <= 4"));
                Assert.IsTrue(properties.EvaluateCondition("PROP5 >= 4"));
                Assert.IsTrue(properties.EvaluateCondition("PROP5 >= 5"));
                Assert.IsFalse(properties.EvaluateCondition("PROP5 >= 6"));

                Assert.IsTrue(properties.EvaluateCondition("PROP4 << \"BEGIN\""));
                Assert.IsFalse(properties.EvaluateCondition("PROP4 << \"END\""));
                Assert.IsTrue(properties.EvaluateCondition("PROP4 >> \"END\""));
                Assert.IsFalse(properties.EvaluateCondition("PROP4 >> \"BEGIN\""));
                Assert.IsTrue(properties.EvaluateCondition("PROP4 >< \"MID\""));
                Assert.IsFalse(properties.EvaluateCondition("PROP4 >< \"NONE\""));

                Assert.IsTrue(properties.EvaluateCondition("PROP16 < v1.1"));
                Assert.IsFalse(properties.EvaluateCondition("PROP16 < v0"));
                Assert.IsTrue(properties.EvaluateCondition("PROP17 > v0.12"));
                Assert.IsFalse(properties.EvaluateCondition("PROP17 > v1"));
                Assert.IsTrue(properties.EvaluateCondition("PROP18 >= v1.0"));
                Assert.IsTrue(properties.EvaluateCondition("PROP18 >= v1.1"));
                Assert.IsFalse(properties.EvaluateCondition("PROP18 >= v2.1"));
                Assert.IsTrue(properties.EvaluateCondition("PROP19 <= v1.1234.1"));
                Assert.IsTrue(properties.EvaluateCondition("PROP19 <= v1.1.1"));
                Assert.IsFalse(properties.EvaluateCondition("PROP19 <= v1.0.123"));

                Assert.IsTrue(properties.EvaluateCondition("PROP6 = \"6\""));
                Assert.IsTrue(properties.EvaluateCondition("\"6\" = PROP6"));
                Assert.IsFalse(properties.EvaluateCondition("PROP6 = \"ABC\""));
                Assert.IsFalse(properties.EvaluateCondition("\"ABC\" = PROP6"));
                Assert.IsFalse(properties.EvaluateCondition("\"ABC\" = PROP6"));

                Assert.IsTrue(properties.EvaluateCondition("PROP13 << 1"));
                Assert.IsFalse(properties.EvaluateCondition("PROP13 << 0"));
                Assert.IsTrue(properties.EvaluateCondition("PROP14 >> 1"));
                Assert.IsFalse(properties.EvaluateCondition("PROP14 >> 0"));
                Assert.IsTrue(properties.EvaluateCondition("PROP15 >< 65537"));
                Assert.IsFalse(properties.EvaluateCondition("PROP15 >< 0"));

                Assert.IsFalse(properties.EvaluateCondition("NOT PROP1"));
                Assert.IsTrue(properties.EvaluateCondition("NOT (PROP1 <> \"VAL1\")"));

                Assert.IsTrue(properties.EvaluateCondition("PROP1 = \"VAL1\" AND PROP2 = \"VAL2\""));
                Assert.IsFalse(properties.EvaluateCondition("PROP1 = \"VAL1\" AND PROP2 = \"NOT\""));
                Assert.IsFalse(properties.EvaluateCondition("PROP1 = \"NOT\" AND PROP2 = \"VAL2\""));
                Assert.IsFalse(properties.EvaluateCondition("PROP1 = \"NOT\" AND PROP2 = \"NOT\""));

                Assert.IsTrue(properties.EvaluateCondition("PROP1 = \"VAL1\" OR PROP2 = \"VAL2\""));
                Assert.IsTrue(properties.EvaluateCondition("PROP1 = \"VAL1\" OR PROP2 = \"NOT\""));
                Assert.IsTrue(properties.EvaluateCondition("PROP1 = \"NOT\" OR PROP2 = \"VAL2\""));
                Assert.IsFalse(properties.EvaluateCondition("PROP1 = \"NOT\" OR PROP2 = \"NOT\""));

                Assert.IsTrue(properties.EvaluateCondition("PROP1 = \"VAL1\" AND PROP2 = \"VAL2\" OR PROP3 = \"NOT\""));
                Assert.IsTrue(properties.EvaluateCondition("PROP1 = \"VAL1\" AND PROP2 = \"NOT\" OR PROP3 = \"VAL3\""));
                Assert.IsFalse(properties.EvaluateCondition("PROP1 = \"VAL1\" AND PROP2 = \"NOT\" OR PROP3 = \"NOT\""));
                Assert.IsTrue(properties.EvaluateCondition("PROP1 = \"VAL1\" AND (PROP2 = \"NOT\" OR PROP3 = \"VAL3\")"));
                Assert.IsTrue(properties.EvaluateCondition("(PROP1 = \"VAL1\" AND PROP2 = \"VAL2\") OR PROP3 = \"NOT\""));

                Assert.IsTrue(properties.EvaluateCondition("PROP3 = \"NOT\" OR PROP1 = \"VAL1\" AND PROP2 = \"VAL2\""));
                Assert.IsTrue(properties.EvaluateCondition("PROP3 = \"VAL3\" OR PROP1 = \"VAL1\" AND PROP2 = \"NOT\""));
                Assert.IsFalse(properties.EvaluateCondition("PROP3 = \"NOT\" OR PROP1 = \"VAL1\" AND PROP2 = \"NOT\""));
                Assert.IsTrue(properties.EvaluateCondition("(PROP3 = \"NOT\" OR PROP1 = \"VAL1\") AND PROP2 = \"VAL2\""));
                Assert.IsTrue(properties.EvaluateCondition("PROP3 = \"NOT\" OR (PROP1 = \"VAL1\" AND PROP2 = \"VAL2\")"));

                Assert.IsFalse(properties.EvaluateCondition("="));
                Assert.IsFalse(properties.EvaluateCondition("(PROP1"));
                Assert.IsFalse(properties.EvaluateCondition("(PROP1 = \""));
                Assert.IsFalse(properties.EvaluateCondition("1A"));
                Assert.IsFalse(properties.EvaluateCondition("*"));
            }
        }
    }
}
