//-----------------------------------------------------------------------
// <copyright file="InstanceTransforms.InstanceTransformTests.cs" company="Microsoft">
//     Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
// <summary>
//     Tests for the building an MSI with an Instance Transform
// </summary>
//-----------------------------------------------------------------------

namespace Microsoft.Tools.WindowsInstallerXml.Test.Tests.Integration.BuildingPackages.InstanceTransforms
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.IO;
    using System.Text;
    using Microsoft.VisualStudio.TestTools.UnitTesting;

    using Microsoft.Tools.WindowsInstallerXml.Test;
    using Microsoft.Deployment.WindowsInstaller;
    using Microsoft.Deployment.WindowsInstaller.Package;

    /// <summary>
    /// Tests for the building an MSI with an Instance Transform
    /// </summary>
    [TestClass]
    public class InstanceTransformTests
    {
        private static readonly string TestDataDirectory = Environment.ExpandEnvironmentVariables(@"%WIX%\test\data\Integration\BuildingPackages\InstanceTransforms\InstanceTransformTests");

        [TestMethod]
        [Description("Verify that a simple MSI with an instance transform can be built")]
        [Priority(1)]
        public void SimpleInstanceTransform()
        {
            string sourceFile = Path.Combine(InstanceTransformTests.TestDataDirectory, @"SimpleInstanceTransform\product.wxs");
            string msi = Builder.BuildPackage(sourceFile);

            // Verify that 3 instance transforms were created
            for (int i = 1; i <= 3; i++)
            {
                string transformName = String.Format("Instance{0}.mst", i);
                string mst = Path.Combine(Path.GetDirectoryName(msi), transformName);

                // Extract the transform
                InstanceTransformTests.ExtractTransform(msi, transformName, mst);

                // Make a copy of the base MSI
                string instanceMSI = Path.Combine(Path.GetDirectoryName(msi), String.Format("{0}{1}.msi", Path.GetFileNameWithoutExtension(msi), i));
                File.Copy(msi, instanceMSI);

                // Apply the transform to the copy of the base MSI
                InstallPackage instanceMSIPackage = new InstallPackage(instanceMSI, DatabaseOpenMode.Direct);
                instanceMSIPackage.ApplyTransform(mst);
                instanceMSIPackage.Close();

                // Verify that the transformed MSI matches what is expected
                string expectedInstanceMSI = Path.Combine(InstanceTransformTests.TestDataDirectory, String.Format(@"SimpleInstanceTransform\expected{0}.msi", i));
                Verifier.VerifyResults(expectedInstanceMSI, instanceMSI);
            }

            // Verify that the base MSI matches what is expected
            Verifier.VerifyResults(Path.Combine(InstanceTransformTests.TestDataDirectory, @"SimpleInstanceTransform\expected.msi"), msi);
        }

        [TestMethod]
        [Description("Verify that Product/@Id element can use a '*' for its GUID")]
        [Priority(1)]
        public void AutoGenProductId()
        {
            string sourceFile = Path.Combine(InstanceTransformTests.TestDataDirectory, @"AutoGenProductId\product.wxs");
            string msi = Builder.BuildPackage(sourceFile);

            // Verify that an instance transforms was created
            string transformName = "Instance1.mst";
            string mst = Path.Combine(Path.GetDirectoryName(msi), transformName);

            // Extract the transform
            InstanceTransformTests.ExtractTransform(msi, transformName, mst);

            // Verify that the transform matches the expected transform
            string expectedTransform = Path.Combine(InstanceTransformTests.TestDataDirectory, @"AutoGenProductId\expected.mst");
            Verifier.VerifyResults(expectedTransform, mst);
        }


        [TestMethod]
        [Description("Verify that there is an error when two instances have the same Id")]
        [Priority(2)]
        public void DuplicateInstanceIds()
        {
            string sourceFile = Path.Combine(InstanceTransformTests.TestDataDirectory, @"DuplicateInstanceIds\product.wxs");

            Candle candle = new Candle();
            candle.SourceFiles.Add(sourceFile);
            candle.Run();

            Light light = new Light(candle);
            light.ExpectedWixMessages.Add(new WixMessage(91, "Duplicate symbol 'WixInstanceTransforms:Instance1' found.", WixMessage.MessageTypeEnum.Error));
            light.ExpectedWixMessages.Add(new WixMessage(92, "Location of symbol related to previous error.", WixMessage.MessageTypeEnum.Error));
            light.ExpectedExitCode = 92;
            light.Run();
        }

        /// <summary>
        /// Extracts a transform from an MSI
        /// </summary>
        /// <param name="msi">The MSI to extract from</param>
        /// <param name="transform">The name of the transform to extract</param>
        /// <param name="extractFile">The location to extract to</param>
        /// <remarks>
        /// Most of this code was copied from DTF's Microsoft.Deployment.WindowsInstaller.Package.PatchPackage class.
        /// </remarks>
        private static void ExtractTransform(string msi, string transform, string extractFile)
        {
            Database msiDatabase = new Database(msi, DatabaseOpenMode.ReadOnly);

            using (View view = msiDatabase.OpenView("SELECT `Name`, `Data` FROM `_Storages` WHERE `Name` = '{0}'", transform))
            {
                view.Execute();
                Record record = view.Fetch();
                if (record == null)
                {
                    Assert.Fail("Transform {0} not found in {1}", transform, msi);
                }
                using (record)
                {
                    Console.WriteLine("Extracting transform {0} to {1}", transform, extractFile);
                    record.GetStream("Data", extractFile);
                }
            }
        }
    }
}
